from .worksheet_outputer import WorksheetOutputer
from xl.common.colors import is_valid_rgb

from openpyxl import Workbook
from openpyxl.styles import PatternFill, Border, Side, Alignment, Font, borders
from openpyxl.styles.numbers import BUILTIN_FORMATS_REVERSE
from openpyxl.styles.fills import FILL_SOLID, FILL_NONE
from openpyxl.styles.cell_style import StyleArray
from openpyxl.utils.cell import SHEETRANGE_RE, quote_sheetname, absolute_coordinate
import json
import os


# pylint: disable=too-many-statements

class WorkbookOutputer:

    def __init__(self, basedir):
        self.wb = Workbook(write_only=True)
        self._base_dir = basedir
        self._config = self._get_json('config')
        self._title_map = {info.get('id'): info.get('title')
                           for info in self._config.get('sheets')}
        self._image_info = self._get_json('image_info')

    def save(self, filename):
        self.write()
        self.wb.save(filename)

    def write(self):
        self._fill_styles(self._get_style())
        self._fill_defined_names(self._get_define())
        sheets_info = self._get_sheets_info()

        for info in sheets_info:
            sheet_id = info.get('id')
            title = info.get('title', '')
            tab_color = info.get('tabColor')
            hidden = info.get('hidden', False)

            ws = self.wb.create_sheet(title=title)
            props = ws.sheet_properties
            if is_valid_rgb(tab_color):
                props.tabColor = tab_color
            if hidden:
                ws.sheet_state = 'hidden'

            snapshot_file = os.path.join(self._base_dir, sheet_id)
            WorksheetOutputer(self, ws, snapshot_file).write()

    def _get_json(self, filename):
        file_path = os.path.join(self._base_dir, filename)
        with open(file_path, 'r', encoding='utf-8') as f:
            return json.load(f)

    def _get_define(self):
        define_path = self._config.get('define')
        return self._get_json(define_path) if define_path else None

    def _get_style(self):
        style_path = self._config.get('style')
        return self._get_json(style_path) if style_path else None

    def _get_sheets_info(self):
        return self._config.get('sheets', [])

    def get_title(self, sheet_id):
        return self._title_map.get(sheet_id, '')

    def get_image_info(self, image_id):
        config = self._image_info.get(image_id)
        if not config or not config.get('name'):
            return None
        path = os.path.join(self._base_dir, config.get('name'))
        return (path, config.get('ext'))

    def _fill_styles(self, stylesheet):
        wb = self.wb
        for f in stylesheet['fonts']:
            if f is None:
                f = {}
            font = Font(name=f.get('n'), b=f.get('b'),
                        i=f.get('i'), strike=f.get('s'))
            # if f.get('sz', 10) != 10:
            font.sz = f.get('sz', 10)
            if f.get('u', False):
                font.u = Font.UNDERLINE_SINGLE
            if is_valid_rgb(f.get('c')):
                font.color = 'FF' + f.get('c').upper()
            list.append(wb._fonts, font)

        for f in stylesheet['fills']:
            if f is None:
                f = {}
            fill = PatternFill(patternType=FILL_NONE)
            start_color = f.get('bg')
            if is_valid_rgb(start_color):
                fill.patternType = FILL_SOLID
                fill.fgColor = 'FF' + start_color.upper()
            list.append(wb._fills, fill)

        for a in stylesheet['alignments']:
            if a is None:
                a = {}
            alignment = Alignment()
            # skip legacy value: start
            if a.get('h') and a.get('h') in ('right', 'left', 'center'):
                alignment.horizontal = a.get('h')
            if a.get('v'):
                alignment.vertical = 'center' if a.get(
                    'v') == 'middle' else a.get('v')
            if a.get('w') == 'wrap':
                alignment.wrapText = True
            list.append(wb._alignments, alignment)

        def make_border_side(color):
            if color and is_valid_rgb(color[1:]):
                return Side(border_style=borders.BORDER_THIN, color='FF' + color[1:].upper())
            return None

        for b in stylesheet['borders']:
            colors = b.get('b') if b else None
            if colors is None:
                colors = []
            while len(colors) < 4:
                colors.append(None)
            # openpyxl Border.bottom can't be None
            border = Border(
                top=make_border_side(colors[0]),
                right=make_border_side(colors[1]),
                bottom=make_border_side(colors[2]) if is_valid_rgb(colors[2]) else Side(),
                left=make_border_side(colors[3]),
                diagonal=None
            )
            list.append(wb._borders, border)

        nf_idx_map = [0]
        for n in stylesheet['numberFormats']:
            fc = n.get('fc') if n else None
            if not fc:
                fc = 'General'
            if fc in BUILTIN_FORMATS_REVERSE:
                idx = BUILTIN_FORMATS_REVERSE[fc]
            else:
                idx = wb._number_formats.add(fc) + 164
            nf_idx_map.append(idx)

        for s in stylesheet['styles']:
            font_id = s[1]
            # two first fills are predefined
            fill_id = s[2] + 1 if s[2] > 0 else 0
            alignment_id = s[3]
            border_id = s[4]
            nf_id = nf_idx_map[s[5]]
            quote_prefix = s[6] if len(s) > 6 else 0
            # style_aray: (font, fill, border, nf, protection, alignment,
            # pivotButton, quotePrefix, xfId)
            style = StyleArray(
                [font_id, fill_id, border_id, nf_id, 0, alignment_id, 0, quote_prefix, 0])
            list.append(wb._cell_styles, style)

    def _fill_defined_names(self, define):
        if not define:
            return
        order_map = {info.get('id'): idx for idx,
                     info in enumerate(self._get_sheets_info())}
        sheet_title_set = set(self._title_map.values())

        for key, nameList in define.items():
            local_sheet_id = order_map.get(key, None)
            for name, value in nameList.items():
                m = SHEETRANGE_RE.match(value)
                if m is None:
                    continue
                sheet_name = m.group('quoted') or m.group('notquoted')
                ref = m.group('cells')
                if sheet_name in sheet_title_set:
                    range_value = '%s!%s' % (quote_sheetname(sheet_name), ':'.join(
                        map(absolute_coordinate, ref.split(':'))))
                    self.wb.create_named_range(
                        name=name, value=range_value, scope=local_sheet_id)
