from openpyxl.comments import Comment
from openpyxl.comments.comment_sheet import CommentRecord
from openpyxl.utils import get_column_letter, range_boundaries
from openpyxl.worksheet.hyperlink import Hyperlink
from openpyxl.worksheet.cell_range import CellRange, MultiCellRange
from openpyxl.formatting import Rule
from openpyxl.styles.differential import DifferentialStyle
from openpyxl.styles import Font, PatternFill
from openpyxl.styles.fills import FILL_SOLID
from openpyxl.worksheet.datavalidation import DataValidationList, DataValidation
from openpyxl.worksheet.filters import (
    FilterColumn,
    Filters,
    CustomFilter,
    CustomFilters,
)
from openpyxl.cell.text import (
    Text,
    RichText,
    InlineFont,
)
from openpyxl.cell.cell import (
    TYPE_STRING,
    TYPE_FORMULA,
    TYPE_NUMERIC,
    TYPE_BOOL,
)
from openpyxl.drawing.spreadsheet_drawing import AbsoluteAnchor
from openpyxl.drawing.xdr import (
    XDRPoint2D,
    XDRPositiveSize2D
)
from openpyxl.utils.units import pixels_to_EMU

from .cell import SynoCell
from .image import SynoImage
from .chart_outputer import create_chart
from ..common.constant import (
    ROW_HEIGHT_RATIO,
    COL_WIDTH_RATIO,
    ROW_HEADER_WIDTH,
    COL_HEADER_HEIGHT,
    COL_WIDTH,
)
from xl.common.colors import is_valid_rgb
import re
from copy import copy
from native.json import get_snapshot_without_cells, parse_cells
import json


def get_coords_text(row, col):
    return '%s%d' % (get_column_letter(col + 1), row + 1)


def get_range_text(min_row, min_col, max_row, max_col):
    return '%s:%s' % (get_coords_text(min_row, min_col),
                      get_coords_text(max_row, max_col))


single_word_re = re.compile(r'^\w$')


def escape_sheet_name(name):
    if re.match(r'^\w+$', name):
        return name
    return "'" + str.replace("'", "''", name) + "'"


hyperlink_re = re.compile(r'=\s*HYPERLINK', re.I)
internal_hyperlink_re = re.compile(
    r'=\s*HYPERLINK\(\s*"#tid=(\d+)"\s*,\s*"(([^"]|"")+)"\s*\)', re.I)


def pos_to_anchor(pos):
    offset = XDRPoint2D(pixels_to_EMU(pos.get('left')),
                        pixels_to_EMU(pos.get('top')))
    ext = XDRPositiveSize2D(pixels_to_EMU(
        pos.get('width')), pixels_to_EMU(pos.get('height')))
    return AbsoluteAnchor(pos=offset, ext=ext)


class WorksheetOutputer:

    def __init__(self, parent, ws, snapshot_file):
        self.parent = parent
        self.ws = ws
        self._cell = SynoCell(ws)
        self._snapshot_file = snapshot_file
        self._snapshot = None

    def write(self):
        json_str, bound = get_snapshot_without_cells(self._snapshot_file)
        dimension = get_range_text(*bound)
        self._snapshot = json.loads(json_str)

        self._freeze_panes()
        self._fill_dimensions()
        self._fill_cfs()
        self._fill_validations()
        self._fill_filter()
        self._fill_drawings()
        self.ws.write_data(self, dimension=dimension)

    def parse_cells(self, on_cell):
        parse_cells(self._snapshot_file, on_cell)

    def cells(self, row, row_idx):
        for col_idx, cell_json_str in row:
            cell_json = json.loads(cell_json_str)
            if cell_json is None:
                continue

            xf = cell_json.get('s', 0)
            value = cell_json.get('v', None)

            cell = self._cell
            cell.row = row_idx
            cell.column = col_idx + 1

            cell.xf_id = xf

            if cell_json.get('n'):
                self._create_comment(cell.coordinate, cell_json.get('n'))

            value_type = None

            if value is not None and value != '':
                if cell.quotePrefix:
                    value_type = TYPE_STRING
                elif isinstance(value, (int, float)):
                    value_type = TYPE_NUMERIC
                elif isinstance(value, (str, bytes)):
                    value_type = TYPE_STRING
                    if len(value) > 1 and value.startswith("="):
                        value_type = TYPE_FORMULA
                elif value is True or value is False:
                    value_type = TYPE_BOOL
                elif isinstance(value, dict):
                    value_type = TYPE_STRING
                    value = self._create_richtext(value)

            if value_type is not None:
                cell.set_explicit_value(value, value_type)
            else:
                cell.value = value

            if value_type is TYPE_FORMULA and re.match(hyperlink_re, value):
                self._create_hyperlink(cell, value)

            yield cell

    def _create_richtext(self, value):
        if value.get('t') != 'r' or not isinstance(value.get('v'), list):
            return ''

        texts = []
        for text_run in value.get('v'):
            tx = text_run.get('tx')
            if tx is None or not len:
                continue
            font = None
            s = text_run.get('s')
            if s is not None:
                font = InlineFont(rFont=s.get('n'), b=s.get('b'),
                                  i=s.get('i'), strike=s.get('s'))
                if s.get('sz') != 10:
                    font.sz = s.get('sz')
                if s.get('u', False):
                    font.u = Font.UNDERLINE_SINGLE
                if is_valid_rgb(s.get('c')):
                    font.color = s.get('c').upper()
            texts.append(RichText(t=tx, rPr=font))

        return Text(r=texts) if texts else ''

    def _create_hyperlink(self, cell, value):
        self._apply_hyperlink_style(cell)

        matches = re.match(internal_hyperlink_re, value)
        if matches:
            sheet_id = 'sh_' + matches.group(1)
            display = str.replace('""', '"', matches.group(2))
            sheet_name = self.parent.get_title(sheet_id)
            cell.set_explicit_value(display, TYPE_STRING)
            if sheet_name:
                location = 'sheet://%s!A1' % escape_sheet_name(sheet_name)
                hyperlink = Hyperlink(location=location, tooltip=sheet_name, ref=cell.coordinate)
                self.ws._hyperlinks.append(hyperlink)

    def _apply_hyperlink_style(self, cell):
        wb = self.ws.parent
        style = wb._cell_styles[cell.xf_id]
        font_id = style[0]
        font = wb._fonts[font_id]
        BLACK = '000000'
        font_color = BLACK
        if font.color and isinstance(font.color.value, str):
            font_color = font.color.value[-6:]
        if not font.u or font.color is None or font_color == BLACK:
            new_font = copy(font)
            new_font.u = Font.UNDERLINE_SINGLE
            # BLUE
            new_font.color = 'FF0000FF'
            new_font_id = font_id
            if new_font in wb._fonts:
                new_font_id = wb._fonts.index(new_font)
            else:
                new_font_id = len(wb._fonts)
                wb._fonts.append(new_font)
            new_style = copy(style)
            new_style[0] = new_font_id
            new_xf_id = cell.xf_id
            if new_style in wb._cell_styles:
                new_xf_id = wb._cell_styles.index(new_style)
            else:
                new_xf_id = len(wb._cell_styles)
                wb._cell_styles.append(new_style)
            cell.xf_id = new_xf_id

    def _create_comment(self, ref, content):
        comment = Comment(content, 'author')
        record = CommentRecord(ref=ref, author=comment.author)
        record.text.t = comment.content
        record.height = comment.height
        record.width = comment.width
        self.ws._comments.append(record)

    def _freeze_panes(self):
        left = self._snapshot.get('fixedColumnLeft', 0)
        top = self._snapshot.get('fixedRowTop', 0)
        if left > 0 or top > 0:
            self.ws.freeze_panes = get_coords_text(top, left)

    def _fill_dimensions(self):
        global_xf = (self._snapshot.get('global') or {}).get('s', 0)

        self.ws.sheet_format.defaultColWidth = COL_WIDTH * COL_WIDTH_RATIO
        cols = self._snapshot.get('cols', {})
        col_indexes = []
        for c, col in cols.items():
            if col is None:
                continue
            col_idx = int(c)
            col_dim = self.ws.column_dimensions[get_column_letter(col_idx + 1)]
            col_dim.min = col_dim.max = col_idx + 1
            col_dim.width = (col.get('w') if col.get(
                'w') else COL_WIDTH) * COL_WIDTH_RATIO
            if col.get('hidden', False):
                col_dim.hidden = True
            if col.get('s', 0) > 0:
                col_dim.xf_id = col.get('s')
            col_indexes.append(col_idx)

        if global_xf > 0:
            # fill empty cols
            last_col_index = -1
            col_indexes.sort()
            for col_idx in col_indexes:
                if col_idx > last_col_index + 1:
                    col = self.ws.column_dimensions[get_column_letter(
                        last_col_index + 2)]
                    col.xf_id = global_xf
                    col.min = last_col_index + 2
                    col.max = col_idx
                    col.width = COL_WIDTH * COL_WIDTH_RATIO
                last_col_index = col_idx
            last = self.ws.column_dimensions[get_column_letter(
                last_col_index + 2)]
            last.xf_id = global_xf
            last.width = COL_WIDTH * COL_WIDTH_RATIO
            last.min = last_col_index + 2
            last.max = 16384

        rows = self._snapshot.get('rows', {})
        for r, row in rows.items():
            if row is None:
                continue
            row_idx = int(r)
            row_dim = self.ws.row_dimensions[row_idx + 1]
            if row.get('h'):
                row_dim.ht = row.get('h') * ROW_HEIGHT_RATIO
            if row.get('hidden', False):
                row_dim.hidden = True
            if row.get('s', 0) > 0:
                row_dim.xf_id = row.get('s')

    def get_merged_cells(self):
        return [get_range_text(*info) for info in self._snapshot.get('mergeCells', [])]

    def _get_sqref(self, ranges):
        result = []
        for r in ranges:
            min_col, min_row, max_col, max_row = range_boundaries(r)
            if min_col is None:
                min_col = 1
            if min_row is None:
                min_row = 1
            if max_col is None:
                max_col = self._snapshot.get('colCount')
            if max_row is None:
                max_row = self._snapshot.get('rowCount')
            cr = CellRange(min_col=min_col, min_row=min_row, max_col=max_col, max_row=max_row)
            result.append(cr)
        return MultiCellRange(result)

    def _fill_cfs(self):
        cfs = self._snapshot.get('cfs')
        if cfs is None or not cfs:
            return

        default_font = Font()
        for index, cf in enumerate(cfs):
            ranges = cf.get('ranges')
            if ranges is None or not ranges:
                continue

            style = cf['style']
            font = Font()
            if style.get('bold'):
                font.b = True
            if style.get('italic'):
                font.i = True
            if style.get('strike'):
                font.strike = True
            if style.get('underline'):
                font.u = Font.UNDERLINE_SINGLE
            if is_valid_rgb(style.get('color')):
                font.color = 'FF' + style['color'].upper()

            if font == default_font:
                font = None

            fill = None
            if is_valid_rgb(style.get('bgColor')):
                fill = PatternFill(fill_type=FILL_SOLID,
                                   end_color=style['bgColor'].upper())

            dxf = DifferentialStyle(font=font, fill=fill)
            rule = Rule(type=cf['type'], formula=[cf['f1']],
                        priority=index + 1, dxf=dxf)
            if cf.get('op'):
                rule.operator = cf['op']
            if cf.get('f2'):
                rule.formula.append(cf['f2'])
            if cf.get('text'):
                rule.text = cf['text']

            sqref_str = str(self._get_sqref(ranges))
            self.ws.conditional_formatting.add(sqref_str, rule)

    def _fill_validations(self):
        validations = self._snapshot.get('validation')
        if not validations:
            return

        data_validations = DataValidationList()

        TYPES = ('list', 'decimal', 'date', 'text', 'custom')
        OPERATORS = ('between', 'notBetween', 'equal', 'notEqual',
                     'greaterThan', 'greaterThanOrEqual', 'lessThan', 'lessThanOrEqual')
        for _, validation in validations.items():
            if not validation.get('ranges', []):
                continue

            valid_type = None
            if validation.get('type') in TYPES:
                valid_type = validation.get('type')
                if valid_type == 'text':
                    valid_type = 'custom'

            valid_op = None
            op = validation.get('operator')
            if op in OPERATORS:
                valid_op = op
            elif op == 'list':
                validation['f1'] = '"%s"' % (validation.get('f1'))

            dv = DataValidation(
                type=valid_type,
                operator=valid_op,
                formula1=validation.get('f1'),
                formula2=validation['f2'] if validation.get('f2') else None,
                # Excel show dropdown menu when showDropDown is false
                showDropDown=False if valid_type == 'list' else None,
                showErrorMessage=True, allowBlank=True,
                errorStyle='stop' if validation.get(
                    'reject', False) else 'warning',
                sqref=self._get_sqref(validation.get('ranges'))
            )
            if validation.get('message'):
                dv.showInputMessage = True
                dv.prompt = validation['message']
            data_validations.append(dv)

        self.ws.data_validations = data_validations

    def _fill_filter(self):
        filter_data = self._snapshot.get('filter')
        if not filter_data or not filter_data.get('enabled'):
            return

        self.ws.auto_filter.ref = get_range_text(*filter_data.get('range'))
        for col_idx, cond in enumerate(filter_data.get('filter', [])):
            if cond is None:
                continue
            cond_type = cond.get('type')
            column = None
            values = cond.get('value', [])
            if cond_type == 'value':
                column = FilterColumn(
                    colId=col_idx, filters=Filters(filter=values))
            elif cond_type == 'custom':
                custom_filters = []
                operators = cond.get('operator', [])
                for idx, values in enumerate(values):
                    if idx > len(operators):
                        break
                    custom_filters.append(CustomFilter(
                        operator=operators[idx], val=values[idx]))

                column = FilterColumn(colId=col_idx,
                                      customFilters=CustomFilters(
                                              _and=cond.get('and', False), customFilter=custom_filters
                                          ))
            elif cond_type == 'text' and values:
                text = values[0]
                op = cond.get('operator', [])[0]
                if op in ('containsTexts', 'notContainsTexts'):
                    text = '*' + text + '*'
                elif op == 'beginsWith':
                    text = text + '*'
                elif op == 'endsWith':
                    text = '*' + text
                elif op == 'exactTexts':
                    pass
                custom_filter = CustomFilter(
                    operator='notEqual' if op == 'notContainsTexts' else 'equal', val=text)
                column = FilterColumn(colId=col_idx,
                                      customFilters=CustomFilters(customFilter=[custom_filter]))

            if column:
                self.ws.auto_filter.filterColumn.append(column)

    def _fill_drawings(self):
        drawings = self._snapshot.get('drawing')
        if not drawings:
            return
        for drawing in drawings:
            drawing_type = drawing.get('type')
            if drawing_type == 'img':
                self._fill_image(drawing)
            elif drawing_type == 'chart':
                self._fill_chart(drawing)

    def _fill_image(self, drawing):
        image_id = drawing.get('id')
        image_info = self.parent.get_image_info(image_id)
        if image_info is None:
            return
        (path, file_ext) = image_info
        pos = drawing.get('pos')
        pos['left'] -= ROW_HEADER_WIDTH
        pos['top'] -= COL_HEADER_HEIGHT
        anchor = pos_to_anchor(pos)
        self.ws.add_image(SynoImage(path, format=file_ext, width=pos.get(
            'width'), height=pos.get('height')), anchor=anchor)

    def _fill_chart(self, drawing):
        chart = create_chart(drawing.get('chartObj'))
        if chart is None:
            return
        self.ws.add_chart(chart, pos_to_anchor(drawing.get('pos')))
