<?php
/**
 * This file is part of PHPPresentation - A pure PHP library for reading and writing
 * presentations documents.
 *
 * PHPPresentation is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPPresentation/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPPresentation
 * @copyright   2009-2015 PHPPresentation contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpPresentation\Reader;

use PhpOffice\PhpPresentation\DocumentLayout;
use PhpOffice\PhpPresentation\PhpPresentation;
use PhpOffice\PhpPresentation\Shape\Chart\Series;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Area;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Bar;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Line;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Pie;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Scatter;
use PhpOffice\PhpPresentation\Shape\Comment\Author;
use PhpOffice\PhpPresentation\Shape\Group;
use PhpOffice\PhpPresentation\Shape\Placeholder;
use PhpOffice\PhpPresentation\Shape\RichText;
use PhpOffice\PhpPresentation\Shape\RichText\Paragraph;
use PhpOffice\PhpPresentation\Shape\Table\Cell;
use PhpOffice\PhpPresentation\Slide;
use PhpOffice\PhpPresentation\Slide\AbstractSlide;
use PhpOffice\PhpPresentation\Slide\Note;
use PhpOffice\PhpPresentation\Slide\SlideLayout;
use PhpOffice\PhpPresentation\Slide\SlideMaster;
use PhpOffice\PhpPresentation\Slide\Transition;
use PhpOffice\PhpPresentation\Slide\Animation;
use PhpOffice\PhpPresentation\Shape\Drawing\Gd;
use PhpOffice\PhpPresentation\Style\Alignment;
use PhpOffice\PhpPresentation\Style\Bullet;
use PhpOffice\PhpPresentation\Style\Border;
use PhpOffice\PhpPresentation\Style\Borders;
use PhpOffice\PhpPresentation\Style\Color;
use PhpOffice\PhpPresentation\Style\Fill;
use PhpOffice\PhpPresentation\Style\Font;
use PhpOffice\PhpPresentation\Style\SchemeColor;
use PhpOffice\PhpPresentation\Style\TextStyle;
use PhpOffice\PhpPresentation\Style\TableStyles;
use PhpOffice\PhpPresentation\Style\TableStyle;
use PhpOffice\Common\XMLReader;
use PhpOffice\Common\Drawing as CommonDrawing;
use ZipArchive;

/**
 * Serialized format reader
 */
class PowerPoint2007 implements ReaderInterface
{
    /**
     * Output Object
     * @var PhpPresentation
     */
    protected $oPhpPresentation;
    /**
     * Output Object
     * @var \ZipArchive
     */
    protected $oZip;
    /**
     * @var array[]
     */
    protected $arrayRels = array();
    /**
     * @var SlideLayout[]
     */
    protected $arraySlideLayouts = array();
    /*
     * @var string
     */
    protected $filename;
    /*
     * @var string
     */
    protected $fileRels;

    private $authorMap = array();
    private $loadingCount = 0;
    private $shapeMap = array();

    /**
     * Can the current \PhpOffice\PhpPresentation\Reader\ReaderInterface read the file?
     *
     * @param  string $pFilename
     * @throws \Exception
     * @return boolean
     */
    public function canRead($pFilename)
    {
        return $this->fileSupportsUnserializePhpPresentation($pFilename);
    }

    /**
     * Does a file support UnserializePhpPresentation ?
     *
     * @param  string $pFilename
     * @throws \Exception
     * @return boolean
     */
    public function fileSupportsUnserializePhpPresentation($pFilename = '')
    {
        // Check if file exists
        if (!file_exists($pFilename)) {
            throw new \Exception("Could not open " . $pFilename . " for reading! File does not exist.");
        }

        $oZip = new ZipArchive();
        // Is it a zip ?
        if ($oZip->open($pFilename) === true) {
            // Is it an OpenXML Document ?
            // Is it a Presentation ?
            if (is_array($oZip->statName('[Content_Types].xml')) && is_array($oZip->statName('ppt/presentation.xml'))) {
                return true;
            }
        }
        return false;
    }

    /**
     * Loads PhpPresentation Serialized file
     *
     * @param  string $pFilename
     * @return \PhpOffice\PhpPresentation\PhpPresentation
     * @throws \Exception
     */
    public function load($pFilename)
    {
        // Unserialize... First make sure the file supports it!
        if (!$this->fileSupportsUnserializePhpPresentation($pFilename)) {
            throw new \Exception("Invalid file format for PhpOffice\PhpPresentation\Reader\PowerPoint2007: " . $pFilename . ".");
        }

        return $this->loadFile($pFilename);
    }

    /**
     * Load PhpPresentation Serialized file
     *
     * @param  string $pFilename
     * @return \PhpOffice\PhpPresentation\PhpPresentation
     */
    protected function loadFile($pFilename)
    {
        try {
            $this->oPhpPresentation = new PhpPresentation();
            $this->oPhpPresentation->removeSlideByIndex();
            $this->oPhpPresentation->setAllMasterSlides(array());
            $this->filename = $pFilename;

            $this->oZip = new ZipArchive();
            $this->oZip->open($this->filename);
            $docPropsCore = $this->oZip->getFromName('docProps/core.xml');
            if ($docPropsCore !== false) {
                $this->loadDocumentProperties($docPropsCore);
            }

            $docPropsCustom = $this->oZip->getFromName('docProps/custom.xml');
            if ($docPropsCustom !== false) {
                $this->loadCustomProperties($docPropsCustom);
            }

            $pptViewProps = $this->oZip->getFromName('ppt/viewProps.xml');
            if ($pptViewProps !== false) {
                $this->loadViewProperties($pptViewProps);
            }

            if (false !== $this->oZip->locateName('ppt/tableStyles.xml')) {
                $pptTableStyles = $this->oZip->getFromName('ppt/tableStyles.xml');
                $this->loadTableStyles($pptTableStyles);
            }

            if (false !== $this->oZip->locateName('ppt/commentAuthors.xml')) {
                $commentAuthors = $this->oZip->getFromName('ppt/commentAuthors.xml');
                $this->loadCommentAuthors($commentAuthors);
            }

            $pptPresentation = $this->oZip->getFromName('ppt/presentation.xml');
            if ($pptPresentation !== false) {
                $this->loadDocumentLayout($pptPresentation);
                $this->loadDefaultTextStyle($pptPresentation);
                $this->loadSlides($pptPresentation);
            }
        } catch (\Exception $e) {
            echo $e->getTraceAsString() . "\n";
            echo "loading count: " . $this->loadingCount . "\n";
        }

        return $this->oPhpPresentation;
    }

    /**
     * Read Document Layout
     * @param $sPart
     */
    protected function loadDocumentLayout($sPart)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            foreach ($xmlReader->getElements('/p:presentation/p:sldSz') as $oElement) {
                if (!($oElement instanceof \DOMElement)) {
                    continue;
                }
                $type = $oElement->getAttribute('type');
                $oLayout = $this->oPhpPresentation->getLayout();
                if ($type == DocumentLayout::LAYOUT_CUSTOM) {
                    $oLayout->setCX($oElement->getAttribute('cx'));
                    $oLayout->setCY($oElement->getAttribute('cy'));
                } else {
                    $oLayout->setDocumentLayout($type, true);
                    if ($oElement->getAttribute('cx') < $oElement->getAttribute('cy')) {
                        $oLayout->setDocumentLayout($type, false);
                    }
                }
            }
        }
    }

    private function loadDefaultTextStyle($sPart)
    {
        $xmlReader = new XMLReader();
        if (!$xmlReader->getDomFromString($sPart)) {
            return;
        }
        foreach ($xmlReader->getElements('/p:presentation/p:defaultTextStyle/*') as $oElement) {
            if ('a:defPPr' === $oElement->nodeName) {
                continue;
            }
            $level = $this->parseParagraphLevel($oElement);
            $oParagraph = new Paragraph();
            if ($oElement->hasAttribute('algn')) {
                $oParagraph->getAlignment()->setHorizontal($oElement->getAttribute('algn'));
            }
            $oDefRunProp = $xmlReader->getElement('a:defRPr', $oElement);
            if (!($oDefRunProp instanceof \DOMElement)) {
                continue;
            }
            $this->loadTextRunProperty($xmlReader, $oDefRunProp, $oParagraph->getFont(), null);
            $this->oPhpPresentation->getListStyle()->setStyleAtLvl($level, $oParagraph);
        }
    }

    /**
     * Read Document Properties
     * @param string $sPart
     */
    protected function loadDocumentProperties($sPart)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            $arrayProperties = array(
                '/cp:coreProperties/dc:creator' => 'setCreator',
                '/cp:coreProperties/cp:lastModifiedBy' => 'setLastModifiedBy',
                '/cp:coreProperties/dc:title' => 'setTitle',
                '/cp:coreProperties/dc:description' => 'setDescription',
                '/cp:coreProperties/dc:subject' => 'setSubject',
                '/cp:coreProperties/cp:keywords' => 'setKeywords',
                '/cp:coreProperties/cp:category' => 'setCategory',
                '/cp:coreProperties/dcterms:created' => 'setCreated',
                '/cp:coreProperties/dcterms:modified' => 'setModified',
            );
            $oProperties = $this->oPhpPresentation->getDocumentProperties();
            foreach ($arrayProperties as $path => $property) {
                $oElement = $xmlReader->getElement($path);
                if ($oElement instanceof \DOMElement) {
                    if ($oElement->hasAttribute('xsi:type') && $oElement->getAttribute('xsi:type') == 'dcterms:W3CDTF') {
                        $oDateTime = \DateTime::createFromFormat(\DateTime::W3C, $oElement->nodeValue);
                        $oProperties->{$property}($oDateTime->getTimestamp());
                    } else {
                        $oProperties->{$property}($oElement->nodeValue);
                    }
                }
            }
        }
    }

    /**
     * Read Custom Properties
     * @param string $sPart
     */
    protected function loadCustomProperties($sPart)
    {
        $xmlReader = new XMLReader();
        $sPart = str_replace(' xmlns="http://schemas.openxmlformats.org/officeDocument/2006/custom-properties"', '', $sPart);
        if ($xmlReader->getDomFromString($sPart)) {
            $pathMarkAsFinal = '/Properties/property[@pid="2"][@fmtid="{D5CDD505-2E9C-101B-9397-08002B2CF9AE}"][@name="_MarkAsFinal"]/vt:bool';
            if (is_object($oElement = $xmlReader->getElement($pathMarkAsFinal))) {
                if ($oElement->nodeValue == 'true') {
                    $this->oPhpPresentation->getPresentationProperties()->markAsFinal(true);
                }
            }
        }
    }

    /**
     * Read View Properties
     * @param string $sPart
     */
    protected function loadViewProperties($sPart)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            $pathZoom = '/p:viewPr/p:slideViewPr/p:cSldViewPr/p:cViewPr/p:scale/a:sx';
            $oElement = $xmlReader->getElement($pathZoom);
            if ($oElement instanceof \DOMElement) {
                if ($oElement->hasAttribute('d') && $oElement->hasAttribute('n')) {
                    $this->oPhpPresentation->getPresentationProperties()->setZoom($oElement->getAttribute('n') / $oElement->getAttribute('d'));
                }
            }
        }
    }

    /**
     * Extract all slides
     */
    protected function loadSlides($sPart)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            $fileRels = 'ppt/_rels/presentation.xml.rels';
            $this->loadRels($fileRels);
            // Load the Masterslides
            $this->loadMasterSlides($xmlReader, $fileRels);

            $this->loadingCount = 0;
            // Continue with loading the slides
            foreach ($xmlReader->getElements('/p:presentation/p:sldIdLst/p:sldId') as $oElement) {
                $this->loadingCount++;
                if (!($oElement instanceof \DOMElement)) {
                    continue;
                }
                $rId = $oElement->getAttribute('r:id');
                $pathSlide = isset($this->arrayRels[$fileRels][$rId]) ? $this->arrayRels[$fileRels][$rId]['Target'] : '';
                if (!empty($pathSlide)) {
                    $pptSlide = $this->oZip->getFromName('ppt/' . $pathSlide);
                    if ($pptSlide !== false) {
                        $slideRels = 'ppt/slides/_rels/' . basename($pathSlide) . '.rels';
                        $this->loadRels($slideRels);
                        $this->loadSlide($pptSlide, basename($pathSlide));
                        foreach ($this->arrayRels[$slideRels] as $rel) {
                            if ($rel['Type'] == 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/notesSlide') {
                                $this->loadSlideNote(basename($rel['Target']), $this->oPhpPresentation->getActiveSlide());
                            } else if ($rel['Type'] === 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/comments') {
                                $this->loadComments(basename($rel['Target']), $this->oPhpPresentation->getActiveSlide());
                            }
                        }
                    }
                }
            }
        }
    }

    /**
     * Extract all MasterSlides
     * @param XMLReader $xmlReader
     * @param string $fileRels
     */
    protected function loadMasterSlides(XMLReader $xmlReader, $fileRels)
    {
        // Get all the MasterSlide Id's from the presentation.xml file
        foreach ($xmlReader->getElements('/p:presentation/p:sldMasterIdLst/p:sldMasterId') as $oElement) {
            if (!($oElement instanceof \DOMElement)) {
                continue;
            }
            $rId = $oElement->getAttribute('r:id');
            // Get the path to the masterslide from the array with _rels files
            $pathMasterSlide = isset($this->arrayRels[$fileRels][$rId]) ?
                $this->arrayRels[$fileRels][$rId]['Target'] : '';
            if (!empty($pathMasterSlide)) {
                $pptMasterSlide = $this->oZip->getFromName('ppt/' . $pathMasterSlide);
                if ($pptMasterSlide !== false) {
                    $this->loadRels('ppt/slideMasters/_rels/' . basename($pathMasterSlide) . '.rels');
                    $this->loadMasterSlide($pptMasterSlide, basename($pathMasterSlide));
                }
            }
        }
    }

    /**
     * Extract data from slide
     * @param string $sPart
     * @param string $baseFile
     */
    protected function loadSlide($sPart, $baseFile)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            // Core
            $oSlide = $this->oPhpPresentation->createSlide();
            $this->oPhpPresentation->setActiveSlideIndex($this->oPhpPresentation->getSlideCount() - 1);
            $oSlide->setRelsIndex('ppt/slides/_rels/' . $baseFile . '.rels');
            $this->fileRels = $oSlide->getRelsIndex();

            $oElement = $xmlReader->getElement('/p:sld');
            if ($oElement instanceof \DOMElement && '0' === $oElement->getAttribute('show')) {
                $oSlide->setIsVisible(false);
            }

            // Background
            $oElement = $xmlReader->getElement('/p:sld/p:cSld/p:bg');
            if ($oElement instanceof \DOMElement) {
                $this->loadSlideBackground($xmlReader, $oElement, $oSlide);
            }

            // Shapes
            $arrayElements = $xmlReader->getElements('/p:sld/p:cSld/p:spTree/*');
            if ($arrayElements) {
                $this->loadSlideShapes($oSlide, $arrayElements, $xmlReader);
            }

            // Transition
            $xmlReader->registerNs('mc', 'http://schemas.openxmlformats.org/markup-compatibility/2006');
            ($oElement = $xmlReader->getElement('/p:sld/mc:AlternateContent/mc:Choice/p:transition')) || ($oElement = $xmlReader->getElement('/p:sld/p:transition'));
            if ($oElement instanceof \DOMElement) {
                $this->loadTransition($xmlReader, $oElement, $oSlide);
            }

            // Animations
            $oElement = $xmlReader->getElement('/p:sld/p:timing');
            if ($oElement instanceof \DOMElement) {
                $this->loadAnimations($xmlReader, $oElement, $oSlide);
            }

            // Layout
            foreach ($this->arrayRels['ppt/slides/_rels/' . $baseFile . '.rels'] as $valueRel) {
                if ($valueRel['Type'] == 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/slideLayout') {
                    $layoutBasename = basename($valueRel['Target']);
                    if (array_key_exists($layoutBasename, $this->arraySlideLayouts)) {
                        $oSlide->setSlideLayout($this->arraySlideLayouts[$layoutBasename]);
                    }
                    break;
                }
            }
        }
    }

    protected function loadTransition(XMLReader $xmlReader, \DOMElement $oElement, AbstractSlide $oSlide)
    {
        $oTypeElement = $xmlReader->getElement('*', $oElement);
        if (!$oTypeElement instanceof \DOMElement) {
            return;
        }
        $type;
        switch ($oTypeElement->nodeName) {
            case 'p:blinds':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'horz';
                switch ($dir) {
                    case 'horz':
                        $type = Transition::TRANSITION_BLINDS_HORIZONTAL;
                        break;
                    case 'vert':
                        $type = Transition::TRANSITION_BLINDS_VERTICAL;
                        break;
                }
                break;
            case 'p:checker':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'horz';
                switch ($dir) {
                    case 'horz':
                        $type = Transition::TRANSITION_CHECKER_HORIZONTAL;
                        break;
                    case 'vert':
                        $type = Transition::TRANSITION_CHECKER_VERTICAL;
                        break;
                }
                break;
            case 'p:circle':
                $type = Transition::TRANSITION_CIRCLE;
                break;
            case 'p:comb':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'horz';
                switch ($dir) {
                    case 'horz':
                        $type = Transition::TRANSITION_COMB_HORIZONTAL;
                        break;
                    case 'vert':
                        $type = Transition::TRANSITION_COMB_VERTICAL;
                        break;
                }
                break;
            case 'p:cover':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'l';
                switch ($dir) {
                    case 'd':
                        $type = Transition::TRANSITION_COVER_DOWN;
                        break;
                    case 'l':
                        $type = Transition::TRANSITION_COVER_LEFT;
                        break;
                    case 'ld':
                        $type = Transition::TRANSITION_COVER_LEFT_DOWN;
                        break;
                    case 'lu':
                        $type = Transition::TRANSITION_COVER_LEFT_UP;
                        break;
                    case 'r':
                        $type = Transition::TRANSITION_COVER_RIGHT;
                        break;
                    case 'rd':
                        $type = Transition::TRANSITION_COVER_RIGHT_DOWN;
                        break;
                    case 'ru':
                        $type = Transition::TRANSITION_COVER_RIGHT_UP;
                        break;
                    case 'u':
                        $type = Transition::TRANSITION_COVER_UP;
                        break;
                }
                break;
            case 'p:cut':
                $thruBlk = $oTypeElement->hasAttribute('thruBlk') ? $oTypeElement->getAttribute('thruBlk') : 'false';
                switch ($thruBlk) {
                    case '1':
                    case 'true':
                        $type = Transition::TRANSITION_CUT_THROUGH_BLACK;
                        break;
                    case '0':
                    case 'false':
                        $type = Transition::TRANSITION_CUT;
                        break;
                }
                break;
            case 'p:diamond':
                $type = Transition::TRANSITION_DIAMOND;
                break;
            case 'p:dissolve':
                $type = Transition::TRANSITION_DISSOLVE;
                break;
            case 'p:fade':
                $thruBlk = $oTypeElement->hasAttribute('thruBlk') ? $oTypeElement->getAttribute('thruBlk') : 'false';
                switch ($thruBlk) {
                    case '1':
                    case 'true':
                        $type = Transition::TRANSITION_FADE_THROUGH_BLACK;
                        break;
                    case '0':
                    case 'false':
                        $type = Transition::TRANSITION_FADE;
                        break;
                }
                break;
            case 'p:newsflask':
                $type = Transition::TRANSITION_NEWSFLASH;
                break;
            case 'p:plus':
                $type = Transition::TRANSITION_PLUS;
                break;
            case 'p:pull':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'l';
                switch ($dir) {
                    case 'd':
                        $type = Transition::TRANSITION_PULL_DOWN;
                        break;
                    case 'l':
                        $type = Transition::TRANSITION_PULL_LEFT;
                        break;
                    case 'ld':
                        $type = Transition::TRANSITION_PULL_LEFT_DOWN;
                        break;
                    case 'lu':
                        $type = Transition::TRANSITION_PULL_LEFT_UP;
                        break;
                    case 'r':
                        $type = Transition::TRANSITION_PULL_RIGHT;
                        break;
                    case 'rd':
                        $type = Transition::TRANSITION_PULL_RIGHT_DOWN;
                        break;
                    case 'ru':
                        $type = Transition::TRANSITION_PULL_RIGHT_UP;
                        break;
                    case 'u':
                        $type = Transition::TRANSITION_PULL_UP;
                        break;
                }
                break;
            case 'p:push':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'l';
                switch ($dir) {
                    case 'd':
                        $type = Transition::TRANSITION_PUSH_DOWN;
                        break;
                    case 'l':
                        $type = Transition::TRANSITION_PUSH_LEFT;
                        break;
                    case 'r':
                        $type = Transition::TRANSITION_PUSH_RIGHT;
                        break;
                    case 'u':
                        $type = Transition::TRANSITION_PUSH_UP;
                        break;
                }
                break;
            case 'p:random':
                $type = Transition::TRANSITION_RANDOM;
                break;
            case 'p:randomBar':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'horz';
                switch ($dir) {
                    case 'horz':
                        $type = Transition::TRANSITION_RANDOMBAR_HORIZONTAL;
                        break;
                    case 'vert':
                        $type = Transition::TRANSITION_RANDOMBAR_VERTICAL;
                        break;
                }
                break;
            case 'p:split':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'out';
                $orient = $oTypeElement->hasAttribute('orient') ? $oTypeElement->getAttribute('orient') : 'horz';
                switch ($dir . $orient) {
                    case 'inhorz':
                        $type = Transition::TRANSITION_SPLIT_IN_HORIZONTAL;
                        break;
                    case 'outhorz':
                        $type = Transition::TRANSITION_SPLIT_OUT_HORIZONTAL;
                        break;
                    case 'invert':
                        $type = Transition::TRANSITION_SPLIT_IN_VERTICAL;
                        break;
                    case 'outvert':
                        $type = Transition::TRANSITION_SPLIT_OUT_VERTICAL;
                        break;
                }
                break;
            case 'p:strips':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'lu';
                switch ($dir) {
                    case 'ld':
                        $type = Transition::TRANSITION_STRIPS_LEFT_DOWN;
                        break;
                    case 'lu':
                        $type = Transition::TRANSITION_STRIPS_LEFT_UP;
                        break;
                    case 'rd':
                        $type = Transition::TRANSITION_STRIPS_RIGHT_DOWN;
                        break;
                    case 'ru':
                        $type = Transition::TRANSITION_STRIPS_RIGHT_UP;
                        break;
                }
                break;
            case 'p:wedge':
                $type = Transition::TRANSITION_WEDGE;
                break;
            case 'p:wipe':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'l';
                switch ($dir) {
                    case 'd':
                        $type = Transition::TRANSITION_WIPE_DOWN;
                        break;
                    case 'l':
                        $type = Transition::TRANSITION_WIPE_LEFT;
                        break;
                    case 'r':
                        $type = Transition::TRANSITION_WIPE_RIGHT;
                        break;
                    case 'u':
                        $type = Transition::TRANSITION_WIPE_UP;
                        break;
                }
                break;
            case 'p:zoom':
                $dir = $oTypeElement->hasAttribute('dir') ? $oTypeElement->getAttribute('dir') : 'out';
                switch ($dir) {
                    case 'in':
                        $type = Transition::TRANSITION_ZOOM_IN;
                        break;
                    case 'out':
                        $type = Transition::TRANSITION_ZOOM_OUT;
                        break;
                }
                break;
        }
        if (is_null($type)) {
            return;
        }
        $oTransition = new Transition();
        $advClick = $oElement->hasAttribute('advClick') ? $oElement->getAttribute('advClick') : 'true';
        $oTransition->setManualTrigger($advClick == '1' || $advClick == 'true');
        if ($oElement->hasAttribute('advTm')) {
            $oTransition->setTimeTrigger(true, intval($oElement->getAttribute('advTm')));
        }
        $spd = $oElement->hasAttribute('spd') ? $oElement->getAttribute('spd') : 'fast';
        $oTransition->setSpeed($spd);
        if ($oElement->hasAttribute('p14:dur')) {
            $oTransition->setDuration(intval($oElement->getAttribute('p14:dur')));
        }
        $oTransition->setTransitionType($type);
        $oSlide->setTransition($oTransition);
    }

    protected function loadAnimations(XMLReader $xmlReader, \DOMElement $oElement, AbstractSlide $oSlide)
    {
        // Microsoft & Google output animations in this form
        $oAnimationElements = $xmlReader->getElements('p:tnLst/p:par/p:cTn/p:childTnLst/p:seq/p:cTn/p:childTnLst/p:par/p:cTn/p:childTnLst/p:par/p:cTn/p:childTnLst/p:par/p:cTn', $oElement);
        foreach ($oAnimationElements as $oAnimationElement) {
            $presetID;
            if ($oAnimationElement->hasAttribute('presetID')) {
                $presetID = $oAnimationElement->getAttribute('presetID');
            }
            $presetClass;
            if ($oAnimationElement->hasAttribute('presetClass')) {
                $presetClass = $oAnimationElement->getAttribute('presetClass');
            }
            $nodeType;
            if ($oAnimationElement->hasAttribute('nodeType')) {
                $nodeType = $oAnimationElement->getAttribute('nodeType');
            }
            $delay;
            $oCondElement = $xmlReader->getElement('p:stCondLst/p:cond', $oAnimationElement);
            if ($oCondElement && $oCondElement->hasAttribute('delay')) {
                $delay = $oCondElement->getAttribute('delay');
            }
            $spid;
            $oSpTgtElement = $xmlReader->getElement('p:childTnLst/p:set/p:cBhvr/p:tgtEl/p:spTgt', $oAnimationElement);
            if ($oSpTgtElement && $oSpTgtElement->hasAttribute('spid')) {
                $spid = $oSpTgtElement->getAttribute('spid');
            }
            $dur;
            $oCtnElement = $xmlReader->getElement('p:childTnLst/p:animEffect/p:cBhvr/p:cTn', $oAnimationElement);
            if ($oCtnElement && $oCtnElement->hasAttribute('dur')) {
                $dur = $oCtnElement->getAttribute('dur');
            }
            $oAnimation = new Animation();
            if (!array_key_exists($spid, $this->shapeMap)) {
                continue;
            }
            $oAnimation->setShape($this->shapeMap[$spid]);
            switch ($presetID . $presetClass) {
                case '1entr':
                    $oAnimation->setType(Animation::ANIMATION_APPEAR);
                    $oAnimation->setDuration($delay);
                    break;
                case '1exit':
                    $oAnimation->setType(Animation::ANIMATION_DISAPPEAR);
                    $oAnimation->setDuration($delay);
                    break;
                case '10entr':
                    $oAnimation->setType(Animation::ANIMATION_FADE_IN);
                    $oAnimation->setDuration($dur);
                    break;
                case '10exit':
                    $oAnimation->setType(Animation::ANIMATION_FADE_OUT);
                    $oAnimation->setDuration($dur);
                    break;
                default:
                    continue 2;
                    break;
            }
            switch ($nodeType) {
                case 'clickEffect':
                    $oAnimation->setCondition(Animation::CONDITION_CLICK);
                    break;
                case 'withEffect':
                    $oAnimation->setCondition(Animation::CONDITION_WITH);
                    break;
                case 'afterEffect':
                    $oAnimation->setCondition(Animation::CONDITION_AFTER);
                    break;
                default:
                    continue 2;
                    break;
            }
            $oSlide->addAnimation($oAnimation);
        }
    }

    /**
     * @param string $sPart
     * @param string $baseFile
     */
    protected function loadMasterSlide($sPart, $baseFile)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            // Core
            $oSlideMaster = $this->oPhpPresentation->createMasterSlide();
            $oSlideMaster->setTextStyles(new TextStyle(false));
            $oSlideMaster->setRelsIndex('ppt/slideMasters/_rels/' . $baseFile . '.rels');
            $this->fileRels = $oSlideMaster->getRelsIndex();

            // Load the theme
            foreach ($this->arrayRels[$oSlideMaster->getRelsIndex()] as $arrayRel) {
                if ($arrayRel['Type'] == 'http://schemas.openxmlformats.org/officeDocument/2006/relationships/theme') {
                    $pptTheme = $this->oZip->getFromName('ppt/' . substr($arrayRel['Target'], strrpos($arrayRel['Target'], '../') + 3));
                    if ($pptTheme !== false) {
                        $this->loadTheme($pptTheme, $oSlideMaster);
                    }
                    break;
                }
            }

            // TextStyles
            $arrayElementTxStyles = $xmlReader->getElements('/p:sldMaster/p:txStyles/*');
            if ($arrayElementTxStyles) {
                foreach ($arrayElementTxStyles as $oElementTxStyle) {
                    $arrayElementsLvl = $xmlReader->getElements('/p:sldMaster/p:txStyles/' . $oElementTxStyle->nodeName . '/*');
                    foreach ($arrayElementsLvl as $oElementLvl) {
                        if (!($oElementLvl instanceof \DOMElement) || $oElementLvl->nodeName == 'a:extLst') {
                            continue;
                        }
                        $oRTParagraph = new Paragraph();

                        $level = $this->parseParagraphLevel($oElementLvl);
                        if (false === $level) {
                            continue;
                        }
                        $this->loadParagraphProperty($xmlReader, $oElementLvl, $oRTParagraph, $oSlideMaster);

                        switch ($oElementTxStyle->nodeName) {
                            case 'p:bodyStyle':
                                $oSlideMaster->getTextStyles()->setBodyStyleAtLvl($oRTParagraph, $level);
                                break;
                            case 'p:otherStyle':
                                $oSlideMaster->getTextStyles()->setOtherStyleAtLvl($oRTParagraph, $level);
                                break;
                            case 'p:titleStyle':
                                $oSlideMaster->getTextStyles()->setTitleStyleAtLvl($oRTParagraph, $level);
                                break;
                        }
                    }
                }
            }

            // Background
            $oElement = $xmlReader->getElement('/p:sldMaster/p:cSld/p:bg');
            if ($oElement instanceof \DOMElement) {
                $this->loadSlideBackground($xmlReader, $oElement, $oSlideMaster);
            }

            // Shapes
            $arrayElements = $xmlReader->getElements('/p:sldMaster/p:cSld/p:spTree/*');
            if ($arrayElements) {
                $this->loadSlideShapes($oSlideMaster, $arrayElements, $xmlReader);
            }
            // Header & Footer

            // ColorMapping
            $colorMap = array();
            $oElement = $xmlReader->getElement('/p:sldMaster/p:clrMap');
            if ($oElement->hasAttributes()) {
                foreach ($oElement->attributes as $attr) {
                    $colorMap[$attr->nodeName] = $attr->nodeValue;
                }
                $oSlideMaster->colorMap->setMapping($colorMap);
            }

            $this->loadingCount = 0;
            // Load the Layoutslide
            foreach ($xmlReader->getElements('/p:sldMaster/p:sldLayoutIdLst/p:sldLayoutId') as $oElement) {
                $this->loadingCount++;
                if (!($oElement instanceof \DOMElement)) {
                    continue;
                }
                $rId = $oElement->getAttribute('r:id');
                // Get the path to the masterslide from the array with _rels files
                $pathLayoutSlide = isset($this->arrayRels[$oSlideMaster->getRelsIndex()][$rId]) ?
                    $this->arrayRels[$oSlideMaster->getRelsIndex()][$rId]['Target'] : '';
                if (!empty($pathLayoutSlide)) {
                    $pptLayoutSlide = $this->oZip->getFromName('ppt/' . substr($pathLayoutSlide, strrpos($pathLayoutSlide, '../') + 3));
                    if ($pptLayoutSlide !== false) {
                        $this->loadRels('ppt/slideLayouts/_rels/' . basename($pathLayoutSlide) . '.rels');
                        $oSlideMaster->addSlideLayout(
                            $this->loadLayoutSlide($pptLayoutSlide, basename($pathLayoutSlide), $oSlideMaster)
                        );
                    }
                }
            }
        }
    }

    /**
     * @param string $sPart
     * @param string $baseFile
     * @param SlideMaster $oSlideMaster
     * @return SlideLayout|null
     */
    protected function loadLayoutSlide($sPart, $baseFile, SlideMaster $oSlideMaster)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            // Core
            $oSlideLayout = new SlideLayout($oSlideMaster);
            $oSlideLayout->setRelsIndex('ppt/slideLayouts/_rels/' . $baseFile . '.rels');
            $this->fileRels = $oSlideLayout->getRelsIndex();

            // Name
            $oElement = $xmlReader->getElement('/p:sldLayout/p:cSld');
            if ($oElement instanceof \DOMElement && $oElement->hasAttribute('name')) {
                $oSlideLayout->setLayoutName($oElement->getAttribute('name'));
            }

            // Background
            $oElement = $xmlReader->getElement('/p:sldLayout/p:cSld/p:bg');
            if ($oElement instanceof \DOMElement) {
                $this->loadSlideBackground($xmlReader, $oElement, $oSlideLayout);
            }

            // ColorMapping
            $oElement = $xmlReader->getElement('/p:sldLayout/p:clrMapOvr/a:overrideClrMapping');
            if ($oElement instanceof \DOMElement && $oElement->hasAttributes()) {
                $colorMap = array();
                foreach ($oElement->attributes as $attr) {
                    $colorMap[$attr->nodeName] = $attr->nodeValue;
                }
                $oSlideLayout->colorMap->setMapping($colorMap);
            }

            // Shapes
            $oElements = $xmlReader->getElements('/p:sldLayout/p:cSld/p:spTree/*');
            if ($oElements) {
                $this->loadSlideShapes($oSlideLayout, $oElements, $xmlReader);
            }
            $this->arraySlideLayouts[$baseFile] = &$oSlideLayout;
            return $oSlideLayout;
        }
        return null;
    }

    /**
     * @param string $sPart
     * @param SlideMaster $oSlideMaster
     */
    protected function loadTheme($sPart, SlideMaster $oSlideMaster)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            $oElements = $xmlReader->getElements('/a:theme/a:themeElements/a:clrScheme/*');
            if ($oElements) {
                foreach ($oElements as $oElement) {
                    $oSchemeColor = new SchemeColor();
                    $oSchemeColor->setValue(str_replace('a:', '', $oElement->tagName));
                    $colorElement = $xmlReader->getElement('*', $oElement);
                    if ($colorElement instanceof \DOMElement) {
                        if ($colorElement->hasAttribute('lastClr')) {
                            $oSchemeColor->setRGB($colorElement->getAttribute('lastClr'));
                        } elseif ($colorElement->hasAttribute('val')) {
                            $oSchemeColor->setRGB($colorElement->getAttribute('val'));
                        }
                    }
                    $oSlideMaster->addSchemeColor($oSchemeColor);
                }
            }
            $oMajorFonts = $xmlReader->getElements('/a:theme/a:themeElements/a:fontScheme/a:majorFont/*');
            if ($oMajorFonts) {
                foreach ($oMajorFonts as $oFont) {
                    if ('a:font' !== $oFont->nodeName) {
                        $oSlideMaster->addSchemeFont(true, $oFont->nodeName, $oFont->getAttribute('typeface'));
                    } else {
                        $oSlideMaster->addSchemeFont(true, $oFont->getAttribute('script'), $oFont->getAttribute('typeface'));
                    }
                }
            }
            $oMinorFonts = $xmlReader->getElements('/a:theme/a:themeElements/a:fontScheme/a:minorFont/*');
            if ($oMinorFonts) {
                foreach ($oMinorFonts as $oFont) {
                    if ('a:font' !== $oFont->nodeName) {
                        $oSlideMaster->addSchemeFont(false, $oFont->nodeName, $oFont->getAttribute('typeface'));
                    } else {
                        $oSlideMaster->addSchemeFont(false, $oFont->nodeName, $oFont->getAttribute('typeface'));
                    }
                }
            }
        }
    }

    /**
     * @param XMLReader $xmlReader
     * @param \DOMElement $oElement
     * @param AbstractSlide $oSlide
     */
    protected function loadSlideBackground(XMLReader $xmlReader, \DOMElement $oElement, AbstractSlide $oSlide)
    {
        $oElementNoFill = $xmlReader->getElement('p:bgPr/a:noFill', $oElement);
        if ($oElementNoFill instanceof \DOMElement) {
            $oColor = new Color();
            $oColor->setRGB('000000', '00');
            $oBackground = new Slide\Background\Color();
            $oBackground->setColor($oColor);
            $oSlide->setBackground($oBackground);
        }

        // Background color
        $oElementSolidFill = $xmlReader->getElement('p:bgPr/a:solidFill', $oElement);
        if ($oElementSolidFill instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oElementSolidFill);
            if (isset($oColor)) {
                if ($oColor instanceof SchemeColor) {
                    $oBgSchemeColor = new Slide\Background\SchemeColor();
                    $oBgSchemeColor->setSchemeColor($oColor);
                    $oSlide->setBackground($oBgSchemeColor);
                } else {
                    $oBgColor = new Slide\Background\Color();
                    $oBgColor->setColor($oColor);
                    $oSlide->setBackground($oBgColor);
                }
            }
        }

        // Background scheme color
        $oElementSchemeColor = $xmlReader->getElement('p:bgRef/a:schemeClr', $oElement);
        if ($oElementSchemeColor instanceof \DOMElement) {
            // Color
            $oColor = new SchemeColor();
            $oColor->setValue($oElementSchemeColor->hasAttribute('val') ? $oElementSchemeColor->getAttribute('val') : null);
            // Background
            $oBackground = new Slide\Background\SchemeColor();
            $oBackground->setSchemeColor($oColor);
            // Slide Background
            $oSlide->setBackground($oBackground);
        }

        // Background image
        $oElementImage = $xmlReader->getElement('p:bgPr/a:blipFill/a:blip', $oElement);
        if ($oElementImage instanceof \DOMElement) {
            $relImg = $this->arrayRels[$oSlide->getRelsIndex()][$oElementImage->getAttribute('r:embed')];
            if (is_array($relImg)) {
                // File
                $pathImage = 'ppt/slides/' . $relImg['Target'];
                $pathImage = explode('/', $pathImage);
                foreach ($pathImage as $key => $partPath) {
                    if ($partPath == '..') {
                        unset($pathImage[$key - 1]);
                        unset($pathImage[$key]);
                    }
                }
                $pathImage = implode('/', $pathImage);
                $contentImg = $this->oZip->getFromName($pathImage);

                $tmpBkgImg = tempnam(sys_get_temp_dir(), 'PhpPresentationReaderPpt2007Bkg');
                file_put_contents($tmpBkgImg, $contentImg);
                // Background
                $oBackground = new Slide\Background\Image();
                $oBackground->setPath($tmpBkgImg);

                $oElementAlpha = $xmlReader->getElement('a:alphaModFix', $oElementImage);
                if ($oElementAlpha && $oElementAlpha->hasAttribute('amt')) {
                    $amt = $oElementAlpha->getAttribute('amt');
                    if (preg_match('/\d+/', $amt)) {
                        $oBackground->setAlpha(floatVal($amt) / 1000);
                    } else if (preg_match('/\d+%/', $amt)) {
                        $oBackground->setAlpha(floatVal($amt));
                    }
                }

                // Slide Background
                $oSlide->setBackground($oBackground);
            }
        }
    }

    /**
     * @param string $baseFile
     * @param Slide $oSlide
     */
    protected function loadSlideNote($baseFile, Slide $oSlide)
    {
        $sPart = $this->oZip->getFromName('ppt/notesSlides/' . $baseFile);
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sPart)) {
            $oNote = $oSlide->getNote();

            $arrayElements = $xmlReader->getElements('/p:notes/p:cSld/p:spTree/*');
            if ($arrayElements) {
                $this->loadSlideShapes($oNote, $arrayElements, $xmlReader);
            }
        }
    }

    private function loadComments($baseFile, Slide $oSlide)
    {
        $sComment = $this->oZip->getFromName('ppt/comments/' . $baseFile);
        $xmlReader = new XMLReader();
        if (!$xmlReader->getDomFromString($sComment)) {
            return;
        }

        foreach ($xmlReader->getElements('/p:cmLst/p:cm') as $oElementCm) {
            $oComment = $oSlide->createComment();

            $authorName = '';
            if ($oElementCm->hasAttribute('authorId')) {
                $authorName = $this->authorMap[$oElementCm->getAttribute('authorId')];
            }
            $oAuthor = new Author();
            $oAuthor->setName($authorName);
            $oComment->setAuthor($oAuthor);

            if ($oElementCm->hasAttribute('dt')) {
                $oComment->setDate(strtotime($oElementCm->getAttribute('dt')));
            }

            if ($oElementCm->hasAttribute('idx')) {
                $oComment->setIndex($oElementCm->getAttribute('idx'));
            }

            $oElementText = $xmlReader->getElement('p:text', $oElementCm);
            if ($oElementText instanceof \DOMElement) {
                $oComment->setText($oElementText->nodeValue);
            }

            $xmlReader->registerNs('p15', 'http://schemas.microsoft.com/office/powerpoint/2012/main');
            $oElementParentCm = $xmlReader->getElement('p:extLst/p:ext/p15:threadingInfo/p15:parentCm', $oElementCm);
            if ($oElementParentCm instanceof \DOMElement && $oElementParentCm->hasAttribute('idx')) {
                $oComment->setParentIndex($oElementParentCm->getAttribute('idx'));
                $oComment->setIsReply(true);
            }
        }
    }

    /**
     * @param XMLReader $document
     * @param \DOMElement $node
     * @param AbstractSlide $oSlide
     */
    protected function loadShapeDrawing(XMLReader $document, \DOMElement $node, $oContainer, AbstractSlide $oSlide)
    {
        // Core
        $oShape = new Gd();
        $oShape->getShadow()->setVisible(false);

        $oElement = $document->getElement('p:nvPicPr/p:cNvPr', $node);
        $this->registerShapeId($oElement, $oShape);
        if ($oElement instanceof \DOMElement) {
            $oShape->setName($oElement->hasAttribute('name') ? $oElement->getAttribute('name') : '');
            $oShape->setDescription($oElement->hasAttribute('descr') ? $oElement->getAttribute('descr') : '');

            // Hyperlink
            $oHyperlink = $oShape->getHyperlink();
            $this->loadHyperlink($document, $oElement, $oHyperlink);
        }

        $oElement = $document->getElement('p:nvPicPr/p:nvPr/p:ph', $node);
        if ($oElement instanceof \DOMElement && $oElement->hasAttribute('idx')) {
            $placeholder = new Placeholder(Placeholder::PH_TYPE_PIC);
            $holderIdx = ($oElement->hasAttribute('idx')) ? intval($oElement->getAttribute('idx')) : 0;
            $placeholder->setIdx($holderIdx);
            $oShape->setPlaceHolder($placeholder);
        }

        $oElement = $document->getElement('p:blipFill/a:blip', $node);
        if ($oElement instanceof \DOMElement) {
            $imageSource = $this->loadBlipFill($oElement);
            if (!empty($imageSource)) {
                $oShape->setSource($imageSource);
            }
            $oElementAlpha = $document->getElement('a:alphaModFix', $oElement);
            if ($oElementAlpha && $oElementAlpha->hasAttribute('amt')) {
                $amt = $oElementAlpha->getAttribute('amt');
                if (preg_match('/\d+/', $amt)) {
                    $oShape->setAlpha(floatVal($amt) / 1000);
                } else if (preg_match('/\d+%/', $amt)) {
                    $oShape->setAlpha(floatVal($amt));
                }
            }
        }

        $oElement = $document->getElement('p:blipFill/a:srcRect', $node);
        if ($oElement instanceof \DOMElement) {
            $cropTop = 0;
            $cropRight = 0;
            $cropBottom = 0;
            $cropLeft = 0;
            if ($oElement->hasAttribute('t')) {
                $cropTop = floatval($oElement->getAttribute('t'));
            }
            if ($oElement->hasAttribute('r')) {
                $cropRight = floatval($oElement->getAttribute('r'));
            }
            if ($oElement->hasAttribute('b')) {
                $cropBottom = floatval($oElement->getAttribute('b'));
            }
            if ($oElement->hasAttribute('l')) {
                $cropLeft = floatval($oElement->getAttribute('l'));
            }
            $oShape->setCrop($cropTop, $cropRight, $cropBottom, $cropLeft);
        }

        $oElementTransform = $document->getElement('p:spPr/a:xfrm', $node);
        if ($oElementTransform instanceof \DOMElement) {
            if ($oElementTransform->hasAttribute('rot')) {
                $oShape->setRotation(CommonDrawing::angleToDegrees($oElementTransform->getAttribute('rot')));
            }

            $oElementOffset = $document->getElement('a:off', $oElementTransform);
            if ($oElementOffset instanceof \DOMElement) {
                if ($oElementOffset->hasAttribute('x')) {
                    $oShape->setOffsetX(CommonDrawing::emuToPixels($oElementOffset->getAttribute('x')));
                }
                if ($oElementOffset->hasAttribute('y')) {
                    $oShape->setOffsetY(CommonDrawing::emuToPixels($oElementOffset->getAttribute('y')));
                }
            }

            $oElementExtent = $document->getElement('a:ext', $oElementTransform);
            if ($oElementExtent instanceof \DOMElement) {
                if ($oElementExtent->hasAttribute('cx')) {
                    $oShape->setWidth(CommonDrawing::emuToPixels($oElementExtent->getAttribute('cx')));
                }
                if ($oElementExtent->hasAttribute('cy')) {
                    $oShape->setHeight(CommonDrawing::emuToPixels($oElementExtent->getAttribute('cy')));
                }
            }

            if ($oElementTransform->hasAttribute('flipH')) {
                $oShape->setFlipH();
            }
            if ($oElementTransform->hasAttribute('flipV')) {
                $oShape->setFlipV();
            }
        }

        $oElement = $document->getElement('p:spPr/a:prstGeom', $node);
        if ($oElement instanceof \DOMElement && $oElement->hasAttribute('prst')) {
            $oShape->setShapeType($oElement->getAttribute('prst'));
        }

        $oElement = $document->getElement('p:spPr/a:effectLst', $node);
        if ($oElement instanceof \DOMElement) {
            $oShape->getShadow()->setVisible(true);

            $oSubElement = $document->getElement('a:outerShdw', $oElement);
            if ($oSubElement instanceof \DOMElement) {
                if ($oSubElement->hasAttribute('blurRad')) {
                    $oShape->getShadow()->setBlurRadius(CommonDrawing::emuToPixels($oSubElement->getAttribute('blurRad')));
                }
                if ($oSubElement->hasAttribute('dist')) {
                    $oShape->getShadow()->setDistance(CommonDrawing::emuToPixels($oSubElement->getAttribute('dist')));
                }
                if ($oSubElement->hasAttribute('dir')) {
                    $oShape->getShadow()->setDirection(CommonDrawing::angleToDegrees($oSubElement->getAttribute('dir')));
                }
                if ($oSubElement->hasAttribute('algn')) {
                    $oShape->getShadow()->setAlignment($oSubElement->getAttribute('algn'));
                }
            }

            $oSubElement = $document->getElement('a:outerShdw/a:srgbClr', $oElement);
            if ($oSubElement instanceof \DOMElement) {
                if ($oSubElement->hasAttribute('val')) {
                    $oColor = new Color();
                    $oColor->setRGB($oSubElement->getAttribute('val'));
                    $oShape->getShadow()->setColor($oColor);
                }
            }

            $oSubElement = $document->getElement('a:outerShdw/a:srgbClr/a:alpha', $oElement);
            if ($oSubElement instanceof \DOMElement) {
                if ($oSubElement->hasAttribute('val')) {
                    $oShape->getShadow()->setAlpha((int)$oSubElement->getAttribute('val') / 1000);
                }
            }
        }

        $oContainer->addShape($oShape);
    }

    /**
     * @param XMLReader $document
     * @param \DOMElement $node
     * @param $oContainer
     * @param AbstractSlide $oSlide
     * @throws \Exception
     */
    protected function loadShapeRichText(XMLReader $document, \DOMElement $node, $oContainer, AbstractSlide $oSlide)
    {
        // Core
        $oShape = $oContainer->createRichTextShape();
        $oShape->setParagraphs(array());
        $this->loadShapeProperties($document, $node, $oShape);

        //place holder
        $oElement = $document->getElement('p:nvSpPr/p:nvPr/p:ph', $node);
        if ($oElement instanceof \DOMElement) {
            $holderType = ($oElement->hasAttribute('type')) ? $oElement->getAttribute('type') : Placeholder::PH_TYPE_BODY;
            $holderIdx = ($oElement->hasAttribute('idx')) ? intval($oElement->getAttribute('idx')) : 0;

            $placeholder = new Placeholder($holderType);
            $placeholder->setIdx($holderIdx);
            $oShape->setPlaceHolder($placeholder);
        }

        // Hyperlink
        $oElementNvPr = $document->getElement('p:nvSpPr/p:cNvPr', $node);
        $this->registerShapeId($oElementNvPr, $oShape);
        if ($oElementNvPr instanceof \DOMElement) {
            $oHyperlink = $oShape->getHyperlink();
            $this->loadHyperlink($document, $oElementNvPr, $oHyperlink);
        }

        $oTxBodyProp = $document->getElement('p:txBody/a:bodyPr', $node);
        if ($oTxBodyProp instanceof \DOMElement) {
            if ($oTxBodyProp->hasAttribute('anchor')) {
                $oShape->getAlignment()->setVertical($oTxBodyProp->getAttribute('anchor'));
            }
            if ($oTxBodyProp->hasAttribute('tIns')) {
                $oShape->setInsetTop(CommonDrawing::emuToPixels($oTxBodyProp->getAttribute('tIns')));
            }
            if ($oTxBodyProp->hasAttribute('rIns')) {
                $oShape->setInsetRight(CommonDrawing::emuToPixels($oTxBodyProp->getAttribute('rIns')));
            }
            if ($oTxBodyProp->hasAttribute('bIns')) {
                $oShape->setInsetBottom(CommonDrawing::emuToPixels($oTxBodyProp->getAttribute('bIns')));
            }
            if ($oTxBodyProp->hasAttribute('lIns')) {
                $oShape->setInsetLeft(CommonDrawing::emuToPixels($oTxBodyProp->getAttribute('lIns')));
            }
        }

        $arrayElements = $document->getElements('p:txBody/a:p', $node);
        foreach ($arrayElements as $oElement) {
            $this->loadParagraph($document, $oElement, $oShape, $oSlide);
        }

        $arrayListStyles = $document->getElements('p:txBody/a:lstStyle/*', $node);
        if (0 === count($arrayListStyles)) {
            return;
        }
        foreach ($arrayListStyles as $oElmParaProp) {
            $lvl = $this->parseParagraphLevel($oElmParaProp);
            if (false === $lvl) {
                continue;
            }
            if ($oSlide instanceof SlideMaster && $oShape->isPlaceholder()) {
                if (Placeholder::PH_TYPE_BODY === $oShape->getPlaceholder()->getType()) {
                    $oParagraph = $oSlide->getTextStyles()->getBodyStyleAtLvl($lvl);
                } else if (Placeholder::PH_TYPE_TITLE === $oShape->getPlaceholder()->getType()) {
                    $oParagraph = $oSlide->getTextStyles()->getTitleStyleAtLvl($lvl);
                }
                if (is_null($oParagraph)) {
                    continue;
                }
                $this->loadParagraphProperty($document, $oElmParaProp, $oParagraph, $oSlide);
            } else {
                $oParagraph = new Paragraph();
                $this->loadParagraphProperty($document, $oElmParaProp, $oParagraph, $oSlide);
                $oShape->getListStyle()->setStyleAtLvl($lvl, $oParagraph);
            }
        }
    }

    protected function loadConnector(XMLReader $xmlReader, \DOMElement $node, $oContainer, $oSlide)
    {
        $oLine = $oContainer->createLineShape();
        $this->loadShapeProperties($xmlReader, $node, $oLine);

        $oTailEnd = $xmlReader->getElement('p:spPr/a:ln/a:tailEnd', $node);
        if ($oTailEnd instanceof \DOMElement && (!$oTailEnd->hasAttribute('type') || 'none' !== $oTailEnd->getAttribute('type'))) {
            $oLine->setTailEnd();
        }

        $oHeadEnd = $xmlReader->getElement('p:spPr/a:ln/a:headEnd', $node);
        if ($oHeadEnd instanceof \DOMElement && (!$oHeadEnd->hasAttribute('type') || 'none' !== $oHeadEnd->getAttribute('type'))) {
            $oLine->setHeadEnd();
        }

        // Hyperlink
        $oElementNvPr = $xmlReader->getElement('p:nvCxnSpPr/p:cNvPr', $node);
        $this->registerShapeId($oElementNvPr, $oLine);
        if ($oElementNvPr instanceof \DOMElement) {
            $oHyperlink = $oLine->getHyperlink();
            $this->loadHyperlink($xmlReader, $oElementNvPr, $oHyperlink);
        }
    }

    protected function loadGroup(XMLReader $xmlReader, \DOMElement $node, $oContainer, $oSlide)
    {
        $oGroup = $oContainer->createGroup();
        $oGroup->setSlide($oSlide);
        $this->loadShapeProperties($xmlReader, $node, $oGroup);

        $oElementNvPr = $xmlReader->getElement('p:nvGrpSpPr/p:cNvPr', $node);
        $this->registerShapeId($oElementNvPr, $oGroup);

        // Shapes
        $arrayElements = $xmlReader->getElements('(p:sp|p:grpSp|p:pic|p:cxnSp|p:graphicFrame)', $node);
        if ($arrayElements) {
            $this->loadSlideShapes($oGroup, $arrayElements, $xmlReader);
        }
    }

    protected function registerShapeId($oElementNvPr, $oShape)
    {
        $shapeId;
        if ($oElementNvPr instanceof \DOMElement && $oElementNvPr->hasAttribute('id')) {
            $shapeId = $oElementNvPr->getAttribute('id');
        }
        if (isset($shapeId)) {
            $this->shapeMap[$shapeId] = $oShape;
        }
    }

    protected function loadShapeProperties(XMLReader $xmlReader, \DOMElement $node, $oShape)
    {
        if ($oShape instanceof Group) {
            $prTag = 'p:grpSpPr';
        } else {
            $prTag = 'p:spPr';
        }

        //default border
        $oElement = $xmlReader->getElement('p:style/a:lnRef', $node);
        if ($oElement instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oElement);
            if (isset($oColor)) {
                $oShape->getBorder()->setColor($oColor);
            }
            $oShape->getBorder()->setLineStyle(Border::LINE_SINGLE);
            $oShape->getBorder()->setLineWidth(1);
        }

        //default fill
        $oElement = $xmlReader->getElement('p:style/a:fillRef', $node);
        if ($oElement instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oElement);
            if (isset($oColor)) {
                $oShape->getFill()->setStartColor($oColor);
                $oShape->getFill()->setFillType(Fill::FILL_SOLID);
            }
        }

        //text box
        $oElement = $xmlReader->getElement('p:nvSpPr/p:cNvSpPr', $node);
        if ($oElement instanceof \DOMElement && $oElement->hasAttribute('txBox') && '1' === $oElement->getAttribute('txBox')) {
            $oShape->setTextBox(true);
        }

        //shape type
        $oElement = $xmlReader->getElement('p:spPr/a:prstGeom', $node);
        if ($oElement instanceof \DOMElement && $oElement->hasAttribute('prst')) {
            $oShape->setShapeType($oElement->getAttribute('prst'));
        }

        //customized points
        $arrayElements = $xmlReader->getElements('p:spPr/a:prstGeom/a:avLst/a:gd', $node);
        if ($arrayElements) {
            $custPoints = new \stdClass();
            foreach ($arrayElements as $oElement) {
                $name = $oElement->getAttribute('name');
                $value = $oElement->getAttribute('fmla');
                if (false === strpos($value, 'val ')) {
                    continue;
                }
                $value = substr($value, 4, strlen($value) - 4);
                $custPoints->$name = $value;
            }
            $oShape->setCustPoints($custPoints);
        }

        //fill
        $oElementSpPr = $xmlReader->getElement($prTag, $node);
        if ($oElementSpPr instanceof \DOMElement) {
            $oFill = $this->loadStyleFill($xmlReader, $oElementSpPr);
            if (isset($oFill)) {
                $oShape->setFill($oFill);
            }
        }

        //border
        $oElementLn = $xmlReader->getElement('p:spPr/a:ln', $node);
        if ($oElementLn instanceof \DOMElement) {
            $this->loadStyleBorder($xmlReader, $oElementLn, $oShape->getBorder());
        }

        //transform
        $oElementTransform = $xmlReader->getElement($prTag . '/a:xfrm', $node);
        if ($oElementTransform instanceof \DOMElement) {
            $this->loadTransform($xmlReader, $oElementTransform, $oShape);
        }
    }

    private function loadTransform(XMLReader $xmlReader, $oElementTransform, $oShape)
    {
        //width, height
        $oElementExt = $xmlReader->getElement('a:ext', $oElementTransform);
        if ($oElementExt instanceof \DOMElement) {
            if ($oElementExt->hasAttribute('cx')) {
                $oShape->setWidth(CommonDrawing::emuToPixels($oElementExt->getAttribute('cx')));
            }
            if ($oElementExt->hasAttribute('cy')) {
                $oShape->setHeight(CommonDrawing::emuToPixels($oElementExt->getAttribute('cy')));
            }
        }

        if ($oElementTransform->hasAttribute('rot')) {
            $oShape->setRotation(CommonDrawing::angleToDegrees($oElementTransform->getAttribute('rot')));
        }
        if ($oElementTransform->hasAttribute('flipH')) {
            $oShape->setFlipH();
        }
        if ($oElementTransform->hasAttribute('flipV')) {
            $oShape->setFlipV();
        }

        //offset
        $oElementOff = $xmlReader->getElement('a:off', $oElementTransform);
        if ($oElementOff instanceof \DOMElement) {
            if ($oElementOff->hasAttribute('x')) {
                $x = CommonDrawing::emuToPixels($oElementOff->getAttribute('x'));
                if (0 > $oShape->getWidth()) {
                    $x -= $oShape->getWidth();
                }
                $oShape->setOffsetX($x);
            }
            if ($oElementOff->hasAttribute('y')) {
                $oShape->setOffsetY(CommonDrawing::emuToPixels($oElementOff->getAttribute('y')));
            }
        }

        //child offset
        $oElementChOff = $xmlReader->getElement('a:chOff', $oElementTransform);
        if ($oElementChOff instanceof \DOMElement) {
            if ($oElementChOff->hasAttribute('x')) {
                $oShape->setChildOffsetX(CommonDrawing::emuToPixels($oElementChOff->getAttribute('x')));
            }
            if ($oElementChOff->hasAttribute('y')) {
                $oShape->setChildOffsetY(CommonDrawing::emuToPixels($oElementChOff->getAttribute('y')));
            }
        }

        //child extent
        $oElementChExt = $xmlReader->getElement('a:chExt', $oElementTransform);
        if ($oElementChExt instanceof \DOMElement) {
            if ($oElementChExt->hasAttribute('cx')) {
                $oShape->setChildExtentX(CommonDrawing::emuToPixels($oElementChExt->getAttribute('cx')));
            }
            if ($oElementChExt->hasAttribute('cy')) {
                $oShape->setChildExtentY(CommonDrawing::emuToPixels($oElementChExt->getAttribute('cy')));
            }
        }
    }

    /**
     * @param XMLReader $document
     * @param \DOMElement $node
     * @param AbstractSlide $oSlide
     * @throws \Exception
     */
    protected function loadShapeTable(XMLReader $document, \DOMElement $node, $oContainer, AbstractSlide $oSlide)
    {
        $oShape = $oContainer->createTableShape();

        $oElement = $document->getElement('p:nvGraphicFramePr/p:cNvPr', $node);
        $this->registerShapeId($oElement, $oShape);
        if ($oElement instanceof \DOMElement) {
            if ($oElement->hasAttribute('name')) {
                $oShape->setName($oElement->getAttribute('name'));
            }
            if ($oElement->hasAttribute('descr')) {
                $oShape->setDescription($oElement->getAttribute('descr'));
            }
        }

        $oElementTransform = $document->getElement('p:xfrm', $node);
        if ($oElementTransform instanceof \DOMElement) {
            $this->loadTransform($document, $oElementTransform, $oShape);
        }

        $oElementTblPr = $document->getElement('a:graphic/a:graphicData/a:tbl/a:tblPr', $node);
        if ($oElementTblPr instanceof \DOMElement) {
            $attrList = array('firstRow', 'lastRow', 'bandRow', 'bandCol', 'firstCol', 'lastCol');
            foreach ($attrList as $key) {
                if ($oElementTblPr->hasAttribute($key)) {
                    $value = ('1' === $oElementTblPr->getAttribute($key) || 'on' === $oElementTblPr->getAttribute($key));
                    $oShape->setProperty($key, $value);
                }
            }
            $oElementStyleId = $document->getElement('a:tableStyleId', $oElementTblPr);
            $oShape->setStyleId($oElementStyleId->nodeValue);
        }

        $arrayElements = $document->getElements('a:graphic/a:graphicData/a:tbl/a:tblGrid/a:gridCol', $node);
        $oShape->setNumColumns($arrayElements->length);
        $oShape->createRow();
        foreach ($arrayElements as $key => $oElement) {
            if ($oElement instanceof \DOMElement && $oElement->getAttribute('w')) {
                $oShape->getRow(0)->getCell($key)->setWidth(CommonDrawing::emuToPixels($oElement->getAttribute('w')));
            }
        }

        $totalRowHeight = 0;
        $arrayElements = $document->getElements('a:graphic/a:graphicData/a:tbl/a:tr', $node);
        foreach ($arrayElements as $keyRow => $oElementRow) {
            if (!($oElementRow instanceof \DOMElement)) {
                continue;
            }
            $oRow = $oShape->getRow($keyRow, true);
            if (is_null($oRow)) {
                $oRow = $oShape->createRow();
            }
            if ($oElementRow->hasAttribute('h')) {
                $rowHeight = CommonDrawing::emuToPixels($oElementRow->getAttribute('h'));
                $totalRowHeight += $rowHeight;
                $oRow->setHeight($rowHeight);
            }
            $arrayElementsCell = $document->getElements('a:tc', $oElementRow);
            foreach ($arrayElementsCell as $keyCell => $oElementCell) {
                if (!($oElementCell instanceof \DOMElement)) {
                    continue;
                }
                $oCell = $oRow->getCell($keyCell);
                $oCell->setParagraphs(array());
                if ($oElementCell->hasAttribute('gridSpan')) {
                    $oCell->setColSpan($oElementCell->getAttribute('gridSpan'));
                }
                if ($oElementCell->hasAttribute('rowSpan')) {
                    $oCell->setRowSpan($oElementCell->getAttribute('rowSpan'));
                }

                foreach ($document->getElements('a:txBody/a:p', $oElementCell) as $oElementPara) {
                    $this->loadParagraph($document, $oElementPara, $oCell, $oSlide);
                }

                $oElementTcPr = $document->getElement('a:tcPr', $oElementCell);
                if ($oElementTcPr instanceof \DOMElement) {
                    $numParagraphs = count($oCell->getParagraphs());
                    if ($numParagraphs > 0) {
                        if ($oElementTcPr->hasAttribute('vert')) {
                            $oCell->getParagraph(0)->getAlignment()->setTextDirection($oElementTcPr->getAttribute('vert'));
                        }
                        if ($oElementTcPr->hasAttribute('anchor')) {
                            $oCell->getParagraph(0)->getAlignment()->setVertical($oElementTcPr->getAttribute('anchor'));
                        }
                        if ($oElementTcPr->hasAttribute('marB')) {
                            $oCell->getParagraph(0)->getAlignment()->setMarginBottom(CommonDrawing::emuToPixels($oElementTcPr->getAttribute('marB')));
                        }
                        if ($oElementTcPr->hasAttribute('marL')) {
                            //$oCell->getParagraph(0)->getAlignment()->setMarginLeft(CommonDrawing::emuToPixels($oElementTcPr->getAttribute('marL')));
                        }
                        if ($oElementTcPr->hasAttribute('marR')) {
                            $oCell->getParagraph(0)->getAlignment()->setMarginRight(CommonDrawing::emuToPixels($oElementTcPr->getAttribute('marR')));
                        }
                        if ($oElementTcPr->hasAttribute('marT')) {
                            $oCell->getParagraph(0)->getAlignment()->setMarginTop(CommonDrawing::emuToPixels($oElementTcPr->getAttribute('marT')));
                        }
                    }

                    $oFill = $this->loadStyleFill($document, $oElementTcPr);
                    if ($oFill instanceof Fill) {
                        $oCell->setFill($oFill);
                    }

                    $oBorders = new Borders();
                    $oElementBorderL = $document->getElement('a:lnL', $oElementTcPr);
                    if ($oElementBorderL instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderL, $oBorders->getLeft());
                    }
                    $oElementBorderR = $document->getElement('a:lnR', $oElementTcPr);
                    if ($oElementBorderR instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderR, $oBorders->getRight());
                    }
                    $oElementBorderT = $document->getElement('a:lnT', $oElementTcPr);
                    if ($oElementBorderT instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderT, $oBorders->getTop());
                    }
                    $oElementBorderB = $document->getElement('a:lnB', $oElementTcPr);
                    if ($oElementBorderB instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderB, $oBorders->getBottom());
                    }
                    $oElementBorderDiagDown = $document->getElement('a:lnTlToBr', $oElementTcPr);
                    if ($oElementBorderDiagDown instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderDiagDown, $oBorders->getDiagonalDown());
                    }
                    $oElementBorderDiagUp = $document->getElement('a:lnBlToTr', $oElementTcPr);
                    if ($oElementBorderDiagUp instanceof \DOMElement) {
                        $this->loadStyleBorder($document, $oElementBorderDiagUp, $oBorders->getDiagonalUp());
                    }
                    $oCell->setBorders($oBorders);
                }
            }
            if ($oShape->getHeight() > $totalRowHeight) {
                $oShape->setHeight($totalRowHeight);
            }
        }
    }

    protected function loadShapeChart(XMLReader $xmlReader, \DOMElement $node, $oContainer, AbstractSlide $oSlide)
    {
        $oElementChart = $xmlReader->getElement('a:graphic/a:graphicData/c:chart', $node);
        if (!($oElementChart instanceof \DOMElement) || !$oElementChart->hasAttribute('r:id')) {
            return;
        }
        $chartFileName = basename($this->arrayRels[$oSlide->getRelsIndex()][$oElementChart->getAttribute('r:id')]['Target']);
        $sChart = $this->oZip->getFromName("ppt/charts/$chartFileName");
        $chartReader = new XMLReader();
        if (!$chartReader->getDomFromString($sChart)) {
            return;
        }
        $oChart = $oContainer->createChartShape();

        $oElementNvPr = $xmlReader->getElement('p:nvGraphicFramePr/p:cNvPr', $node);
        $this->registerShapeId($oElementNvPr, $oChart);

        $oElementTransform = $xmlReader->getElement('p:xfrm', $node);
        if ($oElementTransform instanceof \DOMElement) {
            $this->loadTransform($xmlReader, $oElementTransform, $oChart);
        }

        $arrayTitlePara = $chartReader->getElements('/c:chartSpace/c:chart/c:title/c:tx/c:rich/a:p');
        $title = '';
        foreach ($arrayTitlePara as $oElementPara) {
            $arrayTextRun = $chartReader->getElements('a:r', $oElementPara);
            foreach ($arrayTextRun as $oElementTextRun) {
                $oElementText = $chartReader->getElement('a:t', $oElementTextRun);
                if ($oElementText instanceof \DOMElement) {
                    $title = $title . $oElementText->nodeValue;
                }
            }
        }
        $oChart->getTitle()->setText($title);

        $arrayCharts = $chartReader->getElements('/c:chartSpace/c:chart/c:plotArea/*');
        foreach ($arrayCharts as $oElementChart) {
            if ('c:barChart' === $oElementChart->nodeName) {
                $oType = new Bar();
            } else if ('c:lineChart' === $oElementChart->nodeName) {
                $oType = new Line();
            } else if ('c:areaChart' === $oElementChart->nodeName) {
                $oType = new Area();
            } else if ('c:pieChart' === $oElementChart->nodeName) {
                $oType = new Pie();
            } else if ('c:scatterChart' === $oElementChart->nodeName) {
                $oType = new Scatter();
            } else if ('c:catAx' === $oElementChart->nodeName || 'c:valAx' === $oElementChart->nodeName) {
                $arrayAxTitlePara = $chartReader->getElements('c:title/c:tx/c:rich/a:p', $oElementChart);
                $axTitle = '';
                foreach ($arrayAxTitlePara as $oElementPara) {
                    $oElementText = $chartReader->getElement('a:r/a:t', $oElementPara);
                    if ($oElementText instanceof \DOMElement) {
                        $axTitle = $axTitle . $oElementText->nodeValue;
                    }
                }
                if ('c:catAx' === $oElementChart->nodeName) {
                    $oChart->getPlotArea()->getAxisX()->setTitle($axTitle);
                } else {
                    $oChart->getPlotArea()->getAxisY()->setTitle($axTitle);
                }
                continue;
            } else {
                continue;
            }

            $this->loadChart($chartReader, $oElementChart, $oType);
            $oChart->getPlotArea()->addType($oType);
        }
    }

    private function loadChart(XMLReader $xmlReader, \DOMElement $oElementChart, $oType)
    {
        $oElementBarDir = $xmlReader->getElement('c:barDir', $oElementChart);
        if ($oElementBarDir instanceof \DOMElement && $oElementBarDir->hasAttribute('val')) {
            $oType->setDirection($oElementBarDir->getAttribute('val'));
        }

        $oElementGrouping = $xmlReader->getElement('c:grouping', $oElementChart);
        if ($oElementGrouping instanceof \DOMElement && $oElementGrouping->hasAttribute('val')) {
            $oType->setGrouping($oElementGrouping->getAttribute('val'));
        }

        $oElementMarker = $xmlReader->getElement('c:marker', $oElementChart);
        if ($oElementMarker instanceof \DOMElement && $oElementMarker->hasAttribute('val') && '1' == $oElementMarker->getAttribute('val')) {
            $oType->setMarker();
        }

        $arraySeries = $xmlReader->getElements('c:ser', $oElementChart);
        foreach ($arraySeries as $oElementSer) {
            if ($oType instanceof Scatter) {
                $oType->addSeries($this->loadScatterChartSeries($xmlReader, $oElementSer));
            } else {
                $oType->addSeries($this->loadCommonChartSeries($xmlReader, $oElementSer));
            }
        }
    }

    private function loadCommonChartSeries(XMLReader $xmlReader, \DOMElement $oElementSer)
    {
        $title = '';
        $values = array();

        if ($xmlReader->elementExists('c:tx/c:v', $oElementSer)) {
            $oElementSerTitle = $xmlReader->getElement('c:tx/c:v', $oElementSer);
        } else {
            $oElementSerTitle = $xmlReader->getElement('c:tx/c:strRef/c:strCache/c:pt/c:v', $oElementSer);
        }
        if ($oElementSerTitle instanceof \DOMElement) {
            $title = $oElementSerTitle->nodeValue;
        }

        if ($xmlReader->elementExists('c:cat/c:numRef', $oElementSer)) {
            $arrayKeys = $xmlReader->getElements('c:cat/c:numRef/c:numCache/c:pt/c:v', $oElementSer);
        } else if ($xmlReader->elementExists('c:cat/c:numLit', $oElementSer)) {
            $arrayKeys = $xmlReader->getElements('c:cat/c:numLit/c:pt/c:v', $oElementSer);
        } else if ($xmlReader->elementExists('c:cat/c:strRef', $oElementSer)) {
            $arrayKeys = $xmlReader->getElements('c:cat/c:strRef/c:strCache/c:pt/c:v', $oElementSer);
        } else {
            $arrayKeys = $xmlReader->getElements('c:cat/c:strLit/c:pt/c:v', $oElementSer);
        }
        if ($xmlReader->elementExists('c:val/c:numRef', $oElementSer)) {
            $arrayValues = $xmlReader->getElements('c:val/c:numRef/c:numCache/c:pt/c:v', $oElementSer);
        } else {
            $arrayValues = $xmlReader->getElements('c:val/c:numLit/c:pt/c:v', $oElementSer);
        }
        foreach ($arrayKeys as $idx => $oElementKey) {
            $values[$oElementKey->nodeValue] = floatval($arrayValues[$idx]->nodeValue);
        }

        return new Series($title, $values);
    }

    private function loadScatterChartSeries(XMLReader $xmlReader, \DOMElement $oElementSer)
    {
        $title = '';
        $values = array();

        if ($xmlReader->elementExists('c:tx/c:v', $oElementSer)) {
            $oElementSerTitle = $xmlReader->getElement('c:tx/c:v', $oElementSer);
        } else {
            $oElementSerTitle = $xmlReader->getElement('c:tx/c:strRef/c:strCache/c:pt/c:v', $oElementSer);
        }
        if ($oElementSerTitle instanceof \DOMElement) {
            $title = $oElementSerTitle->nodeValue;
        }

        if ($xmlReader->elementExists('c:xVal/c:numRef', $oElementSer)) {
            $arrayKeys = $xmlReader->getElements('c:xVal/c:numRef/c:numCache/c:pt/c:v', $oElementSer);
        } else {
            $arrayKeys = $xmlReader->getElements('c:xVal/c:numLit/c:pt/c:v', $oElementSer);
        }
        if ($xmlReader->elementExists('c:yVal/c:numRef', $oElementSer)) {
            $arrayValues = $xmlReader->getElements('c:yVal/c:numRef/c:numCache/c:pt/c:v', $oElementSer);
        } else {
            $arrayValues = $xmlReader->getElements('c:yVal/c:numLit/c:pt/c:v', $oElementSer);
        }
        foreach ($arrayKeys as $idx => $oElementKey) {
            $values[$oElementKey->nodeValue] = floatval($arrayValues[$idx]->nodeValue);
        }

        return new Series($title, $values);
    }

    /**
     * @param XMLReader $document
     * @param \DOMElement $oElement
     * @param Cell|RichText $oShape
     * @throws \Exception
     */
    protected function loadParagraph(XMLReader $document, \DOMElement $oElement, $oShape, AbstractSlide $oSlide)
    {
        // Core
        $oParagraph = $oShape->createParagraph();
        $oParagraph->setRichTextElements(array());

        $oSubElement = $document->getElement('a:pPr', $oElement);
        if ($oSubElement instanceof \DOMElement) {
            $this->loadParagraphProperty($document, $oSubElement, $oParagraph, $oSlide);
        }
        $arraySubElements = $document->getElements('(a:r|a:br|a:fld)', $oElement);
        foreach ($arraySubElements as $oSubElement) {
            if ($oSubElement->tagName === 'a:br') {
                $oParagraph->createBreak();
                continue;
            }
            if ($oSubElement->tagName === 'a:fld') {
                $oText = $oParagraph->createTextRun();
                $oTextElement = $document->getElement('a:t', $oSubElement);
                $oText->setText($oTextElement->nodeValue);
                continue;
            }

            $oElementRunPr = $document->getElement('a:rPr', $oSubElement);
            $oText = $oParagraph->createTextRun();
            $oFont = $oText->getFont();
            if ($oElementRunPr instanceof \DOMElement) {
                $this->loadTextRunProperty($document, $oElementRunPr, $oFont, $oSlide);
            }

            // Hyperlink
            $oHyperlink = $oText->getHyperlink();
            if ($oElementRunPr instanceof \DOMElement) {
                $this->loadHyperlink($document, $oElementRunPr, $oHyperlink);
            }

            $oTextElement = $document->getElement('a:t', $oSubElement);
            $oText->setText($oTextElement->nodeValue);
        }

        $oEndParaRPr = $document->getElement('a:endParaRPr', $oElement);
        if ($oEndParaRPr instanceof \DOMElement) {
            $this->loadTextRunProperty($document, $oEndParaRPr, $oParagraph->getFont(), $oSlide);
        }
    }

    /**
     * @param XMLReader $xmlReader
     * @param \DOMElement $oElement
     * @param Border $oBorder
     */
    protected function loadStyleBorder(XMLReader $xmlReader, \DOMElement $oElement, Border $oBorder)
    {
        $oElementNoFill = $xmlReader->getElement('a:noFill', $oElement);
        if ($oElementNoFill instanceof \DOMElement) {
            $oBorder->setLineStyle(Border::LINE_NONE);
            return;
        }
        $oBorder->setLineStyle(Border::LINE_SINGLE);

        if ($oElement->hasAttribute('w')) {
            $oBorder->setLineWidth($oElement->getAttribute('w') / 12700);
        }
        if ($oElement->hasAttribute('cmpd')) {
            $oBorder->setLineStyle($oElement->getAttribute('cmpd'));
        }

        $oSolidFill = $xmlReader->getElement('a:solidFill', $oElement);
        if ($oSolidFill instanceof \DOMElement) {
            $oBorder->setColor($this->loadStyleColor($xmlReader, $oSolidFill));
        }

        $oElementDashStyle = $xmlReader->getElement('a:prstDash', $oElement);
        if ($oElementDashStyle instanceof \DOMElement && $oElementDashStyle->hasAttribute('val')) {
            $oBorder->setDashStyle($oElementDashStyle->getAttribute('val'));
        }
    }

    /**
     * @param XMLReader $xmlReader
     * @param \DOMElement $oElement
     * @return Color
     */
    protected function loadStyleColor(XMLReader $xmlReader, \DOMElement $oElement)
    {
        $oElementSrgbColor = $xmlReader->getElement('a:srgbClr', $oElement);
        if ($oElementSrgbColor instanceof \DOMElement) {
            $oSrgbColor = new Color();

            $oElementAlpha = $xmlReader->getElement('a:alpha', $oElementSrgbColor);
            if ($oElementAlpha instanceof \DOMElement && $oElementAlpha->hasAttribute('val')) {
                $alpha = strtoupper(dechex((($oElementAlpha->getAttribute('val') / 1000) / 100) * 255));
                $oSrgbColor->setRGB($oElementSrgbColor->getAttribute('val'), $alpha);
            } else {
                $oSrgbColor->setRGB($oElementSrgbColor->getAttribute('val'));
            }

            return $oSrgbColor;
        }

        $oElementSchemeColor = $xmlReader->getElement('a:schemeClr', $oElement);
        if ($oElementSchemeColor instanceof \DOMElement) {
            $oSchemeColor = new SchemeColor();
            $oSchemeColor->setValue($oElementSchemeColor->getAttribute('val'));

            $oLumMod = $xmlReader->getElement('a:lumMod', $oElementSchemeColor);
            if ($oLumMod instanceof \DOMElement && $oLumMod->hasAttribute('val')) {
                $oSchemeColor->setLumMod(intval($oLumMod->getAttribute('val')) / 1000);
                $oSchemeColor->setTransform(SchemeColor::TRANSFORM_LUM);
            }
            $oLumOff = $xmlReader->getElement('a:lumOff', $oElementSchemeColor);
            if ($oLumOff instanceof \DOMElement && $oLumOff->hasAttribute('val')) {
                $oSchemeColor->setLumOff(intval($oLumOff->getAttribute('val')) / 1000);
            }

            $oTint = $xmlReader->getElement('a:tint', $oElementSchemeColor);
            if ($oTint instanceof \DOMElement && $oTint->hasAttribute('val')) {
                $oSchemeColor->setTint(intval($oTint->getAttribute('val')) / 1000);
                $oSchemeColor->setTransform(SchemeColor::TRANSFORM_TINT);
            }

            $oElementAlpha = $xmlReader->getElement('a:alpha', $oElementSchemeColor);
            if ($oElementAlpha instanceof \DOMElement && $oElementAlpha->hasAttribute('val')) {
                $oSchemeColor->setAlpha(intval($oElementAlpha->getAttribute('val') / 1000));
            }
            return $oSchemeColor;
        }
    }

    /**
     * @param XMLReader $xmlReader
     * @param \DOMElement $oElement
     * @return null|Fill
     */
    protected function loadStyleFill(XMLReader $xmlReader, \DOMElement $oElement)
    {
        // Gradient fill
        $oElementFill = $xmlReader->getElement('a:gradFill', $oElement);
        if ($oElementFill instanceof \DOMElement) {
            $oFill = new Fill();
            $oFill->setFillType(Fill::FILL_GRADIENT_LINEAR);

            $oGradientStop = $xmlReader->getElement('a:gsLst/a:gs[@pos="0"]', $oElementFill);
            if ($oGradientStop instanceof \DOMElement) {
                $oFill->setStartColor($this->loadStyleColor($xmlReader, $oGradientStop));
            }

            $oGradientStop = $xmlReader->getElement('a:gsLst/a:gs[@pos="100000"]', $oElementFill);
            if ($oGradientStop instanceof \DOMElement) {
                $oFill->setEndColor($this->loadStyleColor($xmlReader, $oGradientStop));
            }

            $oRotation = $xmlReader->getElement('a:lin', $oElementFill);
            if ($oRotation instanceof \DOMElement && $oRotation->hasAttribute('ang')) {
                $oFill->setRotation(CommonDrawing::angleToDegrees($oRotation->getAttribute('ang')));
            }
            return $oFill;
        }

        // Solid fill
        $oElementFill = $xmlReader->getElement('a:solidFill', $oElement);
        if ($oElementFill instanceof \DOMElement) {
            $oFill = new Fill();
            $oFill->setFillType(Fill::FILL_SOLID);

            $oFill->setStartColor($this->loadStyleColor($xmlReader, $oElementFill));
            return $oFill;
        }

        $oElementFill = $xmlReader->getElement('a:blipFill/a:blip', $oElement);
        if ($oElementFill instanceof \DOMElement) {
            $imageSource = $this->loadBlipFill($oElementFill);
            if (!empty($imageSource)) {
                $oFill = new Fill();
                $oFill->setFillType(Fill::FILL_BLIP);
                $oFill->setImageSource($imageSource);
                return $oFill;
            }
        }

        // No fill
        $oElementFill = $xmlReader->getElement('a:noFill', $oElement);
        if ($oElementFill instanceof \DOMElement) {
            return new Fill();
        }
        return null;
    }

    /**
     * @param string $fileRels
     * @return string
     */
    protected function loadRels($fileRels)
    {
        $sPart = $this->oZip->getFromName($fileRels);
        if ($sPart !== false) {
            $xmlReader = new XMLReader();
            if ($xmlReader->getDomFromString($sPart)) {
                foreach ($xmlReader->getElements('*') as $oNode) {
                    if (!($oNode instanceof \DOMElement)) {
                        continue;
                    }
                    $this->arrayRels[$fileRels][$oNode->getAttribute('Id')] = array(
                        'Target' => $oNode->getAttribute('Target'),
                        'Type' => $oNode->getAttribute('Type'),
                    );
                }
            }
        }
    }

    /**
     * @param $oContainer
     * @param \DOMNodeList $oElements
     * @param XMLReader $xmlReader
     * @internal param $baseFile
     */
    protected function loadSlideShapes($oContainer, $oElements, $xmlReader)
    {
        if ($oContainer instanceof Group) {
            $oSlide = $oContainer->getSlide();
        } else if ($oContainer instanceof Note) {
            $oSlide = $oContainer->getParent();
        } else {
            $oSlide = $oContainer;
        }

        foreach ($oElements as $oNode) {
            $oPrstGeom = $xmlReader->getElement('p:spPr/a:prstGeom', $oNode);
            $shapeType = '';
            if ($oPrstGeom && $oPrstGeom->hasAttribute('prst')) {
                $shapeType = $oPrstGeom->getAttribute('prst');
            }
            if ('p:graphicFrame' === $oNode->tagName) {
                $xmlReader->registerNs('c', 'http://schemas.openxmlformats.org/drawingml/2006/chart');
                if ($xmlReader->getElement('a:graphic/a:graphicData/a:tbl', $oNode) instanceof \DOMElement) {
                    $this->loadShapeTable($xmlReader, $oNode, $oContainer, $oSlide);
                } else if ($xmlReader->getElement('a:graphic/a:graphicData/c:chart', $oNode) instanceof \DOMElement) {
                    $this->loadShapeChart($xmlReader, $oNode, $oContainer, $oSlide);
                }
            } else if ('p:pic' === $oNode->tagName) {
                $this->loadShapeDrawing($xmlReader, $oNode, $oContainer, $oSlide);
            } else if ('p:sp' === $oNode->tagName && 'line' !== $shapeType) {
                $this->loadShapeRichText($xmlReader, $oNode, $oContainer, $oSlide);
            } else if ('p:cxnSp' === $oNode->tagName || 'line' === $shapeType) {
                $this->loadConnector($xmlReader, $oNode, $oContainer, $oSlide);
            } else if ('p:grpSp' === $oNode->tagName) {
                $this->loadGroup($xmlReader, $oNode, $oContainer, $oSlide);
            }
        }
    }

    private function loadTypeface(XMLReader $xmlReader, \DOMElement $oParentElement, $oSlide, $oFont)
    {
        $arrayElementName = array('a:latin', 'a:ea', 'a:cs', 'a:sym');

        if ($oSlide instanceof Slide) {
            $oMaster = $oSlide->getSlideLayout()->getSlideMaster();
        } else if ($oSlide instanceof SlideLayout) {
            $oMaster = $oSlide->getSlideMaster();
        } else {
            $oMaster = $oSlide;
        }

        foreach ($arrayElementName as $name) {
            $oElement = $xmlReader->getElement($name, $oParentElement);
            if ($oElement instanceof \DOMElement && $oElement->hasAttribute('typeface')) {
                $typeface = $oElement->getAttribute('typeface');
                if (isset($oMaster)) {
                    if (false !== strpos($typeface, '+mj-')) {
                        $typeface = $oMaster->getSchemeFont(true, $name);
                    } else if (false !== strpos($typeface, '+mn-')) {
                        $typeface = $oMaster->getSchemeFont(false, $name);
                    }
                }
                $oFont->setLangFont($name, $typeface);
            }
        }
    }

    private function parseParagraphLevel(\DOMElement $oElmParaProp)
    {
        if ($oElmParaProp->hasAttribute('lvl')) {
            $lvl = intval($oElmParaProp->getAttribute('lvl')) + 1;
        } else {
            $matches = array();
            if (0 === preg_match('/lvl(\d+)pPr/', $oElmParaProp->nodeName, $matches) || 2 !== count($matches)) {
                return false;
            }
            $lvl = intval($matches[1]);
        }

        return $lvl;
    }

    private function loadParagraphProperty(XMLReader $xmlReader, \DOMElement $oElmParaProp, Paragraph $oParagraph, AbstractSlide $oSlide)
    {
        if ($oElmParaProp->hasAttribute('algn')) {
            $oParagraph->getAlignment()->setHorizontal($oElmParaProp->getAttribute('algn'));
        }
        if ($oElmParaProp->hasAttribute('fontAlgn')) {
            $oParagraph->getAlignment()->setVertical($oElmParaProp->getAttribute('fontAlgn'));
        }
        if ($oElmParaProp->hasAttribute('marL')) {
            $oParagraph->getAlignment()->setMarginLeft(CommonDrawing::emuToPixels($oElmParaProp->getAttribute('marL')));
        }
        if ($oElmParaProp->hasAttribute('marR')) {
            $oParagraph->getAlignment()->setMarginRight(CommonDrawing::emuToPixels($oElmParaProp->getAttribute('marR')));
        }
        if ($oElmParaProp->hasAttribute('indent')) {
            $oParagraph->getAlignment()->setIndent(CommonDrawing::emuToPixels($oElmParaProp->getAttribute('indent')));
        }
        if ($oElmParaProp->hasAttribute('lvl')) {
            $oParagraph->getAlignment()->setLevel($oElmParaProp->getAttribute('lvl'));
        }

        $oLnSpc = $xmlReader->getElement('a:lnSpc/a:spcPct', $oElmParaProp);
        if ($oLnSpc instanceof \DOMElement && $oLnSpc->hasAttribute('val')) {
            $oParagraph->setLineSpacing(floatval($oLnSpc->getAttribute('val')));
        }
        $oSpcBef = $xmlReader->getElement('a:spcBef/a:spcPts', $oElmParaProp);
        if ($oSpcBef instanceof \DOMElement && $oSpcBef->hasAttribute('val')) {
            $oParagraph->setSpaceBefore(floatval($oSpcBef->getAttribute('val')));
        }
        $oSpcAft = $xmlReader->getElement('a:spcAft/a:spcPts', $oElmParaProp);
        if ($oSpcAft instanceof \DOMElement && $oSpcAft->hasAttribute('val')) {
            $oParagraph->setSpaceAfter(floatval($oSpcAft->getAttribute('val')));
        }

        $oBuNone = $xmlReader->getElement('a:buNone', $oElmParaProp);
        if ($oBuNone instanceof \DOMElement) {
            $oParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NONE);
        }

        $oElementBuFont = $xmlReader->getElement('a:buFont', $oElmParaProp);
        if ($oElementBuFont instanceof \DOMElement) {
            if ($oElementBuFont->hasAttribute('typeface')) {
                $oParagraph->getBulletStyle()->setBulletFont($oElementBuFont->getAttribute('typeface'));
            }
        }
        $oElementBuChar = $xmlReader->getElement('a:buChar', $oElmParaProp);
        if ($oElementBuChar instanceof \DOMElement) {
            $oParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_BULLET);
            if ($oElementBuChar->hasAttribute('char')) {
                $oParagraph->getBulletStyle()->setBulletChar($oElementBuChar->getAttribute('char'));
            }
        }
        $oElementBuAutoNum = $xmlReader->getElement('a:buAutoNum', $oElmParaProp);
        if ($oElementBuAutoNum instanceof \DOMElement) {
            $oParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NUMERIC);
            if ($oElementBuAutoNum->hasAttribute('type')) {
                $oParagraph->getBulletStyle()->setBulletNumericStyle($oElementBuAutoNum->getAttribute('type'));
            }
            if ($oElementBuAutoNum->hasAttribute('startAt') && $oElementBuAutoNum->getAttribute('startAt') != 1) {
                $oParagraph->getBulletStyle()->setBulletNumericStartAt($oElementBuAutoNum->getAttribute('startAt'));
            }
        }
        $oElementBuClr = $xmlReader->getElement('a:buClr', $oElmParaProp);
        if ($oElementBuClr instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oElementBuClr);
            if (isset($oColor)) {
                $oParagraph->getBulletStyle()->setBulletColor($oColor);
            }
        }

        $oDefRPr = $xmlReader->getElement('a:defRPr', $oElmParaProp);
        if ($oDefRPr instanceof \DOMElement) {
            $this->loadTextRunProperty($xmlReader, $oDefRPr, $oParagraph->getFont(), $oSlide);
        }
    }

    private function loadTextRunProperty(XMLReader $xmlReader, \DOMElement $oElementRunPr, Font $oFont, $oSlide)
    {
        if ($oElementRunPr->hasAttribute('sz')) {
            $oFont->setSize($oElementRunPr->getAttribute('sz') / 100);
        }
        if ($oElementRunPr->hasAttribute('b')) {
            $oFont->setBold($oElementRunPr->getAttribute('b') == 1 ? Font::STATE_ON : Font::STATE_OFF);
        }
        if ($oElementRunPr->hasAttribute('i')) {
            $oFont->setItalic($oElementRunPr->getAttribute('i') == 1 ? Font::STATE_ON : Font::STATE_OFF);
        }
        if ($oElementRunPr->hasAttribute('u')) {
            $oFont->setUnderline($oElementRunPr->getAttribute('u'));
        }
        if ($oElementRunPr->hasAttribute('strike')) {
            $oFont->setStrikethrough($oElementRunPr->getAttribute('strike') == 'noStrike' ? Font::STATE_OFF : Font::STATE_ON);
        }
        if ($oElementRunPr->hasAttribute('baseline')) {
            $oFont->setBaseline(floatval($oElementRunPr->getAttribute('baseline')));
        }
        $oSolidFill = $xmlReader->getElement('a:solidFill', $oElementRunPr);
        if ($oSolidFill instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oSolidFill);
            if (isset($oColor)) {
                $oFont->setColor($oColor);
            }
        }
        $oHighlight = $xmlReader->getElement('a:highlight', $oElementRunPr);
        if ($oHighlight instanceof \DOMElement) {
            $oColor = $this->loadStyleColor($xmlReader, $oHighlight);
            if (isset($oColor)) {
                $oFont->setHighlight($oColor);
            }
        }
        $this->loadTypeface($xmlReader, $oElementRunPr, $oSlide, $oFont);
    }

    private function loadHyperlink(XMLReader $xmlReader, \DOMElement $oElementrPr, $oHyperlink)
    {
        $oElementHlinkClick = $xmlReader->getElement('a:hlinkClick', $oElementrPr);
        if (is_object($oElementHlinkClick)) {
            if ($oElementHlinkClick->hasAttribute('tooltip')) {
                $oHyperlink->setTooltip($oElementHlinkClick->getAttribute('tooltip'));
            }
            if ($oElementHlinkClick->hasAttribute('action')) {
                $oHyperlink->setAction($oElementHlinkClick->getAttribute('action'));
            }
            if ($oElementHlinkClick->hasAttribute('r:id') && isset($this->arrayRels[$this->fileRels][$oElementHlinkClick->getAttribute('r:id')]['Target'])) {
                $oHyperlink->setUrl($this->arrayRels[$this->fileRels][$oElementHlinkClick->getAttribute('r:id')]['Target']);
            }
        }
    }

    private function loadTableStyles($pptTableStyles)
    {
        $xmlReader = new XMLReader();
        if (!$xmlReader->getDomFromString($pptTableStyles)) {
            return;
        }
        $arrayTblStyles = $xmlReader->getElements('/a:tblStyleLst/*');
        foreach ($arrayTblStyles as $oElementTblStyle) {
            if (!$oElementTblStyle->hasAttribute('styleId')) {
                continue;
            }
            $tableStyles = new TableStyles();
            $styleId = $oElementTblStyle->getAttribute('styleId');

            $arrayStyles = $xmlReader->getElements('*', $oElementTblStyle);
            foreach ($arrayStyles as $oElementStyle) {
                if ('a:' !== substr($oElementStyle->nodeName, 0, 2)) {
                    continue;
                }
                $styleName = substr($oElementStyle->nodeName, 2);
                $tableStyle = new TableStyle();

                // tc text
                $oElementTctx = $xmlReader->getElement('a:tcTxStyle', $oElementStyle);
                if ($oElementTctx instanceof \DOMElement) {
                    if ($oElementTctx->hasAttribute('b')) {
                        $val = $oElementTctx->getAttribute('b');
                        if ('on' === $val || '1' == $val) {
                            $tableStyle->font->setBold(Font::STATE_ON);
                        }
                    }
                    if ($oElementTctx->hasAttribute('i')) {
                        $val = $oElementTctx->getAttribute('i');
                        if ('on' === $val || '1' == $val) {
                            $tableStyle->font->setItalic(Font::STATE_ON);
                        }
                    }
                    $oColor = $this->loadStyleColor($xmlReader, $oElementTctx);
                    if (isset($oColor)) {
                        $tableStyle->font->setColor($oColor);
                    }
                    $oElementFont = $xmlReader->getElement('a:fontRef', $oElementTctx);
                    if ($oElementFont instanceof \DOMElement) {
                        if ($oElementFont->hasAttribute('idx')) {
                            $tableStyle->font->setIdx($oElementFont->getAttribute('idx'));
                        }
                    }
                }

                // tc style
                $oElementBorder = $xmlReader->getElement('a:tcStyle/a:tcBdr/a:left/a:ln', $oElementStyle);
                if ($oElementBorder instanceof \DOMElement) {
                    $this->loadStyleBorder($xmlReader, $oElementBorder, $tableStyle->borders->getLeft());
                }
                $oElementBorder = $xmlReader->getElement('a:tcStyle/a:tcBdr/a:right/a:ln', $oElementStyle);
                if ($oElementBorder instanceof \DOMElement) {
                    $this->loadStyleBorder($xmlReader, $oElementBorder, $tableStyle->borders->getRight());
                }
                $oElementBorder = $xmlReader->getElement('a:tcStyle/a:tcBdr/a:top/a:ln', $oElementStyle);
                if ($oElementBorder instanceof \DOMElement) {
                    $this->loadStyleBorder($xmlReader, $oElementBorder, $tableStyle->borders->getTop());
                }
                $oElementBorder = $xmlReader->getElement('a:tcStyle/a:tcBdr/a:bottom/a:ln', $oElementStyle);
                if ($oElementBorder instanceof \DOMElement) {
                    $this->loadStyleBorder($xmlReader, $oElementBorder, $tableStyle->borders->getBottom());
                }
                //TODO: insideH, insideV

                /*$oElementFill = $xmlReader->getElement('a:tcStyle/a:fill/a:solidFill', $oElementStyle);
                if ($oElementFill instanceof \DOMElement) {
                    $oColor = $this->loadStyleColor($xmlReader, $oElementFill);
                    if (isset($oColor)) {
                        $tableStyle->fillColor = $oColor;
                    }
                }*/
                /*$oElementTcStyle = $xmlReader->getElement('a:tcStyle', $oElementStyle);
                if ($oElementTcStyle instanceof \DOMElement) {
                    $oFill = $this->loadStyleFill($xmlReader, $oElementTcStyle);
                    $tableStyle->fill = $oFill;
                }*/
                $oElementFill = $xmlReader->getElement('a:fillRef', $oElementStyle);
                if ($oElementFill instanceof \DOMElement) {
                    $tableStyle->fillColor = $this->loadStyleColor($xmlReader, $oElementFill);
                }
                $oElementFill = $xmlReader->getElement('a:tcStyle/a:fill/a:solidFill', $oElementStyle);
                if ($oElementFill instanceof \DOMElement) {
                    $tableStyle->fillColor = $this->loadStyleColor($xmlReader, $oElementFill);
                }
                $oElementFill = $xmlReader->getElement('a:tcStyle/a:fillRef', $oElementStyle);
                if ($oElementFill instanceof \DOMElement) {
                    $tableStyle->fillColor = $this->loadStyleColor($xmlReader, $oElementFill);
                }
                $oElementFill = $xmlReader->getElement('a:tcStyle/a:fill/a:noFill', $oElementStyle);
                if ($oElementFill instanceof \DOMElement) {
                    $tableStyle->fillColor = $this->loadStyleColor($xmlReader, $oElementFill);
                }

                $tableStyles->$styleName = $tableStyle;
            }
            $this->oPhpPresentation->setTableStyle($styleId, $tableStyles);
        }
    }

    private function loadCommentAuthors($sAuthors)
    {
        $xmlReader = new XMLReader();
        if ($xmlReader->getDomFromString($sAuthors)) {
            foreach ($xmlReader->getElements('/p:cmAuthorLst/p:cmAuthor') as $oElementAuthor) {
                if (!$oElementAuthor->hasAttribute('id')) {
                    continue;
                }
                $this->authorMap[$oElementAuthor->getAttribute('id')] = $oElementAuthor->getAttribute('name');
            }
        }
    }

    private function loadBlipFill(\DOMElement $oElement)
    {
        if (!$oElement->hasAttribute('r:embed')) {
            return '';
        }

        $rels = $this->arrayRels[$this->fileRels];
        if (!isset($rels)) {
            return '';
        }

        $resource = $rels[$oElement->getAttribute('r:embed')];
        if (!isset($resource)) {
            return '';
        }

        $targetPath = $resource['Target'];
        if (!isset($targetPath)) {
            return '';
        }

        $pathImage = 'ppt/slides/' . $targetPath;
        $pathImage = explode('/', $pathImage);
        foreach ($pathImage as $key => $partPath) {
            if ($partPath == '..') {
                unset($pathImage[$key - 1]);
                unset($pathImage[$key]);
            }
        }
        $pathImage = implode('/', $pathImage);
        $imageSource = "zip://{$this->filename}#{$pathImage}";

        return $imageSource;
    }
}
