<?php

namespace PhpOffice\PhpPresentation\Reader;

use PhpOffice\Common\Drawing;
use PhpOffice\PhpPresentation\PhpPresentation;
use PhpOffice\PhpPresentation\DocumentLayout;
use PhpOffice\PhpPresentation\Slide;
use PhpOffice\PhpPresentation\Slide\SlideMaster;
use PhpOffice\PhpPresentation\Slide\SlideLayout;
use PhpOffice\PhpPresentation\Slide\Transition;
use PhpOffice\PhpPresentation\Slide\Animation;
use PhpOffice\PhpPresentation\Style\Alignment;
use PhpOffice\PhpPresentation\Style\Border;
use PhpOffice\PhpPresentation\Style\Bullet;
use PhpOffice\PhpPresentation\Style\Color;
use PhpOffice\PhpPresentation\Style\Fill;
use PhpOffice\PhpPresentation\Style\Font;
use PhpOffice\PhpPresentation\Style\TextStyle;
use PhpOffice\PhpPresentation\Shape\Chart\Gridlines;
use PhpOffice\PhpPresentation\Shape\Chart\Series;
use PhpOffice\PhpPresentation\Shape\Chart\Type\AbstractType;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Area;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Bar;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Line;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Pie;
use PhpOffice\PhpPresentation\Shape\Chart\Type\Scatter;
use PhpOffice\PhpPresentation\Shape\Chart\Marker;
use PhpOffice\PhpPresentation\Shape\Comment\Author;
use PhpOffice\PhpPresentation\Shape\Placeholder;
use PhpOffice\PhpPresentation\Shape\RichText;
use PhpOffice\PhpPresentation\Shape\RichText\Paragraph;
use PhpOffice\PhpPresentation\Shape\Drawing\Gd;

class ProseMirror implements ReaderInterface
{
    const BASE_INDENT_VALUE = -24;
    const BASE_MARGIN_LEFT_VALUE = 24;
    const MARGIN_LEFT_INCREMENT = 48;
    const LEFT_PADDING_INCREMENT = 30;

    const DEFAULT_SHAPE_FILL = '#CCE6FF';
    const DEFAULT_SHAPE_STROKE = '#2E3133';
    const DEFAULT_LINE_STROKE = '#000000';

    private $phpPresentation;
    private $filename = '';
    private $doc;
    private $attachment;
    private $comments;
    private $arraySlideLayouts = array();
    private $arrayHolderUniqueIdMap = array();
    private $holderUniqueId = 0;
    private $slideIdIdxMap = array();
    private $authorMap = array();
    private $authorIndex = 1;
    private $commentIndex = 1;
    private $shapeMap = array();

    private $shapeNameMapping = array(
        'rect'                  => 'rect',
        'rect_round'            => 'roundRect',
        'rect_single_snip'      => 'snip1Rect',
        'rect_snip_same_side'   => 'snip2SameRect',
        'rect_snip_diagonal'    => 'snip2DiagRect',
        'rect_snip_round'       => 'snipRoundRect',
        'rect_single_round'     => 'round1Rect',
        'rect_round_same_side'  => 'round2SameRect',
        'rect_round_diagonal'   => 'round2DiagRect',
        'ellipse'               => 'ellipse',
        'triangle'              => 'triangle',
        'right_triangle'        => 'rtTriangle',
        'parallelogram'         => 'parallelogram',
        'trapezoid'             => 'trapezoid',
        'diamond'               => 'diamond',
        'pentagon'              => 'pentagon',
        'hexagon'               => 'hexagon',
        'heptagon'              => 'heptagon',
        'octagon'               => 'octagon',
        'decagon'               => 'decagon',
        'dodecagon'             => 'dodecagon',
        'pie'                   => 'pie',
        'chord'                 => 'chord',
        'rect_frame'            => 'frame',
        'half_frame'            => 'halfFrame',
        'l_shape'               => 'corner',
        'diagonal_stripe'       => 'diagStripe',
        'cross'                 => 'plus',
        'plaque'                => 'plaque',
        'can'                   => 'can',
        'cube'                  => 'cube',
        'bevel'                 => 'bevel',
        'donut'                 => 'donut',
        'block_arc'             => 'blockArc',
        'folded_corner'         => 'foldedCorner',
        'lightning_bolt'        => 'lightningBolt',
        'moon'                  => 'moon',
        '4_point_star'          => 'star4',
        'arrow'                 => 'rightArrow',
        'arrow_left'            => 'leftArrow',
        'arrow_up'              => 'upArrow',
        'arrow_down'            => 'downArrow',
        'arrow_left_right'      => 'leftRightArrow',
        'arrow_up_down'         => 'upDownArrow',
        'arrow_striped'         => 'stripedRightArrow',
        'arrow_notched'         => 'notchedRightArrow',
        'arrow_pentagon'        => 'homePlate',
        'arrow_chevron'         => 'chevron'
    );

    private $defaultShapeCustPoints = array(
        'bevel' => array('adj' => 12500),
        'blockArc' => array('adj1' => 10800000, 'adj2' => 0, 'adj3' => 25000),
        'can' => array('adj' => 25000),
        'chevron' => array('adj' => 50000),
        'chord' => array('adj1' => 2700000, 'adj2' => 16200000),
        'corner' => array('adj1' => 50000, 'adj2' => 50000),
        'cube' => array('adj' => 25000),
        'decagon' => array('vf' => 105146),
        'diagStripe' => array('adj' => 50000),
        'donut' => array('adj' => 25000),
        'downArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'upArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'foldedCorner' => array('adj' => 16667),
        'frame' => array('adj1' => 12500),
        'halfFrame' => array('adj1' => 33333, 'adj2' => 33333),
        'heptagon' => array('hf' => 102572, 'vf' => 105210),
        'hexagon' => array('adj' => 25000, 'vf' => 115470),
        'homePlate' => array('adj' => 50000),
        'leftArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'leftRightArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'moon' => array('adj' => 50000),
        'notchedRightArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'octagon' => array('adj' => 29289),
        'parallelogram' => array('adj' => 25000),
        'pentagon' => array('hf' => 105146, 'vf' => 110557),
        'pie' => array('adj1' => 0, 'adj2' => 16200000),
        'plaque' => array('adj' => 16667),
        'plus' => array('adj' => 25000),
        'rightArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'round1Rect' => array('adj' => 16667),
        'round2DiagRect' => array('adj1' => 16667, 'adj2' => 0),
        'round2SameRect' => array('adj1' => 16667, 'adj2' => 0),
        'roundRect' => array('adj' => 16667),
        'snip1Rect' => array('adj' => 16667),
        'snip2DiagRect' => array('adj1' => 0, 'adj2' => 16667),
        'snip2SameRect' => array('adj1' => 16667, 'adj2' => 0),
        'star4' => array('adj' => 12500),
        'stripedRightArrow' => array('adj1' => 50000, 'adj2' => 50000),
        'trapezoid' => array('adj' => 25000),
        'triangle' => array('adj' => 50000),
        'upDownArrow' => array('adj1' => 50000, 'adj2' => 50000)
    );

    public function canRead($filename)
    {
        try {
            $handle = fopen($filename, 'r');
            $jsonString = fread($handle, filesize($filename));
            fclose($handle);
            json_decode($jsonString);
        } catch (Exception $e) {
            return false;
        }
        return true;
    }

    public function load($filename)
    {
        $this->phpPresentation = new PhpPresentation();
        $this->phpPresentation->removeSlideByIndex();
        $this->phpPresentation->setAllMasterSlides(array());
        $this->filename = $filename;

        $handle = fopen($filename, 'r');
        $jsonString = fread($handle, filesize($filename));
        fclose($handle);
        $json = json_decode($jsonString);
        $this->doc = $json->content->data;
        $this->attachment = $json->attachment;
        $this->comments = array();
        foreach ($json->comment->data as $comment) {
            $this->comments[$comment->id] = $comment;
        }

        $this->loadDocumentLayout();
        $this->loadMasterSlides();
        $this->loadSlides();

        $created = intval($json->created);
        $this->phpPresentation->getDocumentProperties()->setCreated($created);
        $this->phpPresentation->getDocumentProperties()->setModified($created);

        return $this->phpPresentation;
    }

    private function loadDocumentLayout()
    {
        $layout = $this->phpPresentation->getLayout();
        $layout->setCX(Drawing::pixelsToEmu($this->doc->attrs->cx));
        $layout->setCY(Drawing::pixelsToEmu($this->doc->attrs->cy));
    }

    private function loadMasterSlides()
    {
        $masterContainer = $this->doc->content[0];
        foreach ($masterContainer->content as $master) {
            $this->loadMasterSlide($master);
        }
    }

    private function loadSlides()
    {
        $slideContainer = $this->doc->content[1];
        if (!is_array($slideContainer->content) || 0 === count($slideContainer->content)) {
            return;
        }
        $slideIdx = 0;
        foreach ($slideContainer->content as $slide) {
            $this->slideIdIdxMap[$slide->attrs->id] = $slideIdx;
            $slideIdx++;
        }

        foreach ($slideContainer->content as $slideIndex => $slide) {
            $this->commentIds = array();
            if (property_exists($slide->attrs, 'synocomment')) {
                foreach ($slide->attrs->synocomment as $id) {
                    $this->commentIds[$id] = true;
                }
            }
            $this->loadSlide($slide, $slideContainer);

            $phpSlide = $this->phpPresentation->getSlide($slideIndex);
            foreach ($this->commentIds as $id => $value) {
                $comment = $this->comments[$id];
                if (!isset($comment)) {
                    continue;
                }
                $this->loadComment($comment, $phpSlide);

                $parnetCommentIndex = $this->commentIndex;
                $this->commentIndex++;
                foreach ($comment->replies as $reply) {
                    $this->loadComment($reply, $phpSlide, $parnetCommentIndex);
                    $this->commentIndex++;
                }
            }
        }
    }

    private function loadComment($comment, $phpSlide, $parnetCommentIndex = 0)
    {
        $phpComment = $phpSlide->createComment();

        if (!isset($this->authorMap[$comment->author])) {
            $this->authorMap[$comment->author] = $this->authorIndex;
            $this->authorIndex++;
        }
        $phpAuthor = new Author();
        $phpAuthor->setIndex($this->authorMap[$comment->author]);
        $phpAuthor->setName($comment->author);
        $phpComment->setAuthor($phpAuthor);
        $phpComment->setDate(round($comment->timestamp / 1000));
        $phpComment->setIndex($this->commentIndex);
        $phpComment->setText($comment->text);

        if (0 < $parnetCommentIndex) {
            $phpComment->setParentIndex($parnetCommentIndex);
            $phpComment->setIsReply(true);
        }
    }

    private function loadMasterSlide($master)
    {
        $phpMaster = $this->phpPresentation->createMasterSlide();
        $phpMaster->setTextStyles(new TextStyle(false));
        $this->setOtherStyles($phpMaster);
        $this->loadSlideBackground($master, $phpMaster);

        $this->holderUniqueId = 0;
        $this->loadContent($master, $phpMaster);

        foreach ($master->content[2]->content as $layout) {
            $phpMaster->addSlideLayout($this->loadLayoutSlide($layout, $phpMaster));
        }
    }

    private function loadSlide($slide, $container)
    {
        $phpSlide = $this->phpPresentation->createSlide();
        $phpSlide->setIsVisible($slide->attrs->is_visible);
        $this->loadTransition($slide, $container, $phpSlide);
        $this->loadSlideBackground($slide, $phpSlide);
        $this->loadContent($slide, $phpSlide);
        $this->loadAnimations($slide, $phpSlide);
        $phpSlide->setSlideLayout($this->arraySlideLayouts[$slide->attrs->layout_id]);
        $this->loadNote($slide->content[2], $phpSlide);
    }

    private function loadTransition($slide, $container, $phpSlide)
    {
        ($transition = $slide->attrs->transition) || ($transition = $container->attrs->transition);
        if (is_null($transition) || $transition->type == 'none') {
            return;
        }
        $phpTransition = new Transition();
        $phpTransition->setManualTrigger(true);
        if ($transition->duration <= 500) {
            $phpTransition->setSpeed(Transition::SPEED_FAST);
        } else if ($transition->duration <= 750) {
            $phpTransition->setSpeed(Transition::SPEED_MEDIUM);
        } else {
            $phpTransition->setSpeed(Transition::SPEED_SLOW);
        }
        $phpTransition->setDuration($transition->duration);
        switch ($transition->type) {
            case 'dissolve':
                $phpTransition->setTransitionType(Transition::TRANSITION_FADE);
                break;
            case 'fade':
                $phpTransition->setTransitionType(Transition::TRANSITION_FADE_THROUGH_BLACK);
                break;
            case 'slide':
                switch ($transition->options->from) {
                    case 'left':
                        $phpTransition->setTransitionType(Transition::TRANSITION_PUSH_RIGHT);
                        break;
                    case 'right':
                        $phpTransition->setTransitionType(Transition::TRANSITION_PUSH_LEFT);
                        break;
                    case 'top':
                        $phpTransition->setTransitionType(Transition::TRANSITION_PUSH_DOWN);
                        break;
                    case 'bottom':
                        $phpTransition->setTransitionType(Transition::TRANSITION_PUSH_UP);
                        break;
                }
                break;
        }
        $phpSlide->setTransition($phpTransition);
    }

    private function loadAnimations($slide, $phpSlide)
    {
        if (!(property_exists($slide, 'attrs') && property_exists($slide->attrs, 'animations'))) {
            return;
        }

        foreach ($slide->attrs->animations as $index=>$animation) {
            if (!array_key_exists($animation->nodeId, $this->shapeMap)) {
                continue;
            }

            $phpAnimation = new Animation();
            switch ($animation->start) {
                case 'click':
                    $phpAnimation->setCondition(Animation::CONDITION_CLICK);
                    break;
                case 'with':
                    if ($index === 0) {
                        $phpAnimation->setCondition(Animation::CONDITION_AFTER);
                    } else {
                        $phpAnimation->setCondition(Animation::CONDITION_WITH);
                    }
                    break;
                case 'after':
                    $phpAnimation->setCondition(Animation::CONDITION_AFTER);
                    break;
            }
            $phpAnimation->setDuration($animation->duration);
            switch ($animation->type) {
                case 'appear':
                    $type = $animation->options->entrance ? Animation::ANIMATION_APPEAR : Animation::ANIMATION_DISAPPEAR;
                    $phpAnimation->setType($type);
                    break;
                case 'fade':
                    $type = $animation->options->entrance ? Animation::ANIMATION_FADE_IN : Animation::ANIMATION_FADE_OUT;
                    $phpAnimation->setType($type);
                    break;
            }
            $phpAnimation->setShape($this->shapeMap[$animation->nodeId]);
            $phpSlide->addAnimation($phpAnimation);
        }
    }

    private function loadNote($note, $phpSlide)
    {
        $phpNote = $phpSlide->getNote();
        $phpShape = $phpNote->createRichTextShape();
        $phpShape->setParagraphs(array());

        if (!property_exists($note, 'content') || !is_array($note->content)) {
            return;
        }

        foreach ($note->content as $item) {
            if ('paragraph' === $item->type) {
                $phpParagraph = $phpShape->createParagraph();
                $phpParagraph->setRichTextElements(array());
                $this->loadParagraph($item, $phpParagraph);
            } else {
                $this->loadList($item, $phpShape, 0);
            }
        }
    }

    private function loadLayoutSlide($layout, $phpMaster)
    {
        $phpLayout = new SlideLayout($phpMaster);

        //FIXME: layout name
        $phpLayout->setLayoutName($layout->attrs->id);
        $this->loadSlideBackground($layout, $phpLayout);
        $this->holderUniqueId = 0;
        $this->loadContent($layout, $phpLayout);
        $this->arraySlideLayouts[$layout->attrs->id] = &$phpLayout;

        return $phpLayout;
    }

    private function loadSlideBackground($slide, $phpSlide)
    {
        $background = $slide->content[0];
        if (!property_exists($background, 'attrs')) {
            return;
        }

        if (property_exists($background->attrs, 'file_id') && '' !== $background->attrs->file_id) {
            $fileId = $background->attrs->file_id;
            $attachment = $this->attachment->$fileId;
            $dir = dirname($attachment->path);
            $newPath = $dir . DIRECTORY_SEPARATOR . $fileId . '.' . $attachment->ext;

            copy($attachment->path, $newPath);
            $bgImage = new Slide\Background\Image();
            $bgImage->setPath($newPath);

            if (property_exists($background->attrs, 'opacity')) {
                $bgImage->setAlpha(floatVal($background->attrs->opacity) * 100);
            }

            $phpSlide->setBackground($bgImage);
        } else if (property_exists($background->attrs, 'color')) {
            if (property_exists($background->attrs, 'opacity')) {
                $alpha = strtoupper(dechex(floatval($background->attrs->opacity) * 255));
                if (2 > strlen($alpha)) {
                    $alpha = '0' . $alpha;
                }
            }
            $strColor = $this->parseColor($background->attrs->color);
            if ('transparent' === $strColor) {
                $alpha = '00';
                $strColor = 'FFFFFF';
            }
            $color = new Color();
            $color->setRGB($strColor, $alpha);
            $bgColor = new Slide\Background\Color();
            $bgColor->setColor($color);
            $phpSlide->setBackground($bgColor);
        }
    }

    private function loadContent($slide, $phpSlide)
    {
        $contentContainer = $slide->content[1];
        if (!property_exists($contentContainer, 'content')) {
            return;
        }
        foreach ($contentContainer->content as $floatingObject) {
            if (property_exists($floatingObject->attrs, 'synocomment')) {
                foreach ($floatingObject->attrs->synocomment as $id) {
                    $this->commentIds[$id] = true;
                }
            }
            if ('text_box' === $floatingObject->type || 'shape_box' === $floatingObject->type) {
                $this->loadWithTextBox($floatingObject, $phpSlide);
            } else if ('image_box' === $floatingObject->type) {
                $this->loadImageBox($floatingObject, $phpSlide);
            } else if ('line_box' === $floatingObject->type) {
                $this->loadLineBox($floatingObject, $phpSlide);
            } else if ('table_box' === $floatingObject->type) {
                $this->loadTableBox($floatingObject, $phpSlide);
            } else if ('chart_box' === $floatingObject->type) {
                $this->loadChartBox($floatingObject, $phpSlide);
            } else if ('object_group' === $floatingObject->type) {
                $this->loadObjectGroup($floatingObject, $phpSlide);
            } else if (false !== strpos($floatingObject->type, '_slide_holder')) {
                $this->loadInstance($slide, $floatingObject, $phpSlide);
            } else if (false !== strpos($floatingObject->type, '_holder')) {
                $this->loadHolder($floatingObject, $phpSlide);
            }
        }
    }

    private function loadWithTextBox($box, $phpSlide)
    {
        $phpShape = $phpSlide->createRichTextShape();
        $this->shapeMap[$box->attrs->id] = $phpShape;
        $phpShape->setParagraphs(array());

        $shape = $box->content[0];
        if ('text_box' === $box->type) {
            $phpShape->setTextBox(true);
            $phpShape->setAutoFit(RichText::AUTOFIT_DEFAULT);
            $phpShape->setShapeType('rect');
        } else {
            $phpShape->setShapeType($this->shapeNameMapping[$shape->attrs->shape_type]);
        }
        $this->loadTransform($box, $phpShape);
        $this->loadSynolink($box, $phpShape);
        $this->loadCustPoints($shape, $phpShape);
        $this->loadPathConfig($shape, $phpShape);

        $block = $shape->content[0];
        $vAlign = $this->loadVerticalAlignment($block);
        $this->loadPaddings($block, $phpShape);
        $phpShape->getAlignment()->setVertical($vAlign);

        foreach ($block->content as $item) {
            if ('paragraph' === $item->type) {
                $phpParagraph = $phpShape->createParagraph();
                $phpParagraph->setRichTextElements(array());
                $this->loadParagraph($item, $phpParagraph);
            } else {
                $this->loadList($item, $phpShape, 0);
            }
        }

        if (count($phpShape->getParagraphs()) > 0) {
            $phpShape->setActiveParagraph(0);
        }
    }

    private function loadImageBox($imageBox, $phpSlide)
    {
        $phpShape = new Gd();
        $this->shapeMap[$imageBox->attrs->id] = $phpShape;
        $phpShape->getShadow()->setVisible(false);

        $this->loadTransform($imageBox, $phpShape);
        $this->loadSynolink($imageBox, $phpShape);

        $image = $imageBox->content[0];
        $phpShape->setName($image->attrs->alt);
        $phpShape->setDescription($image->attrs->alt);

        $width = $phpShape->getWidth();
        $height = $phpShape->getHeight();
        $imgTransform = $image->attrs->img_transform;
        $cropTop = 0;
        $cropRight = 0;
        $cropBottom = 0;
        $cropLeft = 0;

        if (1 !== $imgTransform->width) {
            $orgWidth = $width * $imgTransform->width;
            $cropLeft = -1 * $imgTransform->left * 100000 * $width / $orgWidth;
            $cropRight = (100000 * ($orgWidth - $width) - $orgWidth * $cropLeft) / $orgWidth;
        }
        if (1 !== $imgTransform->height) {
            $orgHeight = $height * $imgTransform->height;
            $cropTop = -1 * $imgTransform->top * 100000 * $height / $orgHeight;
            $cropBottom = (100000 * ($orgHeight - $height) - $orgHeight * $cropTop) / $orgHeight;
        }
        $phpShape->setCrop(round($cropTop), round($cropRight), round($cropBottom), round($cropLeft));

        $fileId = $image->attrs->file_id;
        if (isset($this->attachment->$fileId)) {
            $phpShape->setFileId($fileId);
            $phpShape->setSource($this->attachment->$fileId->path);
            $phpShape->setExtension($this->attachment->$fileId->ext);
            //$phpShape->checkImage();
        }
        if (isset($image->attrs->opacity)) {
            $phpShape->setAlpha($image->attrs->opacity * 100);
        }
        $shapeType = $image->attrs->shape_type;
        if (isset($shapeType) && array_key_exists($shapeType, $this->shapeNameMapping)) {
            $phpShape->setShapeType($this->shapeNameMapping[$shapeType]);
        }

        $phpSlide->addShape($phpShape);
    }

    private function loadLineBox($lineBox, $phpSlide)
    {
        $phpLine = $phpSlide->createLineShape();
        $this->shapeMap[$lineBox->attrs->id] = $phpLine;

        $line = $lineBox->content[0];
        $this->loadTransform($lineBox, $phpLine);
        $this->loadSynolink($lineBox, $phpLine);
        $this->loadCustPoints($line, $phpLine);
        $this->loadPathConfig($line, $phpLine);

        $border = $phpLine->getBorder();
        if (Border::LINE_NONE === $border->getLineStyle() || Border::LINE_NOTSET === $border->getLineStyle()) {
            $border->setLineStyle(Border::LINE_SINGLE);
        }

        if (property_exists($line, 'attrs')) {
            $line_attrs = $line->attrs;

            if ('connect_line' === $line_attrs->line_type) {
                $phpLine->setShapeType('bentConnector3');
            } else if (property_exists($line_attrs, 'arrow')) {
                $phpLine->setShapeType('straightConnector1');
            } else {
                $phpLine->setShapeType('line');
            }

            if (property_exists($line_attrs, 'arrow_start') && $line_attrs->arrow_start !== 'none') {
                $phpLine->setHeadEnd(true);
            }
            if (property_exists($line_attrs, 'arrow_end') && $line_attrs->arrow_end !== 'none') {
                $phpLine->setTailEnd(true);
            }
        } else {
            $phpLine->setShapeType('line');
        }
    }

    private function loadTableBox($tableBox, $phpSlide)
    {
        $phpTable = $phpSlide->createTableShape();
        $this->shapeMap[$tableBox->attrs->id] = $phpTable;

        $table = $tableBox->content[0]->content[0];
        $this->loadTransform($tableBox, $phpTable);
        //NOTE: correct width/height is in table attrs transform of table box
        $phpTable->setWidth(floatval($table->attrs->width));
        $phpTable->setHeight(floatval($table->attrs->height));

        $colgroup = $table->attrs->colgroup;
        $phpTable->setNumColumns(count($colgroup));
        $phpTable->createRow();
        foreach ($colgroup as $idx => $col) {
            $phpTable->getRow(0)->getCell($idx)->setWidth(floatval($col->width));
        }

        $rows = $table->content;
        foreach ($rows as $rowIdx => $row) {
            $phpRow = $phpTable->getRow($rowIdx, true);
            if (is_null($phpRow)) {
                $phpRow = $phpTable->createRow();
            }
            $phpRow->setHeight(floatval($row->attrs->height));

            $cells = $row->content;
            foreach($cells as $cellIdx => $cell) {
                $phpCell = $phpRow->getCell($cellIdx);
                if (true === $cell->attrs->hMerge) {
                    $phpCell->setHMerge();
                }
                if (true === $cell->attrs->vMerge) {
                    $phpCell->setVMerge();
                }
                if (true === $cell->attrs->hMerge || true === $cell->attrs->vMerge) {
                    continue;
                }
                if (1 < $cell->attrs->colspan) {
                    $phpCell->setColSpan($cell->attrs->colspan);
                    for ($count = 1; $count < $cell->attrs->colspan; $count++) {
                        $cells[$cellIdx + $count]->attrs->hMerge = true;
                    }
                }
                if (1 < $cell->attrs->rowspan) {
                    $phpCell->setRowSpan($cell->attrs->rowspan);
                    for ($count = 1; $count < $cell->attrs->rowspan; $count++) {
                        $rows[$rowIdx + $count]->content[$cellIdx]->attrs->vMerge = true;
                    }
                }
                $phpCell->setParagraphs(array());
                $this->loadCellStyle($cell, $phpCell);
                foreach ($cell->content as $item) {
                    if ('paragraph' === $item->type) {
                        $phpParagraph = $phpCell->createParagraph();
                        $phpParagraph->setRichTextElements(array());
                        $this->loadParagraph($item, $phpParagraph);
                    } else {
                        $this->loadList($item, $phpCell, 0);
                    }
                }
            }
        }
    }

    private function loadChartBox($chartBox, $phpSlide)
    {
        $phpChart = $phpSlide->createChartShape();
        $this->shapeMap[$chartBox->attrs->id] = $phpChart;
        $this->loadTransform($chartBox, $phpChart);

        $chart = $chartBox->content[0];
        $config = $chart->attrs->config;

        $phpChart->getTitle()->setText($config->title->text);
        $plotArea = $config->plotArea;

        $subCharts = new \stdClass();
        $seriesCount = 0;
        foreach ($plotArea->series as $series) {
            $type = $series->type;
            if (!isset($subCharts->$type)) {
                if ('area' === $type) {
                    $subCharts->$type = new Area();
                } else if ('bar' === $type) {
                    $subCharts->$type = new Bar();
                } else if ('line' === $type) {
                    $subCharts->$type = new Line();
                } else if ('pie' === $type) {
                    $subCharts->$type = new Pie();
                } else if ('scatter' === $type) {
                    $subCharts->$type = new Scatter();
                } else {
                    continue;
                }
            }
            $phpType = $subCharts->$type;
            $title = $series->tx;
            $values = array();
            foreach ($series->cat as $idx => $cat) {
                $values[$cat] = $series->val[$idx];
            }

            $series = new Series($title, $values);
            $series->setIndex($seriesCount);
            $series->setOrder($seriesCount);
            if ($phpType instanceof Scatter) {
                $series->getMarker()->setSymbol(Marker::SYMBOL_CIRCLE);
            }
            $phpType->addSeries($series);
            $seriesCount++;
        }

        $direction = ('row' === $config->barDir) ? AbstractType::DIRECTION_HORIZONTAL : AbstractType::DIRECTION_VERTICAL;
        $grouping = ('stacked' === $config->grouping) ? AbstractType::GROUPING_STACKED : AbstractType::GROUPING_CLUSTERED;
        $phpPlotArea = $phpChart->getPlotArea();
        foreach ($subCharts as $phpType) {
            $phpType->setDirection($direction);
            $phpType->setGrouping($grouping);
            $phpPlotArea->addType($phpType);
            $phpCatAx = $phpPlotArea->getAxisX();
            $phpValAx = $phpPlotArea->getAxisY();

            if ($phpType instanceof Scatter) {
                $phpCatAx->setIsSourceLinked(false);
            }

            $gridlines = new Gridlines();
            $gridlines->getOutline()->setWidth(1);
            $gridlines->getOutline()->getFill()->setFillType(Fill::FILL_SOLID);
            $gridlines->getOutline()->getFill()->setStartColor(new Color('FFE2E2E2'));
            $phpCatAx->setMajorGridlines($gridlines);
            $phpValAx->setMajorGridlines($gridlines);

            $phpCatAx->getOutline()->setWidth(1);
            $phpCatAx->getOutline()->getFill()->setFillType(Fill::FILL_SOLID);
            $phpCatAx->getOutline()->getFill()->setStartColor(new Color('FF282828'));
        }
    }

    private function loadObjectGroup($group, $phpSlide)
    {
        $phpGroup = $phpSlide->createGroup();
        $this->shapeMap[$group->attrs->id] = $phpGroup;
        $this->loadTransform($group, $phpGroup);
        foreach ($group->content as $floatingObject) {
            if (property_exists($floatingObject->attrs, 'synocomment')) {
                foreach ($floatingObject->attrs->synocomment as $id) {
                    $this->commentIds[$id] = true;
                }
            }

            $floatingObject->attrs->transform->top += $group->attrs->transform->top;
            $floatingObject->attrs->transform->left += $group->attrs->transform->left;
            if ('text_box' === $floatingObject->type || 'shape_box' === $floatingObject->type) {
                $this->loadWithTextBox($floatingObject, $phpGroup);
            } else if ('image_box' === $floatingObject->type) {
                $this->loadImageBox($floatingObject, $phpGroup);
            } else if ('line_box' === $floatingObject->type) {
                $this->loadLineBox($floatingObject, $phpGroup);
            } else if ('chart_box' === $floatingObject->type) {
                $this->loadChartBox($floatingObject, $phpSlide);
            } else if ('object_group' === $floatingObject->type) {
                $this->loadObjectGroup($floatingObject, $phpGroup);
            }
        }
    }

    private function loadHolder($holder, $phpSlide)
    {
        global $i18n;
        $text = '';
        if (false !== strpos($holder->type, 'sub_title_')) {
            $type = Placeholder::PH_TYPE_SUBTITLE;
            $text = $i18n['sub_title'];
        } else if (false !== strpos($holder->type, 'title_')) {
            $type = Placeholder::PH_TYPE_TITLE;
            $text = $i18n['title'];
        } else if (false !== strpos($holder->type, 'body_')) {
            $type = Placeholder::PH_TYPE_BODY;
        } else if (false !== strpos($holder->type, 'page_number')) {
            $type = Placeholder::PH_TYPE_SLIDENUM;
        } else {
            return;
        }

        $addToMap = ($phpSlide instanceof SlideMaster) ? false : true;

        $phpShape = $phpSlide->createRichTextShape();
        $this->shapeMap[$holder->attrs->id] = $phpShape;
        $phpShape->setParagraphs(array());
        $shape = $holder->content[0];

        if ($addToMap) {
            $layoutId = $phpSlide->getLayoutName();
            $holderMapId = $type . $holder->attrs->type_id;
            if (!isset($this->arrayHolderUniqueIdMap[$layoutId])) {
                $this->arrayHolderUniqueIdMap[$layoutId] = array();
            }
            $this->arrayHolderUniqueIdMap[$layoutId][$holderMapId] = $this->holderUniqueId;
        }

        $placeholder = new Placeholder($type);
        $placeholder->setIdx($this->holderUniqueId);
        $phpShape->setPlaceHolder($placeholder);

        $this->loadTransform($holder, $phpShape);
        $this->loadPathConfig($shape, $phpShape);

        $textBlock = $shape->content[0];
        $vAlign = $this->loadVerticalAlignment($textBlock);
        $phpShape->getAlignment()->setVertical($vAlign);

        if (Placeholder::PH_TYPE_BODY === $type) {
            $this->loadBodyStyle($textBlock->content[0], $phpShape, $phpSlide);
            $phpShape->setActiveParagraph(0);
        } else {
            $paragraph = $textBlock->content[0];
            $phpParagraph = $phpShape->createParagraph();
            $this->loadParagraph($paragraph, $phpParagraph, false);

            if (!($phpSlide instanceof SlideLayout)) {
                if (Placeholder::PH_TYPE_TITLE === $type) {
                    $phpParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NONE);
                    $phpSlide->getTextStyles()->setTitleStyleAtLvl($phpParagraph, 0);
                }
            }
            if (Placeholder::PH_TYPE_SUBTITLE === $type) {
                $phpParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NONE);
            }

            $phpParagraph->setRichTextElements(array());
            $phpParagraph->createTextRun()->setText($text);
            $phpShape->setActiveParagraph(0);
        }

        $this->holderUniqueId++;
    }

    private function loadInstance($slide, $instance, $phpSlide)
    {
        if (false !== strpos($instance->type, 'sub_title_')) {
            $type = PlaceHolder::PH_TYPE_SUBTITLE;
        } else if (false !== strpos($instance->type, 'title_')) {
            $type = PlaceHolder::PH_TYPE_TITLE;
        } else if (false !== strpos($instance->type, 'body_')) {
            $type = PlaceHolder::PH_TYPE_BODY;
        } else if (false !== strpos($instance->type, 'page_number')) {
            $type = PlaceHolder::PH_TYPE_SLIDENUM;
        } else {
            return;
        }
        $phpShape = $phpSlide->createRichTextShape();
        $this->shapeMap[$instance->attrs->id] = $phpShape;
        $phpShape->setParagraphs(array());
        $shape = $instance->content[0];

        $holderId = $this->arrayHolderUniqueIdMap[$slide->attrs->layout_id][$type . $instance->attrs->type_id];
        $placeholder = new Placeholder($type);
        $placeholder->setIdx($holderId);
        $phpShape->setPlaceHolder($placeholder);

        $this->loadTransform($instance, $phpShape);
        $this->loadPathConfig($shape, $phpShape);

        $block = $shape->content[0];
        $vAlign = $this->loadVerticalAlignment($block);
        $this->loadPaddings($block, $phpShape);
        $phpShape->getAlignment()->setVertical($vAlign);

        foreach ($block->content as $item) {
            if ('paragraph' === $item->type) {
                $phpParagraph = $phpShape->createParagraph();
                $phpParagraph->setRichTextElements(array());
                $phpParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NONE);
                $this->loadParagraph($item, $phpParagraph);
            } else {
                $this->loadList($item, $phpShape, 0);
            }
        }

        if (count($phpShape->getParagraphs()) > 0) {
            $phpShape->setActiveParagraph(0);
        }
    }

    private function loadTransform($floatingObject, $phpShape)
    {
        if (!property_exists($floatingObject->attrs, 'transform')) {
            return;
        }

        $transform = $floatingObject->attrs->transform;
        if (0 !== $transform->deg) {
            $rad = deg2rad($transform->deg);
            $transform->left = $transform->left + $transform->width / 2 * cos($rad) - $transform->width / 2 - $transform->height / 2 * sin($rad);
            $transform->top = $transform->top + $transform->height / 2 * cos($rad) - $transform->height / 2 + $transform->width / 2 * sin($rad);
        }

        $width = $transform->width;
        $height = $transform->height;
        if (0 > $width) {
            $phpShape->setFlipH();
            $transform->left += $width;
            $width = abs($width);
        }
        if (0 > $height) {
            $phpShape->setFlipV();
            $transform->top += $height;
            $height = abs($height);
        }
        $phpShape->setWidth($width);
        $phpShape->setHeight($height);
        $phpShape->setOffsetX($transform->left);
        $phpShape->setOffsetY($transform->top);
        $phpShape->setRotation($transform->deg);
    }

    private function loadCustPoints($shape, $phpShape)
    {
        if (!property_exists($shape, 'attrs') || !property_exists($shape->attrs, 'cust_points')) {
            return;
        }
        $custPoints = new \stdClass();

        foreach ($shape->attrs->cust_points as $name => $value) {
            $custPoints->$name = $value;
        }

        $defaultCustPoints = $this->defaultShapeCustPoints[$phpShape->getShapeType()];
        if (isset($defaultCustPoints)) {
            foreach ($defaultCustPoints as $name => $value) {
                if (!property_exists($custPoints, $name)) {
                    $custPoints->$name = $value;
                }
            }
        }

        $phpShape->setCustPoints($custPoints);
    }

    private function loadPathConfig($shape, $phpShape)
    {
        $strokeDasharray = 'stroke-dasharray';
        $strokeOpacity = 'stroke-opacity';
        $strokeWidth = 'stroke-width';
        $fillOpacity = 'fill-opacity';
        $arrowType = 'arrow-type';

        if (!property_exists($shape, 'attrs')) {
            return;
        }

        $stroke_config = property_exists($shape->attrs, 'stroke_config') ? $shape->attrs->stroke_config : new \stdClass();
        $fill_config = property_exists($shape->attrs, 'fill_config') ? $shape->attrs->fill_config : new \stdClass();
        if ('shape' === $shape->type) {
            if (!property_exists($shape->attrs, 'fill')) {
                $shape->attrs->fill = true;
            }
            if (!property_exists($fill_config, 'fill')) {
                $fill_config->fill = self::DEFAULT_SHAPE_FILL;
            }
            if (!property_exists($shape->attrs, 'stroke')) {
                $shape->attrs->stroke = true;
            }
            if (!property_exists($stroke_config, $strokeWidth)) {
                $stroke_config->$strokeWidth = 1;
            }
            if (!property_exists($stroke_config, 'stroke')) {
                $stroke_config->stroke = self::DEFAULT_SHAPE_STROKE;
            }
        } else if ('line' === $shape->type) {
            if (!property_exists($shape->attrs, 'stroke')) {
                $shape->attrs->stroke = true;
            }
            if (!property_exists($stroke_config, $strokeWidth)) {
                $stroke_config->$strokeWidth = 1;
            }
            if (!property_exists($stroke_config, 'stroke')) {
                $stroke_config->stroke = self::DEFAULT_LINE_STROKE;
            }
        } else {
            if (!property_exists($shape->attrs, 'stroke')) {
                $shape->attrs->stroke = false;
            }
            if (!property_exists($shape->attrs, 'fill')) {
                $shape->attrs->fill = false;
            }
        }

        //stroke
        if (property_exists($shape->attrs, 'stroke') && false === $shape->attrs->stroke) {
            $phpShape->getBorder()->setLineStyle(Border::LINE_NONE);
        } else {
            if (property_exists($stroke_config, $strokeWidth)) {
                $width = $stroke_config->$strokeWidth;
                $phpShape->getBorder()->setLineWidth($width);
                $phpShape->getBorder()->setLineStyle(Border::LINE_SINGLE);
            }

            if (property_exists($stroke_config, 'stroke')) {
                $alpha = 100;
                if (property_exists($stroke_config, $strokeOpacity)) {
                    $alpha = round(floatval($stroke_config->$strokeOpacity) * 100);
                }
                $alpha *= 2.55;
                $strAlpha = dechex($alpha);
                $strColor = $this->parseColor($stroke_config->stroke);
                if ('transparent' === $strColor) {
                    $strAlpha = '00';
                    $strColor = 'FFFFFF';
                }
                $argb = $strAlpha . $strColor;
                $color = new Color();
                $color->setARGB($argb);
                $phpShape->getBorder()->setColor($color);
            }

            if (property_exists($stroke_config, $strokeDasharray)) {
                switch ($stroke_config->$strokeDasharray) {
                    case '2,2':
                        $phpShape->getBorder()->setDashStyle(Border::DASH_SYSDOT);
                        break;
                    case '6,6':
                        $phpShape->getBorder()->setDashStyle(Border::DASH_SYSDASH);
                        break;
                    default:
                        $phpShape->getBorder()->setDashStyle(Border::DASH_SOLID);
                        break;
                }
            }
        }

        //fill
        if (property_exists($shape->attrs, 'fill') && false === $shape->attrs->fill) {
            $phpShape->getFill()->setFillType(Fill::FILL_NONE);
        } else if (property_exists($fill_config, 'fill')) {
            if ('#' == substr($fill_config->fill, 0, 1)) {
                $alpha = 100;
                if (property_exists($fill_config, $fillOpacity)) {
                    $alpha = round(floatval($fill_config->$fillOpacity) * 100);
                }
                $alpha *= 2.55;
                $strAlpha = dechex($alpha);
                $strColor = $this->parseColor($fill_config->fill);
                if ('transparent' === $strColor) {
                    $strAlpha = '00';
                    $strColor = 'FFFFFF';
                }
                $argb = $strAlpha . $strColor;
                $color = new Color();
                $color->setARGB($argb);
                $phpShape->getFill()->setStartColor($color);
                $phpShape->getFill()->setFillType(Fill::FILL_SOLID);
            } else {
                $fileId = $fill_config->fill;
                if (isset($this->attachment->$fileId)) {
                    $phpShape->getFill()->setFillType(Fill::FILL_BLIP);
                    $phpShape->getFill()->setImageFileId($fileId);
                    $phpShape->getFill()->setImageSource($this->attachment->$fileId->path);
                    $phpShape->getFill()->setImageExtension($this->attachment->$fileId->ext);
                }
            }
        }

        //arrow
        if (property_exists($shape->attrs, 'arrow')) {
            if ('end' === $shape->attrs->arrow) {
                $phpShape->setTailEnd();
            }
            if ('both' === $shape->attrs->arrow) {
                $phpShape->setHeadEnd();
                $phpShape->setTailEnd();
            }
        }
    }

    private function loadVerticalAlignment($item)
    {
        $vAlign = Alignment::VERTICAL_NONE;
        if (property_exists($item, 'attrs') && property_exists($item->attrs, 'vertical_align')) {
            if (0 === $item->attrs->vertical_align) {
                $vAlign = Alignment::VERTICAL_CENTER;
            } else if (-1 === $item->attrs->vertical_align) {
                $vAlign = Alignment::VERTICAL_TOP;
            } else {
                $vAlign = Alignment::VERTICAL_BOTTOM;
            }
        }
        return $vAlign;
    }

    private function loadParagraph($paragraph, $phpParagraph, $loadContent = true)
    {
        if (property_exists($paragraph->attrs, 'lineHeight') && is_numeric($paragraph->attrs->lineHeight)) {
            $phpParagraph->setLineSpacing(floatval($paragraph->attrs->lineHeight) * 100000);
        }

        if (property_exists($paragraph->attrs, 'paddingTop') && false !== strpos($paragraph->attrs->paddingTop, 'pt')) {
            $spaceBefore = ('' === $paragraph->attrs->paddingTop) ? 0 : floatval($paragraph->attrs->paddingTop) * 100;
            $phpParagraph->setSpaceBefore($spaceBefore);
        }

        if (property_exists($paragraph->attrs, 'paddingBottom') && false !== strpos($paragraph->attrs->paddingBottom, 'pt')) {
            $spaceAfter = ('' === $paragraph->attrs->paddingBottom) ? 0 : floatval($paragraph->attrs->paddingBottom) * 100;
            $phpParagraph->setSpaceAfter($spaceAfter);
        }

        $tabCount = 0;
        if (property_exists($paragraph->attrs, 'paddingLeft') && '' !== $paragraph->attrs->paddingLeft) {
            $val = intval($paragraph->attrs->paddingLeft);
            $tabCount = $val / self::LEFT_PADDING_INCREMENT;
        }

        if (property_exists($paragraph->attrs, 'textAlign')) {
            switch ($paragraph->attrs->textAlign) {
                case 'center':
                    $phpParagraph->getAlignment()->setHorizontal(Alignment::HORIZONTAL_CENTER);
                    break;
                case 'right':
                    $phpParagraph->getAlignment()->setHorizontal(Alignment::HORIZONTAL_RIGHT);
                    break;
                case 'justify':
                    $phpParagraph->getAlignment()->setHorizontal(Alignment::HORIZONTAL_JUSTIFY);
                    break;
                default:
                    break;
            }
        }

        if (property_exists($paragraph->attrs, 'textMark')) {
            $this->loadMarks($paragraph->attrs->textMark, $phpParagraph->getFont());
        }

        if (!$loadContent || !property_exists($paragraph, 'content')) {
            return;
        }

        foreach ($paragraph->content as $text) {
            if ('page_number' === $text->type) {
                continue;
            }
            $phpText = $phpParagraph->createTextRun();
            $this->loadMarks($text->marks, $phpText->getFont());
            $plainText = $text->text;
            for ($i = 0; $i < $tabCount; $i++) {
                $plainText = "\t" . $plainText;
            }
            $phpText->setText($plainText);
            $this->loadTextLink($text->marks, $phpText);
        }
    }

    private function loadList($list, $phpShape, $level)
    {
        if ('ordered_list' === $list->type) {
            $bulletType = Bullet::TYPE_NUMERIC;
        } else {
            $bulletType = Bullet::TYPE_BULLET;
        }

        foreach ($list->content as $listItem) {
            foreach ($listItem->content as $item) {
                if ('paragraph' === $item->type) {
                    $phpParagraph = $phpShape->createParagraph();
                    $phpParagraph->setRichTextElements(array());
                    $phpParagraph->getBulletStyle()->setBulletType($bulletType);
                    $phpParagraph->getAlignment()->setLevel($level);
                    $this->loadParagraph($item, $phpParagraph);
                    if (method_exists($phpShape, 'isPlaceholder') && !$phpShape->isPlaceholder()) {
                        $phpParagraph->getAlignment()->setMarginLeft(self::BASE_MARGIN_LEFT_VALUE + $level * self::MARGIN_LEFT_INCREMENT);
                        $phpParagraph->getAlignment()->setIndent(self::BASE_INDENT_VALUE);
                    }
                } else {
                    $this->loadList($item, $phpShape, $level + 1);
                }
            }
        }
    }

    private function loadMarks($marks, $phpFontStyle)
    {
        if (!is_array($marks) && !is_object($marks)) {
            return;
        }

        foreach ($marks as $mark) {
            if ('em' === $mark->_) {
                $phpFontStyle->setItalic($this->getMarkState($mark));
            } else if ('strong' === $mark->_) {
                $phpFontStyle->setBold($this->getMarkState($mark));
            } else if ('underline' === $mark->_) {
                $underline = Font::STATE_ON === $this->getMarkState($mark) ? Font::UNDERLINE_SINGLE : Font::UNDERLINE_NONE;
                $phpFontStyle->setUnderline($underline);
            } else if ('strike' === $mark->_) {
                $phpFontStyle->setStrikethrough($this->getMarkState($mark));
            } else if ('color' === $mark->_ && '#' === substr($mark->value, 0, 1)) {
                $color = new Color();
                $color->setRGB(substr($mark->value, 1));
                $phpFontStyle->setColor($color);
            } else if ('background_color' === $mark->_ && '#' === substr($mark->value, 0, 1)) {
                $color = new Color();
                $color->setRGB(substr($mark->value, 1));
                $phpFontStyle->setHighlight($color);
            } else if ('superscript' === $mark->_) {
                $phpFontStyle->setSuperScript($this->getMarkState($mark));
            } else if ('subscript' === $mark->_) {
                $phpFontStyle->setSubScript($this->getMarkState($mark));
            } else if ('font_size' === $mark->_) {
                $phpFontStyle->setSize(floatval($mark->value));
            } else if ('font_family' === $mark->_) {
                $phpFontStyle->setName($mark->value);
            } else if ('discussion' === $mark->_) {
                foreach ($mark->id as $id) {
                    $this->commentIds[$id] = true;
                }
            }
        }
    }

    private function getMarkState($mark)
    {
        if (!property_exists($mark, 'is_on')) {
            return Font::STATE_ON;
        }
        return $mark->is_on ? Font::STATE_ON : Font::STATE_OFF;
    }

    private function loadBodyStyle($order, $phpShape, $phpSlide)
    {
        global $i18n;
        $level = intval(substr($order->type, 6)) - 1;
        $paragraph = $order->content[0]->content[0];

        if ($phpSlide instanceof SlideLayout) {
            $phpParagraph = $phpShape->createParagraph();
        } else {
            $phpParagraph = new Paragraph();
        }
        $this->loadParagraph($paragraph, $phpParagraph, false);
        $phpParagraph->getBulletStyle()->setBulletType(Bullet::TYPE_NUMERIC);
        if ($phpSlide instanceof SlideMaster) {
            $phpSlide->getTextStyles()->setBodyStyleAtLvl($phpParagraph, $level);
        }

        if (!($phpSlide instanceof SlideLayout)) {
            $phpParagraph = $phpShape->createParagraph();
        }
        $phpParagraph->setRichTextElements(array());
        $phpParagraph->getAlignment()->setLevel($level);
        $text = $i18n["level_" . ($level + 1)];
        $phpParagraph->createTextRun()->setText($text);

        if (2 > count($order->content)) {
            return;
        }
        $this->loadBodyStyle($order->content[1]->content[0], $phpShape, $phpSlide);
    }

    private function loadSynolink($box, $phpObject)
    {
        if (!property_exists($box->attrs, 'synolink')) {
            return;
        }
        $phpHyperlink = $phpObject->getHyperlink();
        $synolink = $box->attrs->synolink;
        if ('#' === substr($synolink, 0, 1)) {
            $slideId = substr($synolink, 1);
            $slideNum = $this->slideIdIdxMap[$slideId] + 1;
            $phpHyperlink->setUrl("slide$slideNum.xml");
            $phpHyperlink->setAction('ppaction://hlinksldjump');
        } else {
            $phpHyperlink->setUrl($synolink);
        }
    }

    private function loadTextLink($marks, $textRun)
    {
        if (!is_array($marks) && !is_object($marks)) {
            return;
        }

        foreach ($marks as $mark) {
            if ('link' !== $mark->_) {
                continue;
            }
            $phpHyperlink = $textRun->getHyperlink();
            $href = $mark->href;
            if ('#' === substr($href, 0, 1)) {
                $slideId = substr($href, 1);
                if (array_key_exists($slideId, $this->slideIdIdxMap)) {
                    $slideNum = $this->slideIdIdxMap[$slideId] + 1;
                    $phpHyperlink->setUrl("slide$slideNum.xml");
                    $phpHyperlink->setAction('ppaction://hlinksldjump');
                }
            } else {
                $phpHyperlink->setUrl($href);
            }
            break;
        }
    }

    private function loadPaddings($block, $phpShape)
    {
        if (!property_exists($block, 'attrs') || !property_exists($block->attrs, 'padding')) {
            return;
        }
        $phpShape->setInsetTop($block->attrs->padding->top);
        $phpShape->setInsetRight($block->attrs->padding->right);
        $phpShape->setInsetBottom($block->attrs->padding->bottom);
        $phpShape->setInsetLeft($block->attrs->padding->left);
    }

    private function loadCellStyle($cell, $phpCell)
    {
        $borders = array('Top', 'Right', 'Bottom', 'Left');
        $styles = array('Color', 'Style', 'Width');

        $borderTop = $phpCell->getBorders()->getTop();
        $color = $this->parseColor($cell->attrs->borderTopColor);
        if ('0px' === $cell->attrs->borderTopWidth || 'transparent' === $color) {
            $borderTop->setLineStyle(Border::LINE_NONE);
        } else {
            $borderTop->setLineStyle(Border::LINE_SINGLE);
        }
        if ('transparent' !== $color) {
            $borderTop->setColor(new Color('FF' . $color));
        }
        $borderTop->setDashStyle($cell->attrs->borderTopStyle);

        $borderRight = $phpCell->getBorders()->getRight();
        $color = $this->parseColor($cell->attrs->borderRightColor);
        if ('0px' === $cell->attrs->borderRightWidth || 'transparent' === $color) {
            $borderRight->setLineStyle(Border::LINE_NONE);
        } else {
            $borderRight->setLineStyle(Border::LINE_SINGLE);
        }
        if ('transparent' !== $color) {
            $borderRight->setColor(new Color('FF' . $color));
        }
        $borderRight->setDashStyle($cell->attrs->borderRightStyle);

        $borderBottom = $phpCell->getBorders()->getBottom();
        $color = $this->parseColor($cell->attrs->borderBottomColor);
        if ('0px' === $cell->attrs->borderBottomWidth || 'transparent' === $color) {
            $borderBottom->setLineStyle(Border::LINE_NONE);
        } else {
            $borderBottom->setLineStyle(Border::LINE_SINGLE);
        }
        if ('transparent' !== $color) {
            $borderBottom->setColor(new Color('FF' . $color));
        }
        $borderBottom->setDashStyle($cell->attrs->borderBottomStyle);

        $borderLeft = $phpCell->getBorders()->getLeft();
        $color = $this->parseColor($cell->attrs->borderLeftColor);
        if ('0px' === $cell->attrs->borderLeftWidth || 'transparent' === $color) {
            $borderLeft->setLineStyle(Border::LINE_NONE);
        } else {
            $borderLeft->setLineStyle(Border::LINE_SINGLE);
        }
        if ('transparent' !== $color) {
            $borderLeft->setColor(new Color('FF' . $color));
        }
        $borderLeft->setDashStyle($cell->attrs->borderLeftStyle);

        $fill = new Fill();
        if (property_exists($cell->attrs, 'backgroundColor')) {
            $strColor = $this->parseColor($cell->attrs->backgroundColor);
            if (!empty($strColor) && 'transparent' !== $strColor) {
                $fill->setFillType(Fill::FILL_SOLID);
                $color = new Color('FF' . $strColor);
                $fill->setStartColor($color);
                $fill->setEndColor($color);
            } else {
                $fill->setFillType(Fill::FILL_NONE);
            }
        }
        $phpCell->setFill($fill);

        if (property_exists($cell->attrs, 'verticalAlign')) {
            switch ($cell->attrs->verticalAlign) {
                case 'top':
                    $phpCell->getAlignment()->setVertical(Alignment::VERTICAL_TOP);
                    break;
                case 'bottom':
                    $phpCell->getAlignment()->setVertical(Alignment::VERTICAL_BOTTOM);
                    break;
                default:
                    $phpCell->getAlignment()->setVertical(Alignment::VERTICAL_CENTER);
                    break;
            }
        }
    }

    private function parseColor($color)
    {
        $color = strtolower($color);
        if ('rgb(' === substr($color, 0, 4)) {
            $hexColor = '';
            $rgb = explode(',', substr($color, 4, strlen($color) - 5));
            foreach ($rgb as $value) {
                $value = intval($value);
                $hexColor .= str_pad(dechex($value), 2, '0', STR_PAD_LEFT);
            }
            return $hexColor;
        } else if ('#' === substr($color, 0, 1)) {
            return substr($color, 1);
        }
        return $color;
    }

    private function setOtherStyles($phpMaster)
    {
        $baseMargin = 48;
        for ($i = 0; $i < 9; $i++) {
            $phpParagraph = new Paragraph();

            $phpParagraph->getFont()->setSize(11);
            $phpParagraph->getAlignment()->setHorizontal(Alignment::HORIZONTAL_LEFT);
            $phpParagraph->getAlignment()->setMarginLeft($i * $baseMargin);

            $phpMaster->getTextStyles()->setOtherStyleAtLvl($phpParagraph, $i);
        }
    }
}

?>
