<?php

namespace PhpOffice\PhpPresentation\Shape\Drawing;

use ZipArchive;

class Gd extends AbstractDrawingAdapter
{
    /* Rendering functions */
    const RENDERING_DEFAULT = 'imagepng';
    const RENDERING_PNG = 'imagepng';
    const RENDERING_GIF = 'imagegif';
    const RENDERING_JPEG = 'imagejpeg';

    /* MIME types */
    const MIMETYPE_DEFAULT = 'image/png';
    const MIMETYPE_PNG = 'image/png';
    const MIMETYPE_GIF = 'image/gif';
    const MIMETYPE_JPEG = 'image/jpeg';

    /**
     * Image resource
     *
     * @var resource
     */
    protected $imageResource;

    /**
     * Rendering function
     *
     * @var string
     */
    protected $renderingFunction = self::RENDERING_DEFAULT;

    /**
     * Mime type
     *
     * @var string
     */
    protected $mimeType = self::MIMETYPE_DEFAULT;

    /**
     * Unique name
     *
     * @var string
     */
    protected $uniqueName;

    /**
     * Source
     *
     * @var string
     */
    protected $source;

    protected $fileId;

    protected $imageCreateFunc;
    protected $imageFunc;
    protected $imageExtension;
    protected $alpha = 100;
    protected $cropTop = 0;
    protected $cropRight = 0;
    protected $cropBottom = 0;
    protected $cropLeft = 0;
    protected $extension;

    /**
     * Gd constructor
     */
    public function __construct()
    {
        parent::__construct();
        $this->uniqueName = md5(rand(0, 9999) . time() . rand(0, 9999));
    }

    /**
     * Get image resource
     *
     * @return resource
     */
    public function getImageResource()
    {
        return $this->imageResource;
    }

    /**
     * Set image resource
     *
     * @param $value resource
     * @return $this
     */
    public function setImageResource($value = null)
    {
        $this->imageResource = $value;

        if (!is_null($this->imageResource)) {
            // Get width/height
            $this->width  = imagesx($this->imageResource);
            $this->height = imagesy($this->imageResource);
        }

        return $this;
    }

    /**
     * Get rendering function
     *
     * @return string
     */
    public function getRenderingFunction()
    {
        return $this->renderingFunction;
    }

    /**
     * Set rendering function
     *
     * @param  string                            $value
     * @return $this
     */
    public function setRenderingFunction($value = self::RENDERING_DEFAULT)
    {
        $this->renderingFunction = $value;
        return $this;
    }

    /**
     * Get mime type
     *
     * @return string
     */
    public function getMimeType()
    {
        return $this->mimeType;
    }

    /**
     * Set mime type
     *
     * @param  string $value
     * @return $this
     */
    public function setMimeType($value = self::MIMETYPE_DEFAULT)
    {
        $this->mimeType = $value;
        return $this;
    }

    /**
     * @return string
     */
    public function getContents()
    {
        ob_start();
        call_user_func($this->getRenderingFunction(), $this->getImageResource());
        $imageContents = ob_get_contents();
        ob_end_clean();
        return $imageContents;
    }

    /**
     * @return string
     */
    public function getExtension()
    {
        if (is_null($this->extension)) {
            $extension = strtolower($this->getMimeType());
            $extension = explode('/', $extension);
            $this->extension = $extension[1];
        }
        return $this->extension;
    }

    public function setExtension($value)
    {
        if (isset($value) && '' != $value) {
            $this->extension = $value;
        }
        return $this;
    }

    /**
     * @return string
     */
    public function getIndexedFilename()
    {
        return $this->fileId . '.' . $this->getExtension();
    }

    /**
     * Set image functions and extensions.
     *
     * @return void
     */
    private function setFunctions()
    {
        switch ($this->mimeType) {
            case 'image/png':
                $this->imageCreateFunc = 'imagecreatefrompng';
                $this->imageFunc = 'imagepng';
                $this->imageExtension = 'png';
                break;
            case 'image/gif':
                $this->imageCreateFunc = 'imagecreatefromgif';
                $this->imageFunc = 'imagegif';
                $this->imageExtension = 'gif';
                break;
            case 'image/jpeg':
            case 'image/jpg':
                $this->imageCreateFunc = 'imagecreatefromjpeg';
                $this->imageFunc = 'imagejpeg';
                $this->imageExtension = 'jpg';
                break;
            case 'image/bmp':
            case 'image/x-ms-bmp':
                $this->mimeType = 'image/bmp';
                $this->imageExtension = 'bmp';
                break;
            case 'image/tiff':
                $this->imageExtension = 'tif';
                break;
        }
    }

    /**
     * Get image string data
     *
     * @return string|null
     * @since 0.11.0
     */
    public function getImageStringData()
    {
        global $synoTempPath;
        $source = $this->source;
        $actualSource = null;
        $imageBinary = null;
        $imageData = null;

        if ('zip://' === substr($this->source, 0, 6)) {
            $source = substr($source, 6);
            list($zipFilename, $imageFilename) = explode('#', $source);

            $tmpPath;
            if (isset($synoTempPath)) {
                $tmpPath = $synoTempPath;
            } else {
                $tmpPath = sys_get_temp_dir();
            }

            $zip = new ZipArchive();
            if ($zip->open($zipFilename) !== false) {
                if ($zip->locateName($imageFilename) !== false) {
                    $imageContent = $zip->getFromName($imageFilename);
                    //$zip->extractTo($tmpPath, $imageFilename);
                    //$actualSource = $tmpPath . DIRECTORY_SEPARATOR . $imageFilename;
                }
            }
            $zip->close();
        } else {
            if (isset($this->source)) {
                $imageContent = file_get_contents($this->source);
            }
        }

        return $imageContent;
    }

    public function getSource()
    {
        return $this->source;
    }

    public function setSource($source = '')
    {
        $this->source = $source;
    }

    public function setFileId($fileId = '')
    {
        $this->fileId = $fileId;
    }

    public function getFileId()
    {
        return $this->fileId;
    }

    public function checkImage()
    {
        global $synoTempPath;
        if (empty($this->source)) {
            return;
        }

        $removeImage = false;
        if ('zip://' === substr($this->source, 0, 6)) {
            $imagePath = $this->extractImage($this->source);
            if ('' === $imagePath) {
                return;
            }
            $removeImage = true;
        } else {
            $imagePath = $this->source;
        }

        $imageData = getimagesize($imagePath);
        if ($removeImage) {
            unlink($imagePath);
        }

        if (!is_array($imageData)) {
            throw new \Exception('Invalid image');
        }

        list($actualWidth, $actualHeight, $imageType) = $imageData;
        $supportedTypes = array(IMAGETYPE_JPEG, IMAGETYPE_GIF, IMAGETYPE_PNG);
        if (!in_array($imageType, $supportedTypes)) {
            throw new \Exception('Unsupported image type');
        }

        $this->mimeType = image_type_to_mime_type($imageType);
        $this->setFunctions();
    }

    private function extractImage($source)
    {
        $source = substr($this->source, 6);
        list($zipFilename, $imageFilename) = explode('#', $source);
        
        if (isset($synoTempPath)) {
            $tempFileName = tempnam($synoTempPath, 'PHPPresentationImage');
        } else {
            $tempFileName = tempnam(sys_get_temp_dir(), 'PHPPresentationImage');
        }
        if (false === $tempFileName) {
            throw new \Exception('Fail to create temp image file');
        }

        $zip = new ZipArchive();
        if (false === $zip->open($zipFilename)) {
            throw new \Exception("Fail to open archive file");
        }

        $imageContent = $zip->getFromName($imageFilename);
        $zip->close();
        if (empty($imageContent)) {
            return '';
        }

        file_put_contents($tempFileName, $imageContent);

        return $tempFileName;
    }

    public function getAlpha()
    {
        return $this->alpha;
    }

    public function setAlpha($pValue = 100)
    {
        $this->alpha = $pValue;
    }

    public function setCrop($cropTop = 0, $cropRight = 0, $cropBottom = 0, $cropLeft = 0)
    {
        $this->cropTop = $cropTop;
        $this->cropRight = $cropRight;
        $this->cropBottom = $cropBottom;
        $this->cropLeft = $cropLeft;
    }

    public function getCrop()
    {
        return Array($this->cropTop, $this->cropRight, $this->cropBottom, $this->cropLeft);
    }
}
