<?php
/**
 * This file is part of PHPWord - A pure PHP library for reading and writing
 * word processing documents.
 *
 * PHPWord is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPWord/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPWord
 * @copyright   2010-2014 PHPWord contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpWord\Writer\Word2007\Element;

use PhpOffice\PhpWord\Element\Image as ImageElement;
use PhpOffice\PhpWord\Shared\XMLWriter;
use PhpOffice\PhpWord\Writer\Word2007\Style\Image as ImageStyleWriter;

/**
 * Image element writer
 *
 * @since 0.10.0
 */
class Image extends Inline
{
    /**
     * Write element.
     *
     * @return void
     */
    public function write()
    {
        $xmlWriter = $this->getXmlWriter();
        $element = $this->getElement();
        if (!$element instanceof ImageElement) {
            return;
        }

        if ($element->isWatermark()) {
            $this->writeWatermark($xmlWriter, $element);
        } else {
            $config = $element->getConfig();
            if (empty($config)) {
                $this->writeImage($xmlWriter, $element);
            } else {
                $this->writeDrawingImage($xmlWriter, $element);
            }
        }
    }

    /**
     * Write image element.
     *
     * @return void
     */
    private function writeImage(XMLWriter $xmlWriter, ImageElement $element)
    {
        $rId = $element->getRelationId() + ($element->isInSection() ? 9 : 0);
        $style = $element->getStyle();
        $styleWriter = new ImageStyleWriter($xmlWriter, $style);
        $isBroken = $element->getBroken();

        if (!$this->withoutP) {
            $xmlWriter->startElement('w:p');
            $styleWriter->writeAlignment();
        }

        $this->writeCommentRangeStart($xmlWriter, $element);
        $this->writeRevisionOpenTag($xmlWriter, $element);

        $xmlWriter->startElement('w:r');
        $xmlWriter->startElement('w:pict');
        $xmlWriter->startElement('v:shape');
        $xmlWriter->writeAttribute('type', '#_x0000_t75');

        if ($isBroken) {
            foreach (array('left', 'top', 'right', 'bottom') as $border) {
                $xmlWriter->writeAttribute("o:border{$border}color", '#d8dadc');
            }
        }

        $styleWriter->write();

        $xmlWriter->startElement('v:imagedata');
        $xmlWriter->writeAttribute('r:id', 'rId' . $rId);
        $xmlWriter->writeAttribute('o:title', '');

        if ($isBroken) {
            $hcrop = (($style->getWidth() - 24) / 2) * 65536 / 24;
            foreach (array('left', 'right') as $crop) {
                $xmlWriter->writeAttribute("crop{$crop}", "-{$hcrop}f");
            }
            $vcrop = (($style->getHeight() - 24) / 2) * 65536 / 24;
            foreach (array('top', 'bottom') as $crop) {
                $xmlWriter->writeAttribute("crop{$crop}", "-{$hcrop}f");
            }
        }

        $xmlWriter->endElement(); // v:imagedata

        if ($isBroken) {
            foreach (array('left', 'top', 'right', 'bottom') as $border) {
                $xmlWriter->startElement("w10:border{$border}");
                $xmlWriter->writeAttribute('type', 'single');
                $xmlWriter->writeAttribute('width', '4');
                $xmlWriter->endElement(); // w10:border{$border}
            }
        }

        $xmlWriter->endElement(); // v:shape
        $xmlWriter->endElement(); // w:pict
        $xmlWriter->endElement(); // w:r

        $this->writeRevisionCloseTag($xmlWriter, $element);
        $this->writeCommentRangeEnd($xmlWriter, $element);

        $this->endElementP();
    }

    /**
     * Write watermark element.
     *
     * @return void
     */
    private function writeWatermark(XMLWriter $xmlWriter, ImageElement $element)
    {
        $rId = $element->getRelationId();
        $style = $element->getStyle();
        $style->setPositioning('absolute');
        $styleWriter = new ImageStyleWriter($xmlWriter, $style);

        $xmlWriter->startElement('w:p');
        $xmlWriter->startElement('w:r');
        $xmlWriter->startElement('w:pict');
        $xmlWriter->startElement('v:shape');
        $xmlWriter->writeAttribute('type', '#_x0000_t75');

        $styleWriter->write();

        $xmlWriter->startElement('v:imagedata');
        $xmlWriter->writeAttribute('r:id', 'rId' . $rId);
        $xmlWriter->writeAttribute('o:title', '');
        $xmlWriter->endElement(); // v:imagedata
        $xmlWriter->endElement(); // v:shape
        $xmlWriter->endElement(); // w:pict
        $xmlWriter->endElement(); // w:r
        $xmlWriter->endElement(); // w:p
    }

    private function writeDrawingImage(XMLWriter $xmlWriter, ImageElement $element)
    {
        $id = $element->getId();
        $config = $element->getConfig();
        $width = intval($config['width'] * 9525);
        $height = intval($config['height'] * 9525);

        $this->writeCommentRangeStart($xmlWriter, $element);
        $this->writeRevisionOpenTag($xmlWriter, $element);

        $xmlWriter->startElement('w:r');
        $xmlWriter->startElement('w:drawing');
        switch ($config['alignment']) {
            case 'break':
                $xmlWriter->startElement('wp:anchor');
                $xmlWriter->writeAttribute('distT', 0);
                $xmlWriter->writeAttribute('distB', 0);
                $xmlWriter->writeAttribute('distL', 114300);
                $xmlWriter->writeAttribute('distR', 114300);
                $xmlWriter->writeAttribute('simplePos', 0);
                $xmlWriter->writeAttribute('relativeHeight', intval($config['order']) + 251658240);
                $xmlWriter->writeAttribute('behindDoc', 0);
                $xmlWriter->writeAttribute('locked', 0);
                $xmlWriter->writeAttribute('layoutInCell', 1);
                $xmlWriter->writeAttribute('allowOverlap', 1);
                $xmlWriter->writeElementBlock('wp:simplePos', array('x' => 0, 'y' => 0));
                $xmlWriter->startElement('wp:positionH');
                $xmlWriter->writeAttribute('relativeFrom', 'page');
                $xmlWriter->startElement('wp:posOffset');
                $xmlWriter->text(intval($config['left'] * 9525));
                $xmlWriter->endElement(); // wp:posOffset
                $xmlWriter->endElement(); // wp:positionH
                $xmlWriter->startElement('wp:positionV');
                $xmlWriter->writeAttribute('relativeFrom', 'page');
                $xmlWriter->startElement('wp:posOffset');
                $xmlWriter->text(intval($config['top'] * 9525));
                $xmlWriter->endElement(); // wp:posOffset
                $xmlWriter->endElement(); // wp:positionV
                // This must be put after <wp:extent>
                // What the fuck Microsoft Word
                // $xmlWriter->writeElementBlock('wp:wrapTopAndBottom', array());
                break;
            case 'inline':
            default:
                $xmlWriter->startElement('wp:inline');
                break;
        }
        $xmlWriter->writeElementBlock('wp:extent', array('cx' => $width, 'cy' => $height));
        switch ($config['alignment']) {
            case 'break':
                $xmlWriter->writeElementBlock('wp:wrapTopAndBottom', array());
                break;
            case 'inline':
            default:
                break;
        }
        $name = empty($config['file_alt'])? 'image': $config['file_alt'];
        $xmlWriter->writeElementBlock('wp:docPr', array(
            'id' => $id,
            'name' => $name
        ));
        $this->writeGraphicPicture($xmlWriter, $config);
        $xmlWriter->endElement(); // wp:inline or wp:anchor
        $xmlWriter->endElement(); // w:drawing
        $xmlWriter->endElement(); // w:r

        $this->writeRevisionCloseTag($xmlWriter, $element);
        $this->writeCommentRangeEnd($xmlWriter, $element);
    }

    private function writeGraphicPicture($xmlWriter, $config) {
        $xmlWriter->startElement('a:graphic');
        $xmlWriter->writeAttribute('xmlns:a', 'http://schemas.openxmlformats.org/drawingml/2006/main');
        $xmlWriter->startElement('a:graphicData');
        $xmlWriter->writeAttribute('uri', 'http://schemas.openxmlformats.org/drawingml/2006/picture');
        $xmlWriter->startElement('pic:pic');
        $xmlWriter->writeAttribute('xmlns:pic', 'http://schemas.openxmlformats.org/drawingml/2006/picture');
        $this->writePicNvpicpr($xmlWriter, $config);
        $this->writePicBlipfill($xmlWriter, $config);
        $this->writePicSppr($xmlWriter, $config);
        $xmlWriter->endElement(); // pic:pic
        $xmlWriter->endElement(); // a:graphicData
        $xmlWriter->endElement(); // a:graphic
    }

    private function writePicNvpicpr($xmlWriter, $config) {
        $element = $this->getElement();
        $id = $element->getId();
        $name = empty($config['file_alt'])? 'image': $config['file_alt'];
        $xmlWriter->startElement('pic:nvPicPr');
        $xmlWriter->writeElementBlock('pic:cNvPr', array(
            'id' => $id,
            'name' => $name
        ));
        $xmlWriter->writeElementBlock('pic:cNvPicPr');
        $xmlWriter->endElement(); // pic:nvPicPr
    }

    private function writePicBlipfill($xmlWriter, $config) {
        // prepare cropping info
        $srcRectArr = array();
        if (!empty($config['crop'])) {
            $cropData = $config['crop'];
            if (!empty($cropData['l'])) {
                $srcRectArr['l'] = intval($cropData['l']);
            }
            if (!empty($cropData['r'])) {
                $srcRectArr['r'] = intval($cropData['r']);
            }
            if (!empty($cropData['t'])) {
                $srcRectArr['t'] = intval($cropData['t']);
            }
            if (!empty($cropData['b'])) {
                $srcRectArr['b'] = intval($cropData['b']);
            }
        }
        $element = $this->getElement();
        $rId = $element->getRelationId() + ($element->isInSection() ? 9 : 0);

        $xmlWriter->startElement('pic:blipFill');
        $xmlWriter->startElement('a:blip');
        $xmlWriter->writeAttribute('r:embed', 'rId'.$rId);
        $xmlWriter->startElement('a:extLst');
        $xmlWriter->startElement('a:ext');
        $xmlWriter->writeAttribute('uri', '{28A0092B-C50C-407E-A947-70E740481C1C}');
        $xmlWriter->writeElementBlock('a14:useLocalDpi', array(
            'xmlns:a14' => 'http://schemas.microsoft.com/office/drawing/2010/main',
            'val' => '0'
        ));
        $xmlWriter->endElement(); // a:ext
        $xmlWriter->endElement(); // a:extLst
        $xmlWriter->endElement(); // a:blip

        $xmlWriter->writeElementBlock('a:srcRect', $srcRectArr);
        $xmlWriter->writeElementBlock('a:stretch', array());
        $xmlWriter->endElement(); // pic:blipFill
    }

    private function writePicSppr($xmlWriter, $config) {
        $width = intval($config['width'] * 9525);
        $height = intval($config['height'] * 9525);
        $xmlWriter->startElement('pic:spPr');
        $xmlWriter->startElement('a:xfrm');
        if ($config['flip_h']) {
            $xmlWriter->writeAttribute('flipH', 1);
        }
        if ($config['flip_v']) {
            $xmlWriter->writeAttribute('flipV', 1);
        }
        if ($config['deg'] !== 0) { // TODO take care about floating point
            $xmlWriter->writeAttribute('rot', intval($config['deg'] * 60000));
        }
        $xmlWriter->writeElementBlock('a:off', array('x' => 0, 'y' => 0)); // TODO
        $xmlWriter->writeElementBlock('a:ext', array('cx' => $width, 'cy' => $height));
        $xmlWriter->endElement(); // a:xfrm

        $xmlWriter->startElement('a:prstGeom');
        $xmlWriter->writeAttribute('prst', 'rect');
        $xmlWriter->writeElementBlock('a:srcRect', array());
        $xmlWriter->endElement(); // a:prstGeom

        // image frame
        $strokeWidth = empty($config['stroke-width'])? '0': $config['stroke-width'] * 9525;
        $xmlWriter->startElement('a:ln');
        if (empty($config['stroke-width'])) {
            $xmlWriter->writeElementBlock('a:noFill', array());
        } else {
            $xmlWriter->writeAttribute('w', $strokeWidth);
            $strokeColor = empty($config['stroke'])? '000000': substr($config['stroke'], 1);
            $xmlWriter->startElement('a:solidFill');
            $xmlWriter->writeElementBlock('a:srgbClr', array(
                'val' => $strokeColor
            ));
            $xmlWriter->endElement(); // a:solidFill
        }
        $xmlWriter->endElement(); // a:ln

        $xmlWriter->startElement('a:extLst');
        $xmlWriter->startElement('a:ext');
        $xmlWriter->writeAttribute('uri', '{53640926-AAD7-44D8-BBD7-CCE9431645EC}');
        $xmlWriter->writeElementBlock('a14:shadowObscured', array(
             'xmlns:a14' => 'http://schemas.microsoft.com/office/drawing/2010/main'
        ));
        $xmlWriter->endElement(); // pic:spPr
        $xmlWriter->endElement(); // a:ext
        $xmlWriter->endElement(); // a:extLst
    }
}
