<?php
/**
 * This file is part of PHPWord - A pure PHP library for reading and writing
 * word processing documents.
 *
 * PHPWord is free software distributed under the terms of the GNU Lesser
 * General Public License version 3 as published by the Free Software Foundation.
 *
 * For the full copyright and license information, please read the LICENSE
 * file that was distributed with this source code. For the full list of
 * contributors, visit https://github.com/PHPOffice/PHPWord/contributors.
 *
 * @link        https://github.com/PHPOffice/PHPWord
 * @copyright   2010-2014 PHPWord contributors
 * @license     http://www.gnu.org/licenses/lgpl.txt LGPL version 3
 */

namespace PhpOffice\PhpWord\Writer\Word2007\Part;

use PhpOffice\PhpWord\Element\Chart as ChartElement;
use PhpOffice\PhpWord\Shared\XMLWriter;

/**
 * Word2007 chart part writer: word/charts/chartx.xml
 *
 * @since 0.12.0
 * @link http://www.datypic.com/sc/ooxml/e-draw-chart_chartSpace.html
 */
class Chart extends AbstractPart
{
    /**
     * Chart element
     *
     * @var \PhpOffice\PhpWord\Element\Chart $element
     */
    private $element;

    /**
     * Type definition
     *
     * @var array
     */
    private $types = array(
        'pie'       => array('type' => 'pie', 'colors' => 1),
        'doughnut'  => array('type' => 'doughnut', 'colors' => 1, 'hole' => 75, 'no3d' => true),
        'bar'       => array('type' => 'bar', 'colors' => 0, 'axes' => true, 'bar' => 'bar', 'grouping' => true),
        'column'    => array('type' => 'bar', 'colors' => 0, 'axes' => true, 'bar' => 'col', 'grouping' => true),
        'line'      => array('type' => 'line', 'colors' => 0, 'axes' => true, 'smooth' => 0, 'grouping' => true),
        'area'      => array('type' => 'area', 'colors' => 0, 'axes' => true, 'grouping' => true),
        'radar'     => array('type' => 'radar', 'colors' => 0, 'axes' => true, 'radar' => 'standard', 'no3d' => true),
        'scatter'   => array('type' => 'scatter', 'colors' => 0, 'axes' => true, 'scatter' => 'marker', 'no3d' => true),
    );

    /**
     * Chart options
     *
     * @var array
     */
    private $options = array();

    /**
     * Set chart element.
     *
     * @param \PhpOffice\PhpWord\Element\Chart $element
     * @return void
     */
    public function setElement(ChartElement $element)
    {
        $this->element = $element;
    }

    /**
     * Write part
     *
     * @return string
     */
    public function write()
    {
        $xmlWriter = $this->getXmlWriter();

        $xmlWriter->startDocument('1.0', 'UTF-8', 'yes');
        $xmlWriter->startElement('c:chartSpace');
        $xmlWriter->writeAttribute('xmlns:c', 'http://schemas.openxmlformats.org/drawingml/2006/chart');
        $xmlWriter->writeAttribute('xmlns:a', 'http://schemas.openxmlformats.org/drawingml/2006/main');
        $xmlWriter->writeAttribute('xmlns:r', 'http://schemas.openxmlformats.org/officeDocument/2006/relationships');

        $this->writeChart($xmlWriter);
        $this->writeShape($xmlWriter);

        $xmlWriter->endElement(); // c:chartSpace

        return $xmlWriter->getData();
    }

    /**
     * Write chart
     *
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_Chart.html
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @return void
     */
    private function writeChart(XMLWriter $xmlWriter)
    {
        $xmlWriter->startElement('c:chart');

        if (!empty($this->element->getTitle())) {
            $this->writeTitleBlock($xmlWriter, $this->element->getTitle());
        }

        $xmlWriter->writeElementBlock('c:autoTitleDeleted', 'val', 1);

        $synochart_config = $this->element->getSynoChartConfig();
        if (!empty($synochart_config)) {
            $this->writeSynoPlotArea($xmlWriter, $synochart_config);
        } else {
            $this->writePlotArea($xmlWriter);
        }

        $this->writeLegend($xmlWriter);

        $xmlWriter->endElement(); // c:chart
    }

    /**
     * Write plot area.
     *
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_PlotArea.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_PieChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_DoughnutChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_BarChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_LineChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_AreaChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_RadarChart.html
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_ScatterChart.html
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @return void
     */
    private function writePlotArea(XMLWriter $xmlWriter)
    {
        $type = $this->element->getType();
        $style = $this->element->getStyle();
        $this->options = $this->types[$type];

        $xmlWriter->startElement('c:plotArea');
        $xmlWriter->writeElement('c:layout');

        // Chart
        $chartType = $this->options['type'];
        $chartType .= $style->is3d() && !isset($this->options['no3d'])? '3D' : '';
        $chartType .= 'Chart';
        $xmlWriter->startElement("c:{$chartType}");

        $xmlWriter->writeElementBlock('c:varyColors', 'val', $this->options['colors']);
        if ($type == 'area') {
            $xmlWriter->writeElementBlock('c:grouping', 'val', 'standard');
        }
        if (isset($this->options['hole'])) {
            $xmlWriter->writeElementBlock('c:holeSize', 'val', $this->options['hole']);
        }
        if (isset($this->options['bar'])) {
            $xmlWriter->writeElementBlock('c:barDir', 'val', $this->options['bar']); // bar|col
            $xmlWriter->writeElementBlock('c:grouping', 'val', 'clustered'); // 3d; standard = percentStacked
        }
        if (isset($this->options['radar'])) {
            $xmlWriter->writeElementBlock('c:radarStyle', 'val', $this->options['radar']);
        }
        if (isset($this->options['scatter'])) {
            $xmlWriter->writeElementBlock('c:scatterStyle', 'val', $this->options['scatter']);
        }

        // Series
        $this->writeSeries($xmlWriter, isset($this->options['scatter']));

        // Axes
        if (isset($this->options['axes'])) {
            $xmlWriter->writeElementBlock('c:axId', 'val', 1);
            $xmlWriter->writeElementBlock('c:axId', 'val', 2);
        }

        $xmlWriter->endElement(); // chart type

        // Axes
        if (isset($this->options['axes'])) {
            $this->writeAxis($xmlWriter, 'cat');
            $this->writeAxis($xmlWriter, 'val');
        }

        $xmlWriter->endElement(); // c:plotArea
    }

    private function writeSynoPlotArea(XMLWriter $xmlWriter, $config)
    {
        $xmlWriter->startElement('c:plotArea');
        $xmlWriter->writeElement('c:layout');

        $subCharts = array();
        foreach ($config['plotArea']['series'] as $series) {
            $type = $series['type'];
            if (!isset($subCharts[$type])) {
                $subCharts[$type] = array();
            }
            $subCharts[$type][] = $series;
        }

        $writeAxes = false;
        foreach ($subCharts as $type => $item) {
            $this->options = $this->types[$type];

            // Chart
            $chartType = $type;
            $chartType .= 'Chart';
            $xmlWriter->startElement("c:{$chartType}");

            $xmlWriter->writeElementBlock('c:varyColors', 'val', $this->options['colors']);
            if ($this->options['grouping'] && !empty($config['grouping'])) {
                $xmlWriter->writeElementBlock('c:grouping', 'val', $config['grouping']);
                if ($config['grouping'] === 'stacked') {
                    $xmlWriter->writeElementBlock('c:overlap', 'val', 100);
                }
            }
            if (isset($this->options['bar']) && !empty($config['barDir'])) {
                $xmlWriter->writeElementBlock('c:barDir', 'val', $config['barDir'] === 'row' ? 'bar' : 'col');
            }

            // Series
            $this->writeComboSeries($xmlWriter, $type === 'scatter', $item);

            // Axes
            if (isset($this->options['axes'])) {
                $writeAxes = true;
                $xmlWriter->writeElementBlock('c:axId', 'val', 1);
                $xmlWriter->writeElementBlock('c:axId', 'val', 2);
            }

            $xmlWriter->endElement(); // chart type
        }

        if ($writeAxes) {
            $this->writeAxis($xmlWriter, 'cat');
            $this->writeAxis($xmlWriter, 'val');
        }

        $xmlWriter->endElement(); // c:plotArea
    }

    private function writeComboSeries(XMLWriter $xmlWriter, $scatter = false, $series)
    {
        static $index = 0;
        foreach ($series as $seriesItem) {
            $categories = $seriesItem['cat'];
            $values = $seriesItem['val'];
            $tx = $seriesItem['tx'];

            $xmlWriter->startElement('c:ser');

            $xmlWriter->writeElementBlock('c:idx', 'val', $index);
            $xmlWriter->writeElementBlock('c:order', 'val', $index);

            if ($this->options['type'] === 'pie') {
                $this->writePieChartLabel($xmlWriter);
            }

            if (isset($this->options['scatter'])) {
                $this->writeShape($xmlWriter);
            }

            if (isset($this->options['smooth'])) {
                $xmlWriter->writeElementBlock('c:smooth', 'val', $this->options['smooth']);
            }

            foreach($values as &$value) {
                if (!is_numeric($value)) {
                    $value = 0;
                }
            }

            if ($scatter === true) {
                $this->writeSeriesItem($xmlWriter, 'xVal', $categories);
                $this->writeSeriesItem($xmlWriter, 'yVal', $values);
            } else {
                $this->writeSeriesItem($xmlWriter, 'cat', $categories);
                $this->writeSeriesItem($xmlWriter, 'val', $values);
            }
            if ($tx) {
                $this->writeSeriesItem($xmlWriter, 'tx', $tx);
            }

            $xmlWriter->endElement(); // c:ser
            $index++;
        }

    }

    /**
     * Write series.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @param bool $scatter
     * @return void
     */
    private function writeSeries(XMLWriter $xmlWriter, $scatter = false)
    {
        $series = $this->element->getSeries();

        $index = 0;
        foreach ($series as $seriesItem) {
            $categories = $seriesItem['categories'];
            $values = $seriesItem['values'];

            $xmlWriter->startElement('c:ser');

            $xmlWriter->writeElementBlock('c:idx', 'val', $index);
            $xmlWriter->writeElementBlock('c:order', 'val', $index);

            if ($this->options['type'] === 'pie') {
                $this->writePieChartLabel($xmlWriter);
            }

            if (isset($this->options['scatter'])) {
                $this->writeShape($xmlWriter);
            }

            if ($scatter === true) {
                $this->writeSeriesItem($xmlWriter, 'xVal', $categories);
                $this->writeSeriesItem($xmlWriter, 'yVal', $values);
            } else {
                $this->writeSeriesItem($xmlWriter, 'cat', $categories);
                $this->writeSeriesItem($xmlWriter, 'val', $values);
            }

            $xmlWriter->endElement(); // c:ser
            $index++;
        }

    }

    /**
     * Write series items.
     *
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @param string $type
     * @param array $values
     * @return void
     */
    private function writeSeriesItem(XMLWriter $xmlWriter, $type, $value)
    {
        $types = array(
            'cat' => array('c:cat', 'c:strLit'),
            'val' => array('c:val', 'c:numLit'),
            'xVal' => array('c:xVal', 'c:strLit'),
            'yVal' => array('c:yVal', 'c:numLit'),
            'tx' => array('c:tx'),
        );
        list($itemType, $itemLit) = $types[$type];

        $xmlWriter->startElement($itemType);

        if ($itemLit) {
            $xmlWriter->startElement($itemLit);
            $xmlWriter->writeElementBlock('c:ptCount', 'val', count($value));

            $index = 0;
            foreach ($value as $val) {
                $xmlWriter->startElement('c:pt');
                $xmlWriter->writeAttribute('idx', $index);

                $xmlWriter->startElement('c:v');
                $xmlWriter->writeRaw($val);
                $xmlWriter->endElement(); // c:v

                $xmlWriter->endElement(); // c:pt
                $index++;
            }

            $xmlWriter->endElement(); // $itemLit
        } else {
            $xmlWriter->startElement('c:v');
            $xmlWriter->writeRaw($value);
            $xmlWriter->endElement(); // c:v
        }

        $xmlWriter->endElement(); // $itemType
    }

    /**
     * Write axis
     *
     * @link http://www.datypic.com/sc/ooxml/t-draw-chart_CT_CatAx.html
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @param string $type
     * @return void
     */
    private function writeAxis(XMLWriter $xmlWriter, $type)
    {
        $types = array(
            'cat' => array('c:catAx', 1, 'b', 2, $this->element->getXAxisTitle()),
            'val' => array('c:valAx', 2, 'l', 1, $this->element->getYAxisTitle()),
        );
        list($axisType, $axisId, $axisPos, $axisCross, $title) = $types[$type];

        $xmlWriter->startElement($axisType);

        $xmlWriter->writeElementBlock('c:axId', 'val', $axisId);
        $xmlWriter->writeElementBlock('c:axPos', 'val', $axisPos);
        $xmlWriter->writeElementBlock('c:crossAx', 'val', $axisCross);
        $xmlWriter->writeElementBlock('c:auto', 'val', 1);

        if (isset($this->options['axes'])) {
            $xmlWriter->writeElementBlock('c:delete', 'val', 0);
            $xmlWriter->writeElementBlock('c:majorTickMark', 'val', 'out');
            $xmlWriter->writeElementBlock('c:minorTickMark', 'val', 'none');
            $xmlWriter->writeElementBlock('c:tickLblPos', 'val', 'nextTo'); // nextTo
            $xmlWriter->writeElementBlock('c:crosses', 'val', 'autoZero');
        }
        if (isset($this->options['radar'])) {
            $xmlWriter->writeElement('c:majorGridlines');
        }

        // draw grid line
        if ($axisType === 'c:valAx') {
            $xmlWriter->startElement('c:majorGridlines');
            $xmlWriter->endElement();
        }

        // write axis title
        if (!empty($title)) {
            $this->writeTitleBlock($xmlWriter, $title);
        }

        $xmlWriter->startElement('c:scaling');
        $xmlWriter->writeElementBlock('c:orientation', 'val', 'minMax');
        $xmlWriter->endElement(); // c:scaling

        $this->writeShape($xmlWriter, true);

        $xmlWriter->endElement(); // $axisType
    }

    /**
     * Write pie chart's value label
     *
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @return void
     */
    private function writePieChartLabel(XMLWriter $xmlWriter)
    {
        $xmlWriter->startElement('c:dLbls');
        $xmlWriter->writeElementBlock('c:dLblPos', 'val', "ctr");
        $xmlWriter->writeElementBlock('c:showLegendKey', 'val', 0);
        $xmlWriter->writeElementBlock('c:showVal', 'val', 0);
        $xmlWriter->writeElementBlock('c:showCatName', 'val', 0);
        $xmlWriter->writeElementBlock('c:showSerName', 'val', 0);
        $xmlWriter->writeElementBlock('c:showPercent', 'val', 1);
        $xmlWriter->writeElementBlock('c:showBubbleSize', 'val', 0);
        $xmlWriter->writeElementBlock('c:showLeaderLines', 'val', 1);
        $xmlWriter->startElement('c:extLst');
        $xmlWriter->writeElementBlock('c:ext', array(
            'uri' => "{CE6537A1-D6FC-4f65-9D91-7224C49458BB}",
            'xmlns:c15' => "http://schemas.microsoft.com/office/drawing/2012/chart"));
        $xmlWriter->endElement(); // c:extLst
        $xmlWriter->endElement(); // c:dLbls
    }

    /**
     * Write a title block
     *
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @param string $title
     * @return void
     */
    private function writeTitleBlock(XMLWriter $xmlWriter, $title)
    {
        $xmlWriter->startElement('c:title');
        $xmlWriter->startElement('c:tx');
        $xmlWriter->startElement('c:rich');
        $xmlWriter->startElement('a:bodyPr');
        $xmlWriter->endElement(); // a:bodyPr
        $xmlWriter->startElement('a:lstStyle');
        $xmlWriter->endElement();
        $xmlWriter->startElement('a:p');
        $xmlWriter->startElement('a:aPr');
        $xmlWriter->startElement('a:defRPr');
        $xmlWriter->endElement(); // a:defRPr
        $xmlWriter->endElement(); // a:pPr
        $xmlWriter->startElement('a:r');
        $xmlWriter->startElement('a:rPr');
        $xmlWriter->endElement(); // a:rPr
        $xmlWriter->startElement('a:t');
        $xmlWriter->writeRaw($title);
        $xmlWriter->endElement(); // a:t
        $xmlWriter->endElement(); // a:r
        $xmlWriter->startElement('a:endParaRPr');
        $xmlWriter->endElement(); // a:endParaRPr
        $xmlWriter->endElement(); // a:p
        $xmlWriter->endElement(); // c:rich
        $xmlWriter->endElement(); // c:tx
        $xmlWriter->writeElementBlock('c:overlay', 'val', 0);
        $xmlWriter->endElement(); // c:title
    }

    /**
     * Write legend
     *
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @return void
     */
    private function writeLegend(XMLWriter $xmlWriter)
    {
        $xmlWriter->startElement('c:legend');
        $xmlWriter->writeElementBlock('c:legendPos', 'val', "r");
        $xmlWriter->writeElementBlock('c:overlay', 'val', 0);
        $xmlWriter->endElement();
    }

    /**
     * Write shape
     *
     * @link http://www.datypic.com/sc/ooxml/t-a_CT_ShapeProperties.html
     * @param \PhpOffice\PhpWord\Shared\XMLWriter $xmlWriter
     * @param bool $line
     * @return void
     */
    private function writeShape(XMLWriter $xmlWriter, $line = false)
    {
        $xmlWriter->startElement('c:spPr');
        $xmlWriter->startElement('a:ln');
        if ($line === true) {
            $xmlWriter->writeElement('a:solidFill');
        } else {
            $xmlWriter->writeElement('a:noFill');
        }
        $xmlWriter->endElement(); // a:ln
        $xmlWriter->endElement(); // c:spPr
    }
}
